/*  Canon LIPS IV Driver for PDAPI Ver. 1.0.0
 *  Copyright CANON INC. 2004
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#include 	<sys/types.h>
#include 	<unistd.h>
#include 	<stdio.h>
#include 	<stdlib.h>
#include 	<errno.h>
#include 	<string.h>
#include 	<stdarg.h>

#include	"pddrv.h"
#include	"pdlips.h"
#include	"pdcommon.h"

extern	int				errorno;
extern	int				opvpErrorNo;


pdColorSpace	available_colorSpace[] = {pdcspaceStandardRGB, pdcspaceBW, pdcspaceDeviceGray};

int				available_ROP[] = {SRCCOPY, SRCPAINT};

int lipsResetCTM(int printerContext)
{
	GraphicsStateData	*gs;

	dbgMessage("vdapi : %s\n", __func__);

	gs = GetGraphicsState(printerContext);
	if (!gs)
		return (pdERROR);

	InitializeCTM(&gs->ctm);

	return (pdOK);
}

int lipsSetCTM(int printerContext, pdCTM *pCTM)
{
	GraphicsStateData	*gs;

	dbgMessage("vdapi : %s  a=%f, b=%f c=%f d=%f e=%f f=%f\n", __func__, pCTM->a, pCTM->b, pCTM->c, pCTM->d, pCTM->e, pCTM->f);

	gs = GetGraphicsState(printerContext);
	if (!gs)
		return (pdERROR);

	gs->ctm.a = pCTM->a, gs->ctm.b = pCTM->b;
	gs->ctm.c = pCTM->c, gs->ctm.d = pCTM->d;
	gs->ctm.e = pCTM->e, gs->ctm.f = pCTM->f;

	return (pdOK);
}

int lipsGetCTM(int printerContext, pdCTM *pCTM)
{
	GraphicsStateData	*gs;

	dbgMessage("vdapi : %s\n", __func__);

	gs = GetGraphicsState(printerContext);
	if (!gs)
		return (pdERROR);

	pCTM->a = gs->ctm.a, pCTM->b = gs->ctm.b;
	pCTM->c = gs->ctm.c, pCTM->d = gs->ctm.d;
	pCTM->e = gs->ctm.e, pCTM->f = gs->ctm.f;

	return (pdOK);
}

int lipsInitGS(int printerContext)
{
	GraphicsStateData	*gs;

	dbgMessage("vdapi : %s\n", __func__);

	gs = GetGraphicsState(printerContext);
	if (!gs)
		return (pdERROR);

	CleanGraphicsState(gs);
	InitializeGraphicsState(gs);
	return (pdOK);
}

int lipsSaveGS(int printerContext)
{
	printerContextData		*pcd;
	GraphicsStateDataList	*src, *dst;

	dbgMessage("vdapi : %s\n", __func__);

	pcd = GetPrinterContext(printerContext);
	if (!pcd)
		return (pdERROR);

	src = &pcd->gslist;
	dst = (GraphicsStateDataList *)malloc(sizeof(GraphicsStateDataList));
	if (!dst) {
		errorno = opvpErrorNo = pdFATALERROR;
		return (pdERROR);
	}

	memset(dst, 0, sizeof(GraphicsStateDataList));

	if (!CopyGraphicsState(dst, src)) {
		errorno = opvpErrorNo = pdFATALERROR;
		return (pdERROR);
	}

	dst->next = pcd->gslist.next;
	pcd->gslist.next = dst;

	return (pdOK);
}

int lipsRestoreGS(int printerContext)
{
	printerContextData	*pcd;
	GraphicsStateDataList	*src, *dst;

	dbgMessage("vdapi : %s\n", __func__);

	pcd = GetPrinterContext(printerContext);
	if (!pcd)
		return (pdERROR);

	if (pcd->gslist.next == NULL) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	dst = &pcd->gslist;
	src = pcd->gslist.next;

	if (!CopyGraphicsState(dst, src)) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	pcd->gslist.next = src->next;
	CleanGraphicsState(&src->gsdata);
	free(src);

	if (pcd->gslist.gsdata.clipPath_registered) {
		if (!pdlips_Set_Clip(pcd, pcd->gslist.gsdata.clipPath_id))
			return (pdERROR);
	}

	return (pdOK);
}

int lipsQueryColorSpace(int printerContext, pdColorSpace *pspace, int *pnum)
{
	GraphicsStateData	*gs;
	int					i, num;

	dbgMessage("vdapi : %s *pnum = %d\n", __func__, *pnum);

	gs = GetGraphicsState(printerContext);
	if (!gs)
		return (pdERROR);

	num = sizeof(available_colorSpace)/sizeof(pdColorSpace);
	if (num > *pnum) {
		errorno = opvpErrorNo = pdPARAMERROR;
		*pnum = num;
		return (pdERROR);
	}

	for (i = 0; i < num; i++)
		pspace[i] = available_colorSpace[i];

	*pnum = num;

	return (pdOK);
}

int lipsQueryColorSpace1(int printerContext, int *pnum, pdColorSpace *pspace)
{
	GraphicsStateData	*gs;
	int					i, num;

	dbgMessage("vdapi : %s *pnum = %d\n", __func__, *pnum);

	gs = GetGraphicsState(printerContext);
	if (!gs)
		return (pdERROR);

	num = sizeof(available_colorSpace)/sizeof(pdColorSpace);
	if (num > *pnum) {
		errorno = opvpErrorNo = pdPARAMERROR;
		*pnum = num;
		return (pdERROR);
	}

	for (i = 0; i < num; i++)
		pspace[i] = available_colorSpace[i];

	*pnum = num;

	return (pdOK);
}

int lipsSetColorSpace(int printerContext, pdColorSpace cspace)
{
	GraphicsStateData	*gs;
	int					i, num;

	dbgMessage("vdapi : %s colorspcae = %d\n", __func__, cspace);

	gs = GetGraphicsState(printerContext);
	if (!gs)
		return (pdERROR);

	num = sizeof(available_colorSpace)/sizeof(pdColorSpace);
	for (i = 0; i < num; i++) {
		if (cspace == available_colorSpace[i])
			break;
	}

	if (i >= num) {
		errorno = opvpErrorNo = pdPARAMERROR;
		return (pdERROR);
	}

	gs->colorspace = cspace;
	return (pdOK);
}

int lipsGetColorSpace(int printerContext, pdColorSpace *pcspace)
{
	GraphicsStateData	*gs;

	dbgMessage("vdapi : %s\n", __func__);

	gs = GetGraphicsState(printerContext);
	if (!gs)
		return (pdERROR);

	*pcspace = gs->colorspace;
	return (pdOK);
}

int lipsQueryROP(int printerContext, int *pnum, int *prop)
{
	GraphicsStateData	*gs;
	int					i, num;

	dbgMessage("vdapi : %s *pnum = %d\n", __func__, *pnum);

	gs = GetGraphicsState(printerContext);
	if (!gs)
		return (pdERROR);

	if (!prop) {
		*pnum = sizeof(available_ROP)/sizeof(int);
		return (pdOK);
	}

	num = sizeof(available_ROP)/sizeof(int);
	if (num > *pnum) {
		*pnum = num;
		errorno = opvpErrorNo = pdPARAMERROR;
		return (pdERROR);
	}

	for (i=0; i<num; i++)
		prop[i] = available_ROP[i];
	*pnum = num;

	return (pdOK);
}

int lipsSetROP(int printerContext, int rop)
{
	GraphicsStateData	*gs;
	int			i, num;

	dbgMessage("vdapi : %s rop=%x\n", __func__, rop);

	gs = GetGraphicsState(printerContext);
	if (!gs)
		return (pdERROR);

	num = sizeof(available_ROP)/sizeof(int);
	for (i = 0; i < num; i++) {
		if (rop == available_ROP[i])
			break;
	}

	if (i >= num) {
		errorno = opvpErrorNo = pdPARAMERROR;
		return (pdERROR);
	}

#if 0
	gs->rop = rop;
#else
	gs->rop = SRCPAINT;
#endif

	return (pdOK);
}

int lipsGetROP(int printerContext, int *prop)
{
	GraphicsStateData	*gs;

	dbgMessage("vdapi : %s\n", __func__);

	gs = GetGraphicsState(printerContext);
	if (!gs)
		return (pdERROR);

	*prop = gs->rop;
	return (pdOK);
}

int lipsSetPaintMode(int printerContext, pdPaintMode pmode)
{
	GraphicsStateData	*gs;

	dbgMessage("vdapi : %s\n", __func__);

	gs = GetGraphicsState(printerContext);
	if (!gs)
		return (pdERROR);

	gs->paintmode = pmode;
	return (pdOK);
}

int lipsGetPaintMode(int printerContext, pdPaintMode *ppmode)
{
	GraphicsStateData	*gs;

	dbgMessage("vdapi : %s\n", __func__);

	gs = GetGraphicsState(printerContext);
	if (!gs)
		return (pdERROR);

	*ppmode = gs->paintmode;
	return (pdOK);
}


int lipsSetFillMode(int printerContext, pdFillMode fillmode)
{
	GraphicsStateData	*gs;

	dbgMessage("vdapi : %s\n", __func__);

	gs = GetGraphicsState(printerContext);
	if (!gs)
		return (pdERROR);

	if ((fillmode != pdFillModeEvenOdd) && (fillmode != pdFillModeWinding)) {
		errorno = opvpErrorNo = pdPARAMERROR;
		return (pdERROR);
	}

	gs->fillmode = fillmode;
	return (pdOK);
}

int lipsGetFillMode(int printerContext, pdFillMode *pfillmode)
{
	GraphicsStateData	*gs;

	dbgMessage("vdapi : %s\n", __func__);

	gs = GetGraphicsState(printerContext);
	if (!gs)
		return (pdERROR);

	*pfillmode = gs->fillmode;
	return (pdOK);
}

int lipsSetLineWidth(int printerContext, pdFix width)
{
	GraphicsStateData	*gs;

	dbgMessage("vdapi : %s\n", __func__);

	gs = GetGraphicsState(printerContext);
	if (!gs)
		return (pdERROR);

	gs->linewidth = width;
	return (pdOK);
}

int lipsGetLineWidth(int printerContext, pdFix *pwidth)
{
	GraphicsStateData	*gs;

	dbgMessage("vdapi : %s\n", __func__);

	gs = GetGraphicsState(printerContext);
	if (!gs)
		return (pdERROR);

	*pwidth = gs->linewidth;
	return (pdOK);
}

int lipsSetLineDash(int printerContext, pdFix pdash[], int num)
{
	GraphicsStateData	*gs;
	int					i, rep;

	dbgMessage("vdapi : %s\n", __func__);

	gs = GetGraphicsState(printerContext);
	if (!gs)
		return (pdERROR);

	if (num > LINE_DASH_MAX) {
		errorno = opvpErrorNo = pdPARAMERROR;
		return (pdERROR);
	}

	rep = 1;
	if ((num % 2) != 0) {
		if (num*2 > LINE_DASH_MAX) {
			errorno = opvpErrorNo = pdPARAMERROR;
			return (pdERROR);
		}
		rep = 2;
	}

	gs->dashnum = num * rep;
	for (i = 0; i < num*rep; i++)
		gs->pDash[i] = pdash[(i % num)];

	return (pdOK);
}

int lipsSetLineDash1(int printerContext, int num, pdFix pdash[])
{
	GraphicsStateData	*gs;
	int					i, rep;

	dbgMessage("vdapi : %s\n", __func__);

	gs = GetGraphicsState(printerContext);
	if (!gs)
		return (pdERROR);

	if (num > LINE_DASH_MAX) {
		errorno = opvpErrorNo = pdPARAMERROR;
		return (pdERROR);
	}

	rep = 1;
	if ((num % 2) != 0) {
		if (num*2 > LINE_DASH_MAX) {
			errorno = opvpErrorNo = pdPARAMERROR;
			return (pdERROR);
		}
		rep = 2;
	}

	gs->dashnum = num * rep;
	for (i = 0; i < num*rep; i++)
		gs->pDash[i] = pdash[(i % num)];

	return (pdOK);
}

int lipsGetLineDash(int printerContext, pdFix pdash[], int *pnum)
{
	GraphicsStateData	*gs;
	int			i;

	dbgMessage("vdapi : %s\n", __func__);

	gs = GetGraphicsState(printerContext);
	if (!gs)
		return (pdERROR);

	*pnum = gs->dashnum;
	for (i = 0; i < gs->dashnum; i++)
		pdash[i] = gs->pDash[i];

	return (pdOK);
}

int lipsGetLineDash1(int printerContext, int *pnum, pdFix pdash[])
{
	GraphicsStateData	*gs;
	int			i;

	dbgMessage("vdapi : %s\n", __func__);

	gs = GetGraphicsState(printerContext);
	if (!gs)
		return (pdERROR);

	*pnum = gs->dashnum;
	for (i = 0; i < gs->dashnum; i++)
		pdash[i] = gs->pDash[i];

	return (pdOK);
}

int lipsSetLineDashOffset(int printerContext, pdFix offset)
{
	GraphicsStateData	*gs;

	dbgMessage("vdapi : %s\n", __func__);

	gs = GetGraphicsState(printerContext);
	if (!gs)
		return (pdERROR);

	gs->dashoffset = offset;

	return (pdOK);
}

int lipsGetLineDashOffset(int printerContext, pdFix *poffset)
{
	GraphicsStateData	*gs;

	dbgMessage("vdapi : %s\n", __func__);

	gs = GetGraphicsState(printerContext);
	if (!gs)
		return (pdERROR);

	*poffset = gs->dashoffset;

	return (pdOK);
}

int lipsSetLineStyle(int printerContext, pdLineStyle linestyle)
{
	GraphicsStateData	*gs;

	dbgMessage("vdapi : %s\n", __func__);

	gs = GetGraphicsState(printerContext);
	if (!gs)
		return (pdERROR);

	if ((linestyle != pdlineStyleSolid) && (linestyle != pdlineStyleDash)) {
		errorno = opvpErrorNo = pdPARAMERROR;
		return (pdERROR);
	}

	gs->linestyle = linestyle;
	return (pdOK);
}

int lipsGetLineStyle(int printerContext, pdLineStyle *plinestyle)
{
	GraphicsStateData	*gs;

	dbgMessage("vdapi : %s\n", __func__);

	gs = GetGraphicsState(printerContext);
	if (!gs)
		return (pdERROR);

	*plinestyle = gs->linestyle;

	return (pdOK);
}

int lipsSetLineCap(int printerContext, pdLineCap linecap)
{
	GraphicsStateData	*gs;

	dbgMessage("vdapi : %s\n", __func__);

	gs = GetGraphicsState(printerContext);
	if (!gs)
		return (pdERROR);

	if ((linecap != pdlineCapButt) && (linecap != pdlineCapRound) && (linecap != pdlineCapSquare)) {
		errorno = opvpErrorNo = pdPARAMERROR;
		return (pdERROR);
	}

	gs->linecap = linecap;
	return (pdOK);
}

int lipsGetLineCap(int printerContext, pdLineCap *linecap)
{
	GraphicsStateData	*gs;

	dbgMessage("vdapi : %s\n", __func__);

	gs = GetGraphicsState(printerContext);
	if (!gs)
		return (pdERROR);

	*linecap = gs->linecap;

	return (pdOK);
}

int lipsSetLineJoin(int printerContext, pdLineJoin linejoin)
{
	GraphicsStateData	*gs;

	dbgMessage("vdapi : %s\n", __func__);

	gs = GetGraphicsState(printerContext);
	if (!gs)
		return (pdERROR);

	if ((linejoin != pdlineJoinMiter) && (linejoin != pdlineJoinRound) && (linejoin != pdlineJoinBevel)) {
		errorno = opvpErrorNo = pdPARAMERROR;
		return (pdERROR);
	}

	gs->linejoin = linejoin;
	return (pdOK);
}

int lipsGetLineJoin(int printerContext, pdLineJoin *plinejoin)
{
	GraphicsStateData	*gs;

	dbgMessage("vdapi : %s\n", __func__);

	gs = GetGraphicsState(printerContext);
	if (!gs)
		return (pdERROR);

	*plinejoin = gs->linejoin;

	return (pdOK);
}

int lipsSetMiterLimit(int printerContext, pdFix miterlimit)
{
	GraphicsStateData	*gs;

	dbgMessage("vdapi : %s\n", __func__);

	gs = GetGraphicsState(printerContext);
	if (!gs)
		return (pdERROR);

	gs->miterlimit = miterlimit;

	return (pdOK);
}

int lipsGetMiterLimit(int printerContext, pdFix *pmiterlimit)
{
	GraphicsStateData	*gs;

	dbgMessage("vdapi : %s\n", __func__);

	gs = GetGraphicsState(printerContext);
	if (!gs)
		return (pdERROR);

	*pmiterlimit = gs->miterlimit;

	return (pdOK);
}

int lipsSetStrokeColor(int printerContext, pdBrush *brush)
{
	GraphicsStateData	*gs;

	dbgMessage("vdapi : %s\n", __func__);

	gs = GetGraphicsState(printerContext);
	if (!gs)
		return (pdERROR);

	if (CopyBrush(&gs->strokecolor, brush))
		return (pdOK);
	else
		return (pdERROR);
}

int lipsSetFillColor(int printerContext, pdBrush *brush)
{
	GraphicsStateData	*gs;

	dbgMessage("vdapi : %s\n", __func__);

	gs = GetGraphicsState(printerContext);
	if (!gs)
		return (pdERROR);

	if (CopyBrush(&gs->fillcolor, brush))
		return (pdOK);
	else
		return (pdERROR);
}

