/*  Canon LIPS IV Driver for PDAPI Ver. 1.0.0
 *  Copyright CANON INC. 2004
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#include 	<sys/types.h>
#include 	<unistd.h>
#include 	<stdio.h>
#include 	<stdlib.h>
#include 	<errno.h>
#include 	<string.h>
#include 	<stdarg.h>

#include	"pddrv.h"
#include	"pdlips.h"
#include	"pdcommon.h"

extern	int	errorno;
extern	int	opvpErrorNo;

int lipsDrawImage(
	int 			printerContext,
	int				sourceWidth,
	int				sourceHeight,
	int				colorDepth,
	pdImageFormat	imageFormat,
	pdRectangle		destinationSize,
	int				count,
	void			*imageData)
{
	printerContextData	*pcd;
	GraphicsStateData	*gs;
	unsigned char		buf[64], param[48];
	int					size;
	pdFix				dstWidth, dstHeight;
	unsigned char		bits, format, process;
	int					i, imgpitch, srcpitch;
	char				*ptr;
	pdCTM				ctm;

	dbgMessage("vdapi : %s width = %d, height = %d, depth = %d, count = %d\n", __func__, sourceWidth, sourceHeight, colorDepth, count);

	if (!count)
		return (pdOK);

	pcd = GetPrinterContext(printerContext);
	if (!pcd)
		return (pdERROR);

	gs = &pcd->gslist.gsdata;

	if (!pcd->page_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if (pcd->drawimage_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if (!pdlips_Set_ROP(pcd))
		return (pdERROR);

	if (colorDepth == 1) {
		if (pcd->pageForm.color_mode) {
			pdConvParam(param, pdConvColorValue(gs->fillcolor.color[2]));
			pdConvParamAdd(param, pdConvColorValue(gs->fillcolor.color[1]));
			pdConvParamAdd(param, pdConvColorValue(gs->fillcolor.color[0]));
		} else {
			pdConvParam(param, pdConvGrayValue(gs->fillcolor.color[2],
						gs->fillcolor.color[1], gs->fillcolor.color[0]));
		}
		size = snprintf(buf, sizeof(buf), "}T%s%c", param, C_IS2);
		if (!pdWrite(pcd, buf, size))
			return (pdERROR);
	}

	ctm.a = gs->ctm.a, ctm.b = gs->ctm.b;
	ctm.c = gs->ctm.c, ctm.d = gs->ctm.d;
	ctm.e = ctm.f = 0;

	dstWidth = abs(destinationSize.p1.x - destinationSize.p0.x);
	dstHeight = abs(destinationSize.p1.y - destinationSize.p0.y);
	pdConvParam(param, fix2int(pcd->currentPoint_x));
	pdConvParamAdd(param, fix2int(pcd->currentPoint_y));
	pdConvParamAdd(param, fix2int(pcd->currentPoint_x + Dev_x_fConvF(dstWidth, dstHeight, &ctm)));
	pdConvParamAdd(param, fix2int(pcd->currentPoint_y + Dev_y_fConvF(dstWidth, dstHeight, &ctm)));
	pdConvParamAdd(param, fix2int(pcd->currentPoint_x + Dev_x_fConvF(0, dstHeight, &ctm)));
	pdConvParamAdd(param, fix2int(pcd->currentPoint_y + Dev_y_fConvF(0, dstHeight, &ctm)));
	pdConvParamAdd(param, sourceHeight);
	pdConvParamAdd(param, sourceWidth);

	if (colorDepth == 1) {
		bits = 0x31, format = 0x30,	process = 0x30;
	} else if (colorDepth == 8) {
		bits = 0x38, format = 0x30, process = 0x30;
	} else if (colorDepth == 24) {
		bits = 0x38, format = 0x3a, process = 0x31;
	} else {
		errorno = opvpErrorNo = pdNOTSUPPORTED;
		return (pdERROR);
	}

	size = snprintf(buf, sizeof(buf), "}U%s%c%c%c1%c", param, bits, format, process, C_IS2);
	if (!pdWrite(pcd, buf, size))
		return (pdERROR);

	if (colorDepth == 1)
		imgpitch = (sourceWidth + 7)/8;
	 else
		imgpitch = sourceWidth * colorDepth / 8;

	pdConvParam(param, imgpitch * sourceHeight);
	size = snprintf (buf, sizeof(buf), "}Q110%s%c", param, C_IS2);
	if (!pdWrite(pcd, buf, size))
		return (pdERROR);

	srcpitch = count / sourceHeight;
	ptr = imageData;
	for (i = 0; i < sourceHeight; i++) {
		if (!pdWrite(pcd, ptr, imgpitch))
			return (pdERROR);
		ptr += srcpitch;
	}

	dbgMessage("vdapi : %s return OK\n", __func__);
	return (pdOK);
}


int lipsDrawImage1(
	int 			printerContext,
	int				sourceWidth,
	int				sourceHeight,
	int				sourcePitch,
	pdImageFormat	imageFormat,
	int				destinationWidth,
	int				destinationHeight,
	void			*imageData)
{
	printerContextData	*pcd;
	GraphicsStateData	*gs;
	unsigned char		buf[64], param[48];
	int					size;
	pdFix				dstWidth, dstHeight;
	unsigned char		bits, format, process;
	int					i, imgpitch;
	char				*ptr;
	pdCTM				ctm;
	pdColorSpace		cs;
	int					colorDepth;

	dbgMessage("vdapi : %s width = %d, height = %d, \n", __func__, sourceWidth, sourceHeight);

	pcd = GetPrinterContext(printerContext);
	if (!pcd)
		return (pdERROR);

	gs = &pcd->gslist.gsdata;

	if (!pcd->page_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if (pcd->drawimage_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if (!pdlips_Set_ROP(pcd))
		return (pdERROR);

	if (imageFormat == pdiformatRaw) {
		lipsGetColorSpace(printerContext, &cs);
		switch (cs) {
			case pdcspaceBW:
				colorDepth = 1;
				break;
			case pdcspaceDeviceGray:
				colorDepth = 8;
				break;
			case pdcspaceStandardRGB:
			defalult:
				colorDepth = 24;
		}
	} else if (imageFormat == pdiformatMask ) {
		colorDepth = 1;
	}

	if (colorDepth == 1) {
		if (pcd->pageForm.color_mode) {
			pdConvParam(param, pdConvColorValue(gs->fillcolor.color[2]));
			pdConvParamAdd(param, pdConvColorValue(gs->fillcolor.color[1]));
			pdConvParamAdd(param, pdConvColorValue(gs->fillcolor.color[0]));
		} else {
			pdConvParam(param, pdConvGrayValue(gs->fillcolor.color[2],
						gs->fillcolor.color[1], gs->fillcolor.color[0]));
		}
		size = snprintf(buf, sizeof(buf), "}T%s%c", param, C_IS2);
		if (!pdWrite(pcd, buf, size))
			return (pdERROR);
	}

	ctm.a = gs->ctm.a, ctm.b = gs->ctm.b;
	ctm.c = gs->ctm.c, ctm.d = gs->ctm.d;
	ctm.e = ctm.f = 0;

	dstWidth = int2fix(destinationWidth);
	dstHeight = int2fix(destinationHeight);

	pdConvParam(param, fix2int(pcd->currentPoint_x));
	pdConvParamAdd(param, fix2int(pcd->currentPoint_y));
	pdConvParamAdd(param, fix2int(pcd->currentPoint_x + Dev_x_fConvF(dstWidth, dstHeight, &ctm)));
	pdConvParamAdd(param, fix2int(pcd->currentPoint_y + Dev_y_fConvF(dstWidth, dstHeight, &ctm)));
	pdConvParamAdd(param, fix2int(pcd->currentPoint_x + Dev_x_fConvF(0, dstHeight, &ctm)));
	pdConvParamAdd(param, fix2int(pcd->currentPoint_y + Dev_y_fConvF(0, dstHeight, &ctm)));
	pdConvParamAdd(param, sourceHeight);
	pdConvParamAdd(param, sourceWidth);

	if (colorDepth == 1) {
		bits = 0x31, format = 0x30,	process = 0x30;
	} else if (colorDepth == 8) {
		bits = 0x38, format = 0x30, process = 0x30;
	} else if (colorDepth == 24) {
		bits = 0x38, format = 0x3a, process = 0x31;
	} else {
		errorno = opvpErrorNo = pdNOTSUPPORTED;
		return (pdERROR);
	}

	size = snprintf(buf, sizeof(buf), "}U%s%c%c%c1%c", param, bits, format, process, C_IS2);
	if (!pdWrite(pcd, buf, size))
		return (pdERROR);

	if (colorDepth == 1)
		imgpitch = (sourceWidth + 7)/8;
	 else
		imgpitch = sourceWidth * colorDepth / 8;

	pdConvParam(param, imgpitch * sourceHeight);
	size = snprintf (buf, sizeof(buf), "}Q110%s%c", param, C_IS2);
	if (!pdWrite(pcd, buf, size))
		return (pdERROR);

	ptr = imageData;
	for (i = 0; i < sourceHeight; i++) {
		if (!pdWrite(pcd, ptr, imgpitch))
			return (pdERROR);
		ptr += sourcePitch;
	}

	dbgMessage("vdapi : %s return OK\n", __func__);
	return (pdOK);
}


int lipsStartDrawImage(
	int				printerContext,
	int				sourceWidth,
	int				sourceHeight,
	int				colorDepth,
	pdImageFormat	imageFormat,
	pdRectangle		destinationSize)
{
	printerContextData	*pcd;
	GraphicsStateData	*gs;
	unsigned char		buf[64], param[48];
	int					size;
	pdFix				dstWidth, dstHeight;
	unsigned char		bits, format, process;
	pdCTM				ctm;

	dbgMessage("vdapi : %s width = %d, height = %d depth = %d\n", __func__, sourceWidth, sourceHeight, colorDepth);

	pcd = GetPrinterContext(printerContext);
	if (!pcd)
		return (pdERROR);

	gs = &pcd->gslist.gsdata;

	if (!pcd->page_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if (pcd->drawimage_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if (!pdlips_Set_ROP(pcd))
		return (pdERROR);

	if (colorDepth == 1) {
		if (pcd->pageForm.color_mode) {
			pdConvParam(param, pdConvColorValue(gs->fillcolor.color[2]));
			pdConvParamAdd(param, pdConvColorValue(gs->fillcolor.color[1]));
			pdConvParamAdd(param, pdConvColorValue(gs->fillcolor.color[0]));
		} else {
			pdConvParam(param, pdConvGrayValue(gs->fillcolor.color[2],
						gs->fillcolor.color[1], gs->fillcolor.color[0]));
		}
		size = snprintf(buf, sizeof(buf), "}T%s%c", param, C_IS2);
		if (!pdWrite(pcd, buf, size))
			return (pdERROR);
	}

	ctm.a = gs->ctm.a, ctm.b = gs->ctm.b;
	ctm.c = gs->ctm.c, ctm.d = gs->ctm.d;
	ctm.e = ctm.f = 0;

	dstWidth = abs(destinationSize.p1.x - destinationSize.p0.x);
	dstHeight = abs(destinationSize.p1.y - destinationSize.p0.y);

	pdConvParam(param, fix2int(pcd->currentPoint_x));
	pdConvParamAdd(param, fix2int(pcd->currentPoint_y));
	pdConvParamAdd(param, fix2int(pcd->currentPoint_x + Dev_x_fConvF(dstWidth, dstHeight, &ctm)));
	pdConvParamAdd(param, fix2int(pcd->currentPoint_y + Dev_y_fConvF(dstWidth, dstHeight, &ctm)));
	pdConvParamAdd(param, fix2int(pcd->currentPoint_x + Dev_x_fConvF(0, dstHeight, &ctm)));
	pdConvParamAdd(param, fix2int(pcd->currentPoint_y + Dev_y_fConvF(0, dstHeight, &ctm)));
	pdConvParamAdd(param, sourceHeight);
	pdConvParamAdd(param, sourceWidth);

	if (colorDepth == 1) {
		bits = 0x31, format = 0x30,	process = 0x30;
	} else if ((colorDepth == 8) || (colorDepth == 24)) {
		bits = 0x38, format = 0x3a, process = 0x31;
	} else {
		errorno = opvpErrorNo = pdNOTSUPPORTED;
		return (pdERROR);
	}

	size = snprintf(buf, sizeof(buf), "}U%s%c%c%c0%c", param, bits, format, process, C_IS2);
	if (!pdWrite(pcd, buf, size))
		return (pdERROR);

	pcd->imageHeight = sourceHeight;
	pcd->currentImageHeight = 0;
	pcd->imageWidth = sourceWidth;
	pcd->imageDepth = colorDepth;
	pcd->drawimage_process_flag = TRUE;

	return (pdOK);
}


int lipsStartDrawImage1(
	int				printerContext,
	int				sourceWidth,
	int				sourceHeight,
	int				sourcePitch,
	pdImageFormat	imageFormat,
	int				destinationWidth,
	int				destinationHeight)

{
	printerContextData	*pcd;
	GraphicsStateData	*gs;
	unsigned char		buf[64], param[48];
	int					size;
	pdFix				dstWidth, dstHeight;
	unsigned char		bits, format, process;
	pdCTM				ctm;
	pdColorSpace		cs;
	int					colorDepth;

	dbgMessage("vdapi : %s width = %d, height = %d \n", __func__, sourceWidth, sourceHeight);

	pcd = GetPrinterContext(printerContext);
	if (!pcd)
		return (pdERROR);

	gs = &pcd->gslist.gsdata;

	if (!pcd->page_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if (pcd->drawimage_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if (!pdlips_Set_ROP(pcd))
		return (pdERROR);

	if (imageFormat == pdiformatRaw) {
		lipsGetColorSpace(printerContext, &cs);
		switch (cs) {
			case pdcspaceBW:
				colorDepth = 1;
				break;
			case pdcspaceDeviceGray:
				colorDepth = 8;
				break;
			case pdcspaceStandardRGB:
			defalult:
				colorDepth = 24;
		}
	} else if (imageFormat == pdiformatMask ) {
		colorDepth = 1;
	}

	if (colorDepth == 1) {
		if (pcd->pageForm.color_mode) {
			pdConvParam(param, pdConvColorValue(gs->fillcolor.color[2]));
			pdConvParamAdd(param, pdConvColorValue(gs->fillcolor.color[1]));
			pdConvParamAdd(param, pdConvColorValue(gs->fillcolor.color[0]));
		} else {
			pdConvParam(param, pdConvGrayValue(gs->fillcolor.color[2],
						gs->fillcolor.color[1], gs->fillcolor.color[0]));
		}
		size = snprintf(buf, sizeof(buf), "}T%s%c", param, C_IS2);
		if (!pdWrite(pcd, buf, size))
			return (pdERROR);
	}

	ctm.a = gs->ctm.a, ctm.b = gs->ctm.b;
	ctm.c = gs->ctm.c, ctm.d = gs->ctm.d;
	ctm.e = ctm.f = 0;

	dstWidth = int2fix(destinationWidth);
	dstHeight = int2fix(destinationHeight);

	pdConvParam(param, fix2int(pcd->currentPoint_x));
	pdConvParamAdd(param, fix2int(pcd->currentPoint_y));
	pdConvParamAdd(param, fix2int(pcd->currentPoint_x + Dev_x_fConvF(dstWidth, dstHeight, &ctm)));
	pdConvParamAdd(param, fix2int(pcd->currentPoint_y + Dev_y_fConvF(dstWidth, dstHeight, &ctm)));
	pdConvParamAdd(param, fix2int(pcd->currentPoint_x + Dev_x_fConvF(0, dstHeight, &ctm)));
	pdConvParamAdd(param, fix2int(pcd->currentPoint_y + Dev_y_fConvF(0, dstHeight, &ctm)));
	pdConvParamAdd(param, sourceHeight);
	pdConvParamAdd(param, sourceWidth);

	if (colorDepth == 1) {
		bits = 0x31, format = 0x30,	process = 0x30;
	} else if ((colorDepth == 8) || (colorDepth == 24)) {
		bits = 0x38, format = 0x3a, process = 0x31;
	} else {
		errorno = opvpErrorNo = pdNOTSUPPORTED;
		return (pdERROR);
	}

	size = snprintf(buf, sizeof(buf), "}U%s%c%c%c0%c", param, bits, format, process, C_IS2);
	if (!pdWrite(pcd, buf, size))
		return (pdERROR);

	pcd->imageHeight = sourceHeight;
	pcd->currentImageHeight = 0;
	pcd->imageWidth = sourceWidth;
	pcd->imageDepth = colorDepth;
	pcd->drawimage_process_flag = TRUE;

	return (pdOK);
}


int lipsTransferDrawImage(
	int		printerContext,
	int		count,
	void	*imageData)
{
	printerContextData	*pcd;
	GraphicsStateData	*gs;
	unsigned char		buf[64], param[48];
	unsigned char		*lbuf, *comp_buf, *img_ptr, *trns_ptr;
	int					size, i, j;
	int					pitch, epitch, line;
	int					byte_count;
	int					buf_size, comp_size;

	dbgMessage("vdapi : %s  count = %d\n", __func__, count);
	if( count == 0 ){
		return (pdOK);
	}

	pcd = GetPrinterContext(printerContext);
	if (!pcd)
		return (pdERROR);

	gs = &pcd->gslist.gsdata;

	if (!pcd->page_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if (!pcd->drawimage_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if (pcd->imageDepth == 1)
		pitch = (pcd->imageWidth + 7)/8;
	else
		pitch = pcd->imageWidth * (pcd->imageDepth / 8);
	epitch = ((pitch + 3)/4) * 4;
	line = count / epitch;
	pcd->currentImageHeight += line;

	if (pcd->imageDepth == 8)
		byte_count = pitch * line * 3;
	else
		byte_count = pitch * line;

	dbgMessage("vdapi : %s pitch = %d, epitch = %d, line = %d\n", __func__, pitch, epitch, line);

	if (pcd->currentImageHeight >= pcd->imageHeight)
	{
		size = snprintf (buf, sizeof(buf), "}Q11%c", 0x3B);
	}
	else
	{
		size = snprintf (buf, sizeof(buf), "}Q10%c", 0x3B);
	}
	if (!pdWrite(pcd, buf, size))
		return (pdERROR);

	buf_size = (pcd->imageDepth == 8)? line*pitch*3 : line*pitch;
	lbuf = (unsigned char *)malloc(buf_size);
		if (!lbuf)
			return (pdERROR);

	trns_ptr = lbuf;
	img_ptr = imageData;
	for (i = 0; i < line; i++) {
		if (pcd->imageDepth == 8)  {
			for (j = 0; j < pitch; j++) {
				trns_ptr[j*3]     = img_ptr[j];
				trns_ptr[j*3 + 1] = img_ptr[j];
				trns_ptr[j*3 + 2] = img_ptr[j];
			}
			trns_ptr += pitch * 3;

		} else {
			memcpy(trns_ptr, img_ptr, pitch);
			trns_ptr += pitch;
		}
		img_ptr += epitch;
	}

	comp_buf = (unsigned char *)malloc(buf_size*2);
	if (!comp_buf) {
		free(lbuf);
		return (pdERROR);
	}

	comp_size = packbitsCompress(comp_buf, lbuf, buf_size);
	free(lbuf);

	pdConvParam(param, comp_size);
	size = snprintf(buf, sizeof(buf), "%s%c", param, C_IS2);
	if (!pdWrite(pcd, buf, size)) {
		free(comp_buf);
		return (pdERROR);
	}

	if (!pdWrite(pcd, comp_buf, comp_size)) {
		free(comp_buf);
		return (pdERROR);
	}

	free (comp_buf);
	return (pdOK);
}

int lipsEndDrawImage(int printerContext)
{
	printerContextData	*pcd;
	GraphicsStateData	*gs;

	dbgMessage("vdapi : %s\n", __func__);

	pcd = GetPrinterContext(printerContext);
	if (!pcd)
		return (pdERROR);

	gs = &pcd->gslist.gsdata;

	if (!pcd->page_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if (!pcd->drawimage_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	pcd->drawimage_process_flag = FALSE;
	return (pdOK);

}

