/*  Canon LIPS IV Driver for PDAPI Ver. 1.0.0
 *  Copyright CANON INC. 2004
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#include 	<sys/types.h>
#include 	<unistd.h>
#include 	<stdio.h>
#include 	<stdlib.h>
#include 	<errno.h>
#include 	<string.h>
#include 	<stdarg.h>

#include	"pddrv.h"
#include	"pdcommon.h"
#include	"pdlips.h"
#include	"cnpkliblips.h"
#include	"apiUtils.h"

extern	int	errorno;
extern	int	opvpErrorNo;

typedef struct _PaperNameTbl {
	char *pName;
	char *pOptName;
} PaperNameTbl;

PaperNameTbl  tPaperNameTbl[] = {
	{ "jpn_you_chou3_120x235mm",    "jenv_you_chou3"  },
	{ NULL, NULL                                },
};

#define	WORK_BUFFER_SIZE			256
#define	CUPSOPTION_STRING_MAXLENGTH		4096

static char cupsOptStr[ CUPSOPTION_STRING_MAXLENGTH ];

bool
make_cups_option(
	char		*jobInfo,
	char		*pcupsOpt,
	int		isize,
	printerContextData	*pcd
)
{
	bool		bRet = FALSE;
	char		*pWorkOpt = NULL;
	char		*pptr = NULL;
	int		totallen = 0;
	int		idx = 0;
	char		work_buff[ WORK_BUFFER_SIZE ];

	if( ( jobInfo == NULL ) || ( pcupsOpt == NULL ) || ( isize == 0 ) ) {
		return bRet;
	}

	memset( pcupsOpt, 0x00, isize );
	memset( work_buff, 0x00, WORK_BUFFER_SIZE );

	pWorkOpt = jobInfo;
	totallen = strlen( pWorkOpt );

	pptr = strstr( pWorkOpt, "CNOrgPageSize" );
	if( pptr != NULL ) {
		pptr += strlen( "CNOrgPageSize" );
		pptr ++;

		for ( idx = 0; tPaperNameTbl[ idx ].pOptName != NULL; idx++ ) {
			if ( strncmp( pptr, tPaperNameTbl[ idx ].pOptName, strlen( tPaperNameTbl[ idx ].pOptName ) ) == 0 ) {
				strcat( work_buff, "MediaSize" );
				strcat( work_buff, "=" );
				strcat( work_buff, tPaperNameTbl[ idx ].pName );

				pcd->org_pagesize_index = idx;
				break;
			}
		}
	}
	else if( pcd->org_pagesize_index != -1 ) {
		idx = pcd->org_pagesize_index;

		strcat( work_buff, "MediaSize" );
		strcat( work_buff, "=" );
		strcat( work_buff, tPaperNameTbl[ idx ].pName );
	}

	totallen += strlen( work_buff );

	if( isize < totallen ) {
		return bRet;
	}

	if( strlen( work_buff ) ) {
		strcat( pcupsOpt, jobInfo );
		strcat( pcupsOpt, ";" );
		strcat( pcupsOpt, work_buff );
		bRet = TRUE;
	}
	else {
		bRet = FALSE;
	}

	dbgMessage( "vdapi : pCupsOpt [size = %d] %s %s\n", strlen( pcupsOpt ), __func__, pcupsOpt );

	return bRet;
}

int lipsStartJob(int printerContext, char *jobInfo)
{
	printerContextData	*pcd;
	GraphicsStateData	*gsd;
	CnpkPageInfo		pageinfo;
	char			*pJobInfo = NULL;
	char			*originalJobInfo = NULL;
	char			*tmpJobInfo = NULL;
	int				ret = 0;

#ifdef DEBUG
#endif

	dbgMessage("vdapi : %s %s\n", __func__, jobInfo);

	pcd = GetPrinterContext(printerContext);
	if (!pcd)
		return (pdERROR);

	pcd->org_pagesize_index = -1;

	gsd = GetGraphicsState(printerContext);
	if (!gsd)
		return (pdERROR);

	if (pcd->job_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	ret = get_orignalJobInfo(jobInfo, &originalJobInfo);
	if (ret == RET_ERROR)
		return (pdERROR);

	tmpJobInfo = (ret == RET_NOHANDLE) ? jobInfo : originalJobInfo;

	if( make_cups_option( tmpJobInfo, cupsOptStr, sizeof( cupsOptStr ), pcd ) ) {
			pJobInfo = cupsOptStr;
	}
	else {
		pJobInfo = tmpJobInfo;
	}

	if (cnpkSetJobAttr(pcd->cnpk_context, pJobInfo) == -1) {
		errorno = opvpErrorNo = pdPARAMERROR;

		if (originalJobInfo != NULL)
			free(originalJobInfo);

		return (pdERROR);
	}else {
		if (originalJobInfo != NULL)
			free(originalJobInfo);
	}

	if (cnpkGetPageInfo(pcd->cnpk_context, &pageinfo) == -1) {
		errorno = opvpErrorNo = pdPARAMERROR;
		return (pdERROR);
	}

	if (cnpkStartOutput(pcd->cnpk_context) == -1) {
		;
	}

	if (!getFormParam(&pageinfo, &pcd->jobForm)) {
		errorno = opvpErrorNo = pdPARAMERROR;
		return (pdERROR);
	}

	if (!pdlips_PrtInit(pcd))
		return (pdERROR);
	if (!pdlips_PrtStart(pcd, &pcd->jobForm))
		return (pdERROR);

	memcpy(&pcd->docForm, &pcd->jobForm, sizeof(pdFormParam));
	memcpy(&pcd->pageForm, &pcd->jobForm, sizeof(pdFormParam));
	memcpy(&pcd->curForm, &pcd->jobForm, sizeof(pdFormParam));

	CleanGraphicsState(gsd);
	InitializeGraphicsState(gsd);

	pcd->job_process_flag = TRUE;
	pcd->first_page_flag = TRUE;

	return (pdOK);
}

int lipsEndJob(int printerContext)
{
	printerContextData	*pcd;

	dbgMessage("vdapi : %s\n", __func__);

	pcd = GetPrinterContext(printerContext);
	if (!pcd)
		return (pdERROR);

	if (!pcd->job_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if (pcd->path_process_flag || pcd->drawimage_process_flag || pcd->scanline_process_flag ||
					pcd->raster_process_flag || pcd->stream_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if (!pdlips_PrtEnd(pcd))
		return (pdERROR);

	if (cnpkEndOutput(pcd->cnpk_context) == -1) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	memcpy(&pcd->jobForm, &pcd->defForm, sizeof(pdFormParam));
	memcpy(&pcd->docForm, &pcd->defForm, sizeof(pdFormParam));
	memcpy(&pcd->pageForm, &pcd->defForm, sizeof(pdFormParam));

	pcd->job_process_flag = FALSE;
	pcd->doc_process_flag = FALSE;
	pcd->page_process_flag = FALSE;

	return (pdOK);
}

int lipsAbortJob(int printerContext)
{
	printerContextData	*pcd;

	dbgMessage("vdapi : %s\n", __func__);

	pcd = GetPrinterContext(printerContext);
	if (!pcd)
		return (pdERROR);

	if (!pcd->job_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if (pcd->path_process_flag || pcd->drawimage_process_flag || pcd->scanline_process_flag ||
					pcd->raster_process_flag || pcd->stream_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if (!pdlips_PrtEnd(pcd))
		return (pdERROR);

	if (cnpkEndOutput(pcd->cnpk_context) == -1) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	memcpy(&pcd->jobForm, &pcd->defForm, sizeof(pdFormParam));
	memcpy(&pcd->docForm, &pcd->defForm, sizeof(pdFormParam));
	memcpy(&pcd->pageForm, &pcd->defForm, sizeof(pdFormParam));

	pcd->job_process_flag = FALSE;
	pcd->doc_process_flag = FALSE;
	pcd->page_process_flag = FALSE;

	return (pdOK);
}

int lipsStartDoc(int printerContext, char *docInfo)
{
	printerContextData	*pcd;
	CnpkPageInfo		pageinfo;
	int			ret;
	char			*pDocInfo = NULL;

	dbgMessage("vdapi : %s %s\n", __func__, docInfo);

	pcd = GetPrinterContext(printerContext);
	if (!pcd)
		return (pdERROR);

	if (!pcd->job_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if (pcd->page_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if (pcd->doc_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if( make_cups_option( docInfo, cupsOptStr, sizeof( cupsOptStr ), pcd ) ) {
		pDocInfo = cupsOptStr;
	}
	else {
		pDocInfo = docInfo;
	}

	if (docInfo != 0) {
		if ((ret = cnpkSetDocAttr(pcd->cnpk_context, pDocInfo)) == -1) {
			errorno = opvpErrorNo = pdPARAMERROR;
			return (pdERROR);
		}

		if (cnpkGetPageInfo(pcd->cnpk_context, &pageinfo) == -1) {
			errorno = opvpErrorNo = pdPARAMERROR;
			return (pdERROR);
		}

		if (!getFormParam(&pageinfo, &pcd->docForm)) {
			errorno = opvpErrorNo = pdPARAMERROR;
			return (pdERROR);
		}
	} else
		memcpy(&pcd->docForm, &pcd->curForm, sizeof(pdFormParam));

	if (memcmp(&pcd->curForm, &pcd->docForm, sizeof(pdFormParam)) != 0) {
		if (!pdlips_PrtEnd(pcd))
			return (pdERROR);
		if (!pdlips_PrtStart(pcd, &pcd->docForm))
			return (pdERROR);

		memcpy (&pcd->curForm, &pcd->docForm, sizeof(pdFormParam));
	}

	pcd->doc_process_flag = TRUE;
	pcd->first_page_flag = TRUE;

	return (pdOK);
}

int lipsEndDoc(int printerContext)
{
		printerContextData	*pcd;

	dbgMessage("vdapi : %s\n", __func__);

	pcd = GetPrinterContext(printerContext);
	if (!pcd)
		return (pdERROR);

	if (!pcd->doc_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if (pcd->path_process_flag || pcd->drawimage_process_flag ||
				pcd->scanline_process_flag || pcd->raster_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	memcpy(&pcd->docForm, &pcd->jobForm, sizeof(pdFormParam));
	memcpy(&pcd->pageForm, &pcd->jobForm, sizeof(pdFormParam));

	pcd->doc_process_flag = FALSE;
	return (pdOK);
}

int lipsStartPage(int printerContext, char *pageInfo)
{
	printerContextData	*pcd;
	CnpkPageInfo		pageinfo;
	GraphicsStateData	*gsd;
	char			*pPageInfo = NULL;

	dbgMessage("vdapi : %s %s\n", __func__, pageInfo);

	pcd = GetPrinterContext(printerContext);
	if (!pcd)
		return (pdERROR);

	gsd = GetGraphicsState(printerContext);
	if (!gsd)
		return (pdERROR);

	if (!pcd->job_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if (pcd->page_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if (!pcd->first_page_flag) {
		if (cnpkNextPage(pcd->cnpk_context) == -1) {
			errorno = opvpErrorNo = pdBADREQUEST;
			return (pdERROR);
		}
	}

	if( make_cups_option( pageInfo, cupsOptStr, sizeof( cupsOptStr ), pcd ) ) {
		pPageInfo = cupsOptStr;
	}
	else {
		pPageInfo = pageInfo;
	}

	if (pageInfo != 0) {
		if (cnpkSetPageAttr(pcd->cnpk_context, pPageInfo) == -1) {
			errorno = opvpErrorNo = pdBADREQUEST;
			return (pdERROR);
		}

		if (cnpkGetPageInfo(pcd->cnpk_context, &pageinfo) == -1) {
			errorno = opvpErrorNo = pdBADREQUEST;
			return (pdERROR);
		}

		if (!getFormParam(&pageinfo, &pcd->pageForm)) {
			errorno = opvpErrorNo = pdBADREQUEST;
			return (pdERROR);
		}
	} else
		memcpy(&pcd->pageForm, &pcd->curForm, sizeof(pdFormParam));

	if (memcmp(&pcd->curForm, &pcd->pageForm, sizeof(pdFormParam)) != 0) {
		if (!pdlips_PrtEnd(pcd))
			return (pdERROR);
		if (!pdlips_PrtStart(pcd, &pcd->pageForm))
			return (pdERROR);

		memcpy (&pcd->curForm, &pcd->pageForm, sizeof(pdFormParam));
	}

	CleanGraphicsState(gsd);
	InitializeGraphicsState(gsd);

	pcd->page_process_flag = TRUE;
	if (pcd->first_page_flag)
		pcd->first_page_flag = FALSE;

	return (pdOK);
}

int lipsEndPage(int printerContext)
{
	printerContextData	*pcd;

	dbgMessage("vdapi : %s\n", __func__);

	pcd = GetPrinterContext(printerContext);
	if (!pcd)
		return (pdERROR);

	if (!pcd->page_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if (pcd->path_process_flag || pcd->drawimage_process_flag ||
				pcd->scanline_process_flag || pcd->raster_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	pdlips_Out_FF(pcd);

	memcpy (&pcd->pageForm, &pcd->docForm, sizeof(pdFormParam));

	pcd->page_process_flag = FALSE;
	return (pdOK);
}
