/*  Canon LIPS IV Driver for PDAPI Ver. 1.0.0
 *  Copyright CANON INC. 2004
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#include 	<sys/types.h>
#include 	<stdio.h>
#include 	<stdlib.h>
#include 	<errno.h>
#include 	<string.h>
#include	<signal.h>
#include 	<stdarg.h>
#include	<unistd.h>

#include	"pddrv.h"
#include	"pdcommon.h"
#include	"pdlips.h"
#include	"cnpkliblips.h"

extern	int (*API_Table[])();
extern	opvp_api_procs_t	api_proc_table;

int		errorno;
int		opvpErrorNo;

int	OpenPrinter(int outputFD, char *printerModel, int *nApiEntry, int (**apiEntry[])())
{
	printerContextData	*pcd;
	struct	sigaction	act;

	dbgMessage("vdapi : %s for Ver 0.2 path id. = %d, modelname = %s\n", __func__, outputFD, printerModel);

	pcd = CreatePrinterContext(outputFD);
	if (pcd == NULL)
		return (pdERROR);

	pcd->vdapi_ver = 0;

	pcd->defForm.page_size_code = paperSize_A4;
	pcd->defForm.page_xsize = 0;
	pcd->defForm.page_ysize = 0;
	pcd->defForm.page_orientation = page_Portrait;
	pcd->defForm.resolution = 600;
	pcd->defForm.color_mode = mode_Color;
	memcpy(&pcd->jobForm, &pcd->defForm, sizeof(pdFormParam));
	memcpy(&pcd->docForm, &pcd->defForm, sizeof(pdFormParam));
	memcpy(&pcd->pageForm, &pcd->defForm, sizeof(pdFormParam));

	pcd->job_process_flag        = FALSE;
	pcd->doc_process_flag       = FALSE;
	pcd->page_process_flag      = FALSE;
	pcd->path_process_flag 	   = FALSE;
	pcd->drawimage_process_flag = FALSE;
	pcd->scanline_process_flag  = FALSE;
	pcd->raster_process_flag    = FALSE;
	pcd->stream_process_flag    = FALSE;
	pcd->currentPoint_x = 0;
	pcd->currentPoint_y = 0;

	InitializeGraphicsState(&pcd->gslist.gsdata);

	*nApiEntry = API_Count;
	*apiEntry = API_Table;

	if ((pcd->cnpk_context = cnpkNew(CNPK_DOC_FORMAT_LIPS, outputFD)) == NULL) {
		errorno = pdBADREQUEST;
		return (pdERROR);
	}

#if 0
	act.sa_handler = sigtermHandler;
	sigemptyset(&act.sa_mask);
	act.sa_flags = 0;
	sigaction(SIGTERM, &act, 0);
#endif

	return (pcd->context_number);
}


int	ClosePrinter(int printerContext)
{
	printerContextData	*pcd;

	dbgMessage("vdapi : %s\n", __func__);

	pcd = GetPrinterContext(printerContext);

	cnpkDestroy (pcd->cnpk_context);

	if (DeletePrinterContext(printerContext))
		return (pdOK);
	else
		return (pdERROR);
}

int	opvpOpenPrinter(int outputFD, char *printerModel, int apiVersion[2], opvp_api_procs_t **apiEntry)
{
	printerContextData	*pcd;
	struct	sigaction	act;

#ifdef DEBUG_SLEEP
sleep(60);
#endif

	dbgMessage("vdapi : %s for Ver 1.0 path id.=%d, modelname=%s, apiVersion = %d.%d\n", __func__, outputFD, printerModel, apiVersion[0], apiVersion[1]);

	pcd = CreatePrinterContext(outputFD);
	if (pcd == NULL)
		return (pdERROR);

	pcd->vdapi_ver = 1;

	pcd->defForm.page_size_code = paperSize_A4;
	pcd->defForm.page_xsize = 0;
	pcd->defForm.page_ysize = 0;
	pcd->defForm.page_orientation = page_Portrait;
	pcd->defForm.resolution = 600;
	pcd->defForm.color_mode = mode_Color;
	memcpy(&pcd->jobForm, &pcd->defForm, sizeof(pdFormParam));
	memcpy(&pcd->docForm, &pcd->defForm, sizeof(pdFormParam));
	memcpy(&pcd->pageForm, &pcd->defForm, sizeof(pdFormParam));

	pcd->job_process_flag        = FALSE;
	pcd->doc_process_flag       = FALSE;
	pcd->page_process_flag      = FALSE;
	pcd->path_process_flag 	   = FALSE;
	pcd->drawimage_process_flag = FALSE;
	pcd->scanline_process_flag  = FALSE;
	pcd->raster_process_flag    = FALSE;
	pcd->stream_process_flag    = FALSE;
	pcd->currentPoint_x = 0;
	pcd->currentPoint_y = 0;

	InitializeGraphicsState(&pcd->gslist.gsdata);

	*apiEntry = &api_proc_table;

	if ((pcd->cnpk_context = cnpkNew(CNPK_DOC_FORMAT_LIPS, outputFD)) == NULL) {
		opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

#if 0
	act.sa_handler = sigtermHandler;
	sigemptyset(&act.sa_mask);
	act.sa_flags = 0;
	sigaction(SIGTERM, &act, 0);
#endif

	return (pcd->context_number);
}


int	opvpClosePrinter(int printerContext)
{
	printerContextData	*pcd;

	dbgMessage("vdapi : %s\n", __func__);

	pcd = GetPrinterContext(printerContext);

	cnpkDestroy (pcd->cnpk_context);

	if (DeletePrinterContext(printerContext))
		return (pdOK);
	else
		return (pdERROR);
}


