/*  Canon LIPS IV Driver for PDAPI Ver. 1.0.0
 *  Copyright CANON INC. 2004
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#include	<sys/types.h>
#include	<stdio.h>
#include	<stdlib.h>
#include	<errno.h>
#include	<string.h>
#include 	<stdarg.h>

#include	"pddrv.h"
#include	"pdlips.h"
#include	"pdcommon.h"

extern	int	errorno;
extern	int	opvpErrorNo;


int GetEmptyPathID(printerContextData *pcd)
{
	GraphicsStateDataList	*gsl;
	bool			id_table[16];
	int				i;

	for (i = 0; i < 16; i++)
		id_table[i] = FALSE;

	for (gsl = pcd->gslist.next; gsl; gsl = gsl->next) {
		if (gsl->gsdata.clipPath_registered)
			id_table[gsl->gsdata.clipPath_id] = TRUE;

		if (gsl->gsdata.drawPath_registered)
			id_table[gsl->gsdata.drawPath_id] = TRUE;
	}

	if (pcd->gslist.gsdata.clipPath_registered)
		id_table[pcd->gslist.gsdata.clipPath_id] = TRUE;

	for (i = 0; i < 16; i++) {
		if (id_table[i] == FALSE)
			return (i);
	}

	return (-1);
}


int lipsNewPath(int printerContext)
{
	printerContextData	*pcd;
	GraphicsStateData	*gs;
	unsigned char	buf[32], param[6];
	int				pid, size;

	dbgMessage("vdapi : %s\n", __func__);

	pcd = GetPrinterContext(printerContext);
	if (!pcd)
		return (pdERROR);

	gs = &pcd->gslist.gsdata;

	if (!pcd->page_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if (pcd->path_process_flag)	{
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	pid = GetEmptyPathID(pcd);

	pdConvParam(param, pid);
	size = snprintf(buf, sizeof(buf), "P(%s0%c", param, C_IS2);
	if (!pdWrite(pcd, buf, size))
		return (pdERROR);

	pcd->path_id = pid;
	pcd->path_process_flag = TRUE;

	return (pdOK);
}

int lipsEndPath(int printerContext)
{
	printerContextData	*pcd;
	GraphicsStateData	*gs;
	u_char				buf[32];
	int					size;

	dbgMessage("vdapi : %s\n", __func__);

	pcd = GetPrinterContext(printerContext);
	if (!pcd)
		return (pdERROR);

	gs = &pcd->gslist.gsdata;

	if (!pcd->page_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if (!pcd->path_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	size = snprintf(buf, sizeof(buf), "P)%c", C_IS2);
	if (!pdWrite(pcd, buf, size))
		return (pdERROR);

	gs->drawPath_registered = TRUE;
	gs->drawPath_id = pcd->path_id;

	pcd->path_process_flag = FALSE;
	return (pdOK);
}

int lipsStrokePath(int printerContext)
{
	printerContextData	*pcd;
	GraphicsStateData	*gs;
	int				pid, size;
	u_char			buf[32], param[6];

	dbgMessage("vdapi : %s\n", __func__);

	pcd = GetPrinterContext(printerContext);
	if (!pcd)
		return (pdERROR);

	gs = &pcd->gslist.gsdata;

	if (!pcd->page_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if (pcd->path_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if (gs->drawPath_registered)
		pid = gs->drawPath_id;
	else if (gs->clipPath_registered)
		pid = gs->clipPath_id;
	else {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}
	if (!pdlips_Set_AllAttr(pcd))
		return (pdERROR);

	pdConvParam(param, pid);
	size = snprintf(buf, sizeof(buf), "PS%s0%c", param, C_IS2);
	if (!pdWrite(pcd, buf, size))
		return (pdERROR);

	return (pdOK);
}

int lipsFillPath(int printerContext)
{
	printerContextData	*pcd;
	GraphicsStateData	*gs;
	int				pid, size;
	u_char			buf[32], param[6];

	dbgMessage("vdapi : %s\n", __func__);

	pcd = GetPrinterContext(printerContext);
	if (!pcd)
		return (pdERROR);

	gs = &pcd->gslist.gsdata;

	if (!pcd->page_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if (pcd->path_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if (gs->drawPath_registered)
		pid = gs->drawPath_id;
	else if (gs->clipPath_registered)
		pid = gs->clipPath_id;
	else {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if (!pdlips_Set_AllAttr(pcd))
		return (pdERROR);

	pdConvParam(param, pid);
	size = snprintf(buf, sizeof(buf), "PF%s0%c", param, C_IS2);
	if (!pdWrite(pcd, buf, size))
		return (pdERROR);

	return (pdOK);
}

int lipsStrokeFillPath(int printerContext)
{
	printerContextData	*pcd;
	GraphicsStateData	*gs;
	int				pid, size;
	u_char			buf[32], param[6];

	dbgMessage("vdapi : %s\n", __func__);

	pcd = GetPrinterContext(printerContext);
	if (!pcd)
		return (pdERROR);

	gs = &pcd->gslist.gsdata;

	if (!pcd->page_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if (pcd->path_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if (gs->drawPath_registered)
		pid = gs->drawPath_id;
	else if (gs->clipPath_registered)
		pid = gs->clipPath_id;
	else {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if (!pdlips_Set_AllAttr(pcd))
		return (pdERROR);

	pdConvParam(param, pid);
	size = snprintf(buf, sizeof(buf), "P&%s0%c", param, C_IS2);
	if (!pdWrite(pcd, buf, size))
		return (pdERROR);

	return (pdOK);
}

int lipsSetClipPath(
	int 		printerContext,
	pdClipRule cliprule)
{
	printerContextData	*pcd;
	GraphicsStateData	*gs;
	int					pid;
	pdFillMode			fillmode;

	dbgMessage("vdapi : %s\n", __func__);

	pcd = GetPrinterContext(printerContext);
	if (!pcd)
		return (pdERROR);

	gs = &pcd->gslist.gsdata;

	if (!pcd->page_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if (pcd->path_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if (gs->drawPath_registered)
		pid = gs->drawPath_id;
	else if (gs->clipPath_registered)
		pid = gs->clipPath_id;
	else {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	fillmode = (cliprule == pdclipRuleEvenOdd)? pdFillModeEvenOdd : pdFillModeWinding;
	if (!pdlips_Set_FillMode(pcd, fillmode))
		return (pdERROR);

	if (!pdlips_Set_Clip(pcd, pid))
		return (pdERROR);

	gs->clipPath_registered = TRUE;
	gs->clipPath_id = pid;
	gs->drawPath_registered = FALSE;

	return (pdOK);
}

int lipsResetClipPath(int printerContext)
{
	printerContextData	*pcd;
	GraphicsStateData	*gs;

	dbgMessage("vdapi : %s\n", __func__);

	pcd = GetPrinterContext(printerContext);
	if (!pcd)
		return (pdERROR);

	gs = &pcd->gslist.gsdata;

	if (!pcd->page_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if (!gs->clipPath_registered) {
#if 1
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
#else
		return (pdOK);
#endif
	}

	if (!pdlips_Reset_Clip(pcd, gs->clipPath_id))
		return (pdERROR);

	gs->clipPath_registered = FALSE;

	return (pdOK);
}

int lipsSetCurrentPoint(
	int 	printerContext,
	pdFix 	x,
	pdFix 	y)
{
	printerContextData	*pcd;
	GraphicsStateData	*gs;
	int					size;
	u_char				buf[32], param[16];

	dbgMessage("vdapi : %s  x = %d, y = %d\n", __func__, x, y);

	pcd = GetPrinterContext(printerContext);
	if (!pcd)
		return (pdERROR);

	gs = &pcd->gslist.gsdata;

	if (!pcd->page_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if (pcd->path_process_flag)	{
		pdConvParam(param, Dev_x_fConv(x, y, &gs->ctm));
		pdConvParamAdd(param, Dev_y_fConv(x, y, &gs->ctm));
		size = snprintf(buf, sizeof(buf), "p10%s%c", param, C_IS2);
		if (!pdWrite(pcd, buf, size))
			return (pdERROR);
	}

	pcd->currentPoint_x = x;
	pcd->currentPoint_y = y;
	return (pdOK);
}

int lipsLinePath(
	int 	printerContext,
	int 	flag,
	int 	npoints,
	pdPoint *points)
{
	printerContextData	*pcd;
	GraphicsStateData	*gs;
	u_char				*buf;
	u_char				*param;
	int					i, size;
	int					bufsize;

	dbgMessage("vdapi : %s\n", __func__);

	pcd = GetPrinterContext(printerContext);
	if (!pcd)
		return (pdERROR);

	gs = &pcd->gslist.gsdata;

	if (!pcd->page_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if (!pcd->path_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	param = malloc(npoints * 12);
	if (!param)
		return (pdERROR);
	bufsize = npoints * 12 +16;
	buf = malloc(bufsize);
	if (!buf) {
		free (param);
		return (pdERROR);
	}

	param[0]='\0';
	for (i = 0; i < npoints; i++) {
		pdConvParamAdd(param, Dev_x_fConv(points[i].x, points[i].y, &gs->ctm));
		pdConvParamAdd(param, Dev_y_fConv(points[i].x, points[i].y, &gs->ctm));
	}

	size = snprintf(buf, bufsize, "p402%s%c", param, C_IS2);

	dbgMessage("vdapi : %s size = %d, buf = %s\n", __func__, size, buf);

	if (!pdWrite(pcd, buf, size)) {
		free (param);
		free (buf);
		return (pdERROR);
	}

	if (flag == pdPathClose) {
		size = snprintf(buf, bufsize, "p0%c", C_IS2);
		if (!pdWrite(pcd, buf, size)) {
			free (param);
			free (buf);
			return (pdERROR);
		}
	}

	if (flag == pdPathOpen) {
		pcd->currentPoint_x = points[npoints - 1].x;
		pcd->currentPoint_y = points[npoints - 1].y;
	}

	free (param);
	free (buf);

	return (pdOK);
}

int SinglePolygonPath(
	int		printerContext,
	int		num,
	pdPoint	*points)
{
	dbgMessage("common : %s\n", __func__);

	if (lipsSetCurrentPoint(printerContext, points[0].x, points[0].y) == pdERROR)
		return (pdERROR);

	if (lipsLinePath(printerContext, pdPathClose, num-1, &points[1]) == pdERROR)
		return (pdERROR);

	return (pdOK);
}

int lipsPolygonPath(
	int 	printerContext,
	int		npolygons,
	int		*nvertexes,
	pdPoint	*points)
{
	printerContextData	*pcd;
	GraphicsStateData	*gs;
	int					i;

	dbgMessage("vdapi : %s\n", __func__);

	pcd = GetPrinterContext(printerContext);
	if (!pcd)
		return (pdERROR);

	gs = &pcd->gslist.gsdata;

	if (!pcd->page_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if (!pcd->path_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	for (i=0; i < npolygons; i++) {
		if (SinglePolygonPath(printerContext, nvertexes[i], points) == pdERROR)
			return (pdERROR);
		points += nvertexes[i];
	}

	return (pdOK);
}

int SingleRectanglePath(
	int		printerContext,
	pdRectangle	*rectangle)
{
	pdPoint	points[3];

	dbgMessage("common : %s\n", __func__);

	if (lipsSetCurrentPoint(printerContext, rectangle->p0.x, rectangle->p0.y) == -1)
		return (pdERROR);

	points[0].x = rectangle->p1.x, points[0].y = rectangle->p0.y;
	points[1].x = rectangle->p1.x, points[1].y = rectangle->p1.y;
	points[2].x = rectangle->p0.x, points[2].y = rectangle->p1.y;

	if (lipsLinePath(printerContext, pdPathClose, 3, points) == -1)
		return (pdERROR);

	return (pdOK);
}


int lipsRectanglePath(
	int			printerContext,
	int			nrectangles,
	pdRectangle	*rectangles)
{
	printerContextData	*pcd;
	GraphicsStateData	*gs;
	int					i;

	dbgMessage("vdapi : %s\n", __func__);

	pcd = GetPrinterContext(printerContext);
	if (!pcd)
		return (pdERROR);

	gs = &pcd->gslist.gsdata;

	if (!pcd->page_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if (!pcd->path_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	for (i=0; i < nrectangles; i++) {
		if (SingleRectanglePath(printerContext, &rectangles[i]) == pdERROR)
			return (pdERROR);
	}

	return (pdOK);
}

int lipsBezierPath(
	int		printerContext,
	int		npoints,
	pdPoint	*points)
{
	printerContextData	*pcd;
	GraphicsStateData	*gs;
	int					i, size;
	int					bufsize;
	u_char				*buf, *param;

	dbgMessage("vdapi : %s\n", __func__);

	pcd = GetPrinterContext(printerContext);
	if (!pcd)
		return (pdERROR);

	gs = &pcd->gslist.gsdata;

	if (!pcd->page_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if (!pcd->path_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if ((npoints % 3) != 0)	{
		errorno = opvpErrorNo = pdPARAMERROR;
		return (pdERROR);
	}

	param = malloc (npoints * 12);
	if (!param)
		return (pdERROR);
	bufsize = npoints * 12 + 16;
	buf = malloc (bufsize);
	if (!buf) {
		free (param);
		return (pdERROR);
	}

	param[0] = '\0';
	for (i=0; i < npoints; i++) {
		pdConvParamAdd(param, Dev_x_fConv(points[i].x, points[i].y, &gs->ctm));
		pdConvParamAdd(param, Dev_y_fConv(points[i].x, points[i].y, &gs->ctm));
	}
	size = snprintf(buf, bufsize, "p404%s%c", param, C_IS2);

	dbgMessage("vdapi : %s size = %d, buf = %s\n", __func__, size, buf);

	if (!pdWrite(pcd, buf, size)) {
		free (param);
		free (buf);
		return (pdERROR);
	}

	pcd->currentPoint_x = points[npoints - 1].x;
	pcd->currentPoint_y = points[npoints - 1].y;

	free (param);
	free (buf);

	return (pdOK);
}
