/*  Canon LIPS IV Driver for PDAPI Ver. 1.0.0
 *  Copyright CANON INC. 2004
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include	<sys/types.h>

#include	"pddrv.h"
#include	"pdlips.h"


int	(*API_Table[])() =
{
	OpenPrinter,
	ClosePrinter,

	lipsStartJob,
	lipsEndJob,
	lipsStartDoc,
	lipsEndDoc,
	lipsStartPage,
	lipsEndPage,
	NOP,
	NOP,

	lipsResetCTM,
	lipsSetCTM,
	lipsGetCTM,
	lipsInitGS,
	lipsSaveGS,
	lipsRestoreGS,
	lipsQueryColorSpace,
	lipsSetColorSpace,
	lipsGetColorSpace,
	lipsQueryROP,
	lipsSetROP,
	lipsGetROP,
	lipsSetFillMode,
	lipsGetFillMode,
	NOP,
	NOP,
	lipsSetLineWidth,
	lipsGetLineWidth,
	lipsSetLineDash,
	lipsGetLineDash,
	lipsSetLineDashOffset,
	lipsGetLineDashOffset,
	lipsSetLineStyle,
	lipsGetLineStyle,
	lipsSetLineCap,
	lipsGetLineCap,
	lipsSetLineJoin,
	lipsGetLineJoin,
	lipsSetMiterLimit,
	lipsGetMiterLimit,
	NOP,
	NOP,
	lipsSetStrokeColor,
	lipsSetFillColor,
	NOP,

	lipsNewPath,
	lipsEndPath,
	lipsStrokePath,
	lipsFillPath,
	lipsStrokeFillPath,
	lipsSetClipPath,
	lipsResetClipPath,
	lipsSetCurrentPoint,
	lipsLinePath,
	lipsPolygonPath,
	lipsRectanglePath,
	NOP,
	lipsBezierPath,
	NOP,

	lipsDrawBitmapText,

	lipsDrawImage,
	lipsStartDrawImage,
	lipsTransferDrawImage,
	lipsEndDrawImage,

	NOP,
	NOP,
	NOP,

	NOP,
	NOP,
	NOP,
	NOP,

	NOP,
	NOP,
	NOP
};

opvp_api_procs_t api_proc_table = {
	opvpOpenPrinter,
	opvpClosePrinter,

	lipsStartJob,
	lipsEndJob,
	lipsAbortJob,
	lipsStartDoc,
	lipsEndDoc,
	lipsStartPage,
	lipsEndPage,
	NOP,
	NOP,

	lipsResetCTM,
	lipsSetCTM,
	lipsGetCTM,
	lipsInitGS,
	lipsSaveGS,
	lipsRestoreGS,
	lipsQueryColorSpace1,
	lipsSetColorSpace,
	lipsGetColorSpace,
	lipsSetFillMode,
	lipsGetFillMode,
	NOP,
	NOP,
	lipsSetLineWidth,
	lipsGetLineWidth,
	lipsSetLineDash1,
	lipsGetLineDash1,
	lipsSetLineDashOffset,
	lipsGetLineDashOffset,
	lipsSetLineStyle,
	lipsGetLineStyle,
	lipsSetLineCap,
	lipsGetLineCap,
	lipsSetLineJoin,
	lipsGetLineJoin,
	lipsSetMiterLimit,
	lipsGetMiterLimit,
	lipsSetPaintMode,
	lipsGetPaintMode,
	lipsSetStrokeColor,
	lipsSetFillColor,
	NOP,

	lipsNewPath,
	lipsEndPath,
	lipsStrokePath,
	lipsFillPath,
	lipsStrokeFillPath,
	lipsSetClipPath,
	lipsResetClipPath,
	lipsSetCurrentPoint,
	lipsLinePath,
	lipsPolygonPath,
	lipsRectanglePath,
	NOP,
	lipsBezierPath,
	NOP,

	lipsDrawImage1,
	lipsStartDrawImage1,
	lipsTransferDrawImage,
	lipsEndDrawImage,

	NOP,
	NOP,
	NOP,

	NOP,
	NOP,
	NOP,
	NOP,

	NOP,
	NOP,
	NOP
};

paperSize	paperSizeConvTable[] = {
				{ "Letter",		paperSize_Letter },
				{ "Ledger",		paperSize_Ledger },
				{ "Legal",		paperSize_Legal },
				{ "Executive",	paperSize_Executive },
				{ "A5",			paperSize_A5 },
				{ "B5",			paperSize_B5 },
				{ "A4",			paperSize_A4 },
				{ "B4",			paperSize_B4 },
				{ "A3",			paperSize_A3 },
				{ "jenv_you4",	paperSize_Envelope_you4 },
				{ "jenv_kaku2",	paperSize_Envelope_kaku2 },
				{ "Postcard",	paperSize_PS },
				{ "11x17",		paperSize_Ledger },
				{ "4x_postcard",paperSize_4xPS },
				{ "dbl_postcard",paperSize_2xPS },
				{ "jenv_chou3",	paperSize_Envelope_chou3 },
				{ "SRA3",		paperSize_SRA3 },
				{ "12x18",		paperSize_12x18 },
				{ "13x19",      paperSize_13x19 },
				{ "jenv_you2",	paperSize_Envelope_you2 },
    			{ "jenv_you_chou3", paperSize_Envelope_you_chou3 },
    			{ "A6",		paperSize_A6 },
				{ "jenv_chou4",		paperSize_Envelope_chou4 },
				{ "jenv_chou40",	paperSize_Envelope_chou40 },
				{ "B6",			paperSize_B6 },
				{ "",          0 }
			};


#define PAPER_SIZE_NUM		25
#define PAPER_SIZE_PARAM	2

int		pageFormat[PAPER_SIZE_NUM][PAPER_SIZE_PARAM] = {
					{ 80, 81 },
					{ 12, 13 },
					{ 14, 15 },
					{ 16, 17 },
					{ 18, 19 },
					{ 24, 25 },
					{ 26, 27 },
					{ 28, 29 },
					{ 30, 31 },
					{ 32, 33 },
					{ 34, 35 },
					{ 40, 41 },
					{ 50, 51 },
					{ 54, 55 },
					{ 22, 23 },
					{ 20, 21 },
					{ 58, 59 },
					{ 42, 43 },
					{ 44, 45 },
					{ 46, 47 },
					{ 52, 53 },
					{ 60, 61 },
					{ 48, 49 },
					{ 62, 63 },
					{ 64, 65 }
		};

double		marginSize[PAPER_SIZE_NUM] = {
					5.0,
					5.0,
					5.0,
					5.0,
					5.0,
					5.0,
					5.0,
					5.0,
					5.0,
					5.0,
					5.0,
					5.0,
					10.0,
					10.0,
					5.0,
					5.0,
					10.0,
					10.0,
					10.0,
					12.6,
					10.0,
					10.0,
					5.0,
					10.0,
					10.0
		};

mediaType	mediaTypeConvTable[] = {
				{ "PlainPaper",		media_PlainPaper },
				{ "PlainPaperL",	media_PlainPaperL },
				{ "HEAVY",			media_HEAVY },
				{ "ThickPaperH",	media_HEAVY2 },
				{ "OHP",			media_OHP },
				{ "GLOSSY",			media_GLOSSY },
				{ "HEAVYL",		media_HEAVYL },
				{ "COATED",		media_GLOSSY },
				{ "", 				0 }
		};

bindType	bindEdgeConvTable[] = {
				{ "Left",	bindedge_Left },
				{ "Right",	bindedge_Right },
				{ "Top",	bindedge_Top },
				{ "Bottom",	bindedge_Bottom },
				{ "", 		0 }
		};
