/*  Canon LIPS IV Driver for PDAPI Ver. 1.0.0
 *  Copyright CANON INC. 2004
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#include 	<sys/types.h>
#include 	<unistd.h>
#include 	<stdio.h>
#include 	<stdlib.h>
#include 	<errno.h>
#include 	<string.h>
#include 	<stdarg.h>

#include	"pddrv.h"
#include	"pdlips.h"
#include	"pdcommon.h"

extern	int	errorno;
extern	int	opvpErrorNo;


int lipsDrawBitmapText(
	int 	printerContext,
	int		width,
	int		height,
	int		pitch,
	void	*fontdata)
{
	printerContextData	*pcd;
	GraphicsStateData	*gs;
	u_char				buf[64], param[48];
	int					size;
	int					imgpitch, srcpitch, i;
	u_char				*ptr;

	dbgMessage("vdapi : %s width = %d height = %d\n", __func__,width,height);
	if( (width == 0) || (height == 0) ){
		return(pdOK);
	}

	pcd = GetPrinterContext(printerContext);
	if (!pcd)
		return (pdERROR);

	gs = &pcd->gslist.gsdata;

	if (!pcd->page_process_flag) {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (pdERROR);
	}

	if (!pdlips_Set_ROP(pcd))
		return (pdERROR);

	pdConvParam(param, pdConvColorValue(gs->fillcolor.color[2]));
	pdConvParamAdd(param, pdConvColorValue(gs->fillcolor.color[1]));
	pdConvParamAdd(param, pdConvColorValue(gs->fillcolor.color[0]));
	size = snprintf(buf, sizeof(buf), "}T%s%c", param, C_IS2);
	if (!pdWrite(pcd, buf, size))
		return (pdERROR);

	pdConvParam(param, Dev_x_fConv(pcd->currentPoint_x, pcd->currentPoint_y - int2fix(height), &gs->ctm));
	pdConvParamAdd(param, Dev_y_fConv(pcd->currentPoint_x, pcd->currentPoint_y - int2fix(height), &gs->ctm));
	pdConvParamAdd(param, Dev_x_fConv(pcd->currentPoint_x + int2fix(width), pcd->currentPoint_y, &gs->ctm));
	pdConvParamAdd(param, Dev_y_fConv(pcd->currentPoint_x + int2fix(width), pcd->currentPoint_y, &gs->ctm));
	pdConvParamAdd(param, Dev_x_fConv(pcd->currentPoint_x, pcd->currentPoint_y, &gs->ctm));
	pdConvParamAdd(param, Dev_y_fConv(pcd->currentPoint_x, pcd->currentPoint_y, &gs->ctm));
	pdConvParamAdd(param, height);
	pdConvParamAdd(param, width);
	size = snprintf(buf, sizeof(buf), "}U%s1001%c", param, C_IS2);
	if (!pdWrite(pcd, buf, size))
		return (pdERROR);

	imgpitch = (width % 8)? width/8 + 1: width/8;
	pdConvParam(param, imgpitch * height);
	size = snprintf (buf, sizeof(buf), "}Q110%s%c", param, C_IS2);
	if (!pdWrite(pcd, buf, size))
		return (pdERROR);

	ptr = fontdata;
	srcpitch = pitch / 8;
	for (i = 0; i < height; i++) {
		if (!pdWrite(pcd, ptr, imgpitch))
			return (pdERROR);
		ptr += srcpitch;
	}

	return (pdOK);
}
