/*  Canon LIPS IV Driver for PDAPI Ver. 1.0.0
 *  Copyright CANON INC. 2004
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#include	<sys/types.h>
#include	<unistd.h>
#include	<stdio.h>
#include	<stdlib.h>
#include	<errno.h>
#include	<string.h>
#include 	<stdarg.h>

#include	"pddrv.h"
#include	"pdcommon.h"
#include	"cnpkliblips.h"

extern	int	errorno;
extern	int	opvpErrorNo;

extern	paperSize	paperSizeConvTable[];
extern	mediaType	mediaTypeConvTable[];
extern	bindType	bindEdgeConvTable[];

printerContextDataList	*g_printerlist;


void sigtermHandler(int sig)
{
	printerContextDataList	**pcl;

	for (pcl = &g_printerlist; *pcl; pcl = &((*pcl)->next)) {
		if (((*pcl)->pcdata.cnpk_context))
			cnpkDestroy((*pcl)->pcdata.cnpk_context);
	}

	exit(0);
}


void dbgMessage(const char *fmt, ...)
{
#ifdef	DEBUG
	va_list	vap;

	va_start(vap, fmt);
	vfprintf(stderr, fmt, vap);
	va_end(vap);
#endif
}


pdFix int2fix(int i)
{
	return (i << FIX_FRAC_WIDTH);
}

int fix2int(pdFix fix)
{
	return ((fix>>FIX_FRAC_WIDTH) + ((fix & FIX_FRAC_RMASK)? 1:0));
}

pdFix float2fix(float f)
{
	int	floor, frac;

	floor = (int)f;
	frac = (f - floor)*FIX_FRAC_DENOM;
	return ((floor << FIX_FRAC_WIDTH) | (frac & 0x000000ff));
}

float fix2float(pdFix fix)
{
	float	f;

	f = (double)(fix >> FIX_FRAC_WIDTH) + (double)(fix & 0x000000ff)/FIX_FRAC_DENOM;
	return (f);
}

int double2int(double d)
{
	bool	neg;
	int		i;

	neg = FALSE;

	if (d < 0) {
		neg = TRUE;
		d = -d;
	}

	i = (int)(d + 0.5);
	if (neg)
		i = -i;

	return (i);
}


void InitializeCTM(pdCTM *ctm)
{
	ctm->a = 1, ctm->b = 0;
	ctm->c = 0, ctm->d = 1;
	ctm->e = 0, ctm->f = 0;
}

int Dev_x_dConv(double dx, double dy, pdCTM *ctm)
{
	return (double2int((double)ctm->a * dx + (double)ctm->c * dy + (double)ctm->e));
}

int Dev_y_dConv(double dx, double dy, pdCTM *ctm)
{
	return (double2int((double)ctm->b * dx + (double)ctm->d * dy + (double)ctm->f));
}
int Dev_x_fConv(pdFix fx, pdFix fy, pdCTM *ctm)
{
	return (Dev_x_dConv(fix2float(fx), fix2float(fy), ctm));
}

int Dev_y_fConv(pdFix fx, pdFix fy, pdCTM *ctm)
{
	return (Dev_y_dConv(fix2float(fx), fix2float(fy), ctm));
}

pdFix Dev_x_dConvF(double dx, double dy, pdCTM *ctm)
{
	return (float2fix((double)ctm->a * dx + (double)ctm->c * dy + (double)ctm->e));
}

pdFix Dev_y_dConvF(double dx, double dy, pdCTM *ctm)
{
	return (float2fix((double)ctm->b * dx + (double)ctm->d * dy + (double)ctm->f));
}
pdFix Dev_x_fConvF(pdFix fx, pdFix fy, pdCTM *ctm)
{
	return (Dev_x_dConvF(fix2float(fx), fix2float(fy), ctm));
}

pdFix Dev_y_fConvF(pdFix fx, pdFix fy, pdCTM *ctm)
{
	return (Dev_y_dConvF(fix2float(fx), fix2float(fy), ctm));
}

printerContextData *CreatePrinterContext(int fd)
{
	int	pct_num;
	printerContextDataList	**pcl, *pcd;

	pct_num = 0;
	for (pcl = &g_printerlist; *pcl; pcl = &((*pcl)->next))
	{
		if ((*pcl)->pcdata.fd == fd) {
			errorno = opvpErrorNo = pdBADREQUEST;
			return (NULL);
		}

		if ((*pcl)->pcdata.context_number > pct_num)
			pct_num = (*pcl)->pcdata.context_number;
	}

	pct_num++;

	pcd = (printerContextDataList *)malloc(sizeof(printerContextDataList));
	if (pcd == NULL) {
		errorno = opvpErrorNo = pdFATALERROR;
		return (NULL);
	}

	memset(pcd, 0, sizeof(printerContextDataList));
	pcd->pcdata.context_number = pct_num;
	pcd->pcdata.fd = fd;
	*pcl = pcd;

	return (&pcd->pcdata);
}

bool DeletePrinterContext(int printerContext)
{
	printerContextDataList	**pcl, *pcd;
	printerContextData		*prd;
	GraphicsStateDataList	*gsl, *gst;
	GraphicsStateData		*gsd;

	for (pcl = &g_printerlist; *pcl; pcl = &((*pcl)->next)) {
		if ((*pcl)->pcdata.context_number == printerContext)
			break;
	}

	if (*pcl == NULL) {
		errorno = opvpErrorNo = pdBADCONTEXT;
		return (FALSE);
	}

	pcd = *pcl;
	prd = &pcd->pcdata;
	*pcl = pcd->next;

#if 0
	for (gsl = prd->gslist.next; gsl; gsl = gsl->next) {
		gsd = &gsl->gsdata;
		CleanGraphicsState(gsd);
		free(gsl);
	}
#else
	for (gsl = prd->gslist.next; gsl;) {
		gst = gsl;
		gsl = gst->next;
		gsd = &gst->gsdata;
		CleanGraphicsState(gsd);
		free(gst);
	}
#endif

	gsd = &prd->gslist.gsdata;
	CleanGraphicsState(gsd);

	free(pcd);

	return (TRUE);
}

printerContextData *GetPrinterContext(int printerContext)
{
	printerContextDataList	*pcl;

	for (pcl = g_printerlist; pcl; pcl = pcl->next) {
		if (pcl->pcdata.context_number == printerContext)
			break;
	}

	if (pcl == NULL) {
		errorno = opvpErrorNo = pdBADCONTEXT;
		return (NULL);
	}

	return (&pcl->pcdata);
}

GraphicsStateData *GetGraphicsState(int printerContext)
{
	printerContextData	*pcd;

	pcd = (printerContextData *)GetPrinterContext(printerContext);
	if (pcd == NULL) {
		errorno = opvpErrorNo = pdBADCONTEXT;
		return (NULL);
	}

	return (&pcd->gslist.gsdata);
}

void CleanGraphicsState(GraphicsStateData *gsdata)
{
	DeleteBrushData(gsdata->strokecolor.pbrush);
	DeleteBrushData(gsdata->fillcolor.pbrush);
	DeleteBrushData(gsdata->bgcolor.pbrush);

	memset (gsdata, 0, sizeof(GraphicsStateData));
}

void InitializeGraphicsState(GraphicsStateData *gsdata)
{
	InitializeCTM(&gsdata->ctm);
	gsdata->colorspace 	= pdcspaceStandardRGB;
	gsdata->rop 		= SRCPAINT;
	gsdata->fillmode 	= pdFillModeEvenOdd;
	gsdata->linewidth 	= int2fix(1);
	gsdata->dashnum 	= 0;
	gsdata->dashoffset 	= int2fix(0);
	gsdata->linestyle 	= pdlineStyleSolid;
	gsdata->linecap 	= pdlineCapButt;
	gsdata->linejoin 	= pdlineJoinMiter;
	gsdata->miterlimit 	= float2fix(10.0);
	SetSolidBrush(&gsdata->strokecolor, 0, 0, 0);
	SetSolidBrush(&gsdata->fillcolor, 0, 0, 0);

	gsdata->clipPath_registered = FALSE;
	gsdata->drawPath_registered = FALSE;
}


pdBrushData *DuplicateBrushData(pdBrushData *src)
{
	pdBrushData	*dst;
	int			size;

	size = sizeof(pdBrushData) + src->height*(src->pitch/8) - 1;
	dst = (pdBrushData *)malloc(size);
	if (!dst) {
		errorno = opvpErrorNo = pdFATALERROR;
		return (NULL);
	}

	memcpy(dst, src, size);
	return (dst);
}

void DeleteBrushData(pdBrushData *ptr)
{
	if (ptr != NULL)
		free (ptr);
}

bool CopyBrush(pdBrush *dst, pdBrush *src)
{
	pdBrushData	*pbd;

	if (dst->pbrush != NULL)
		DeleteBrushData(dst->pbrush);

	memcpy (dst, src, sizeof(pdBrush));

	if (src->pbrush != NULL) {
		pbd = DuplicateBrushData(src->pbrush);
		if (pbd == NULL) {
			errorno = opvpErrorNo = pdFATALERROR;
			return (FALSE);
		}

		dst->pbrush = pbd;
	}

	return (TRUE);
}

void SetSolidBrush(pdBrush *pb, int R, int G, int B)
{
	pb->colorSpace = pdcspaceStandardRGB;
	pb->color[0] = B;
	pb->color[1] = G;
	pb->color[2] = R;
	pb->color[3] = 0;
	pb->xorg = 0;
	pb->yorg = 0;
	pb->pbrush = (pdBrushData *)NULL;
}

bool CopyGraphicsState(GraphicsStateDataList *dst, GraphicsStateDataList *src)
{
	pdBrushData		*pstrokebrush;
	pdBrushData		*pfillbrush;
	pdBrushData		*pbgbrush;

	CleanGraphicsState(&dst->gsdata);
	memcpy(&dst->gsdata, &src->gsdata, sizeof(GraphicsStateData));

	if (src->gsdata.strokecolor.pbrush != NULL) {
		pstrokebrush = (pdBrushData *)DuplicateBrushData(src->gsdata.strokecolor.pbrush);
		if (pstrokebrush == NULL)
			return (FALSE);
	} else
		pstrokebrush = NULL;
	dst->gsdata.strokecolor.pbrush = pstrokebrush;

	if (src->gsdata.fillcolor.pbrush != NULL) {
		pfillbrush = (pdBrushData *)DuplicateBrushData(src->gsdata.fillcolor.pbrush);
		if (pfillbrush == NULL)
			return (FALSE);
	} else
		pfillbrush = NULL;
	dst->gsdata.fillcolor.pbrush = pfillbrush;

	if (src->gsdata.bgcolor.pbrush != NULL) {
		pbgbrush = (pdBrushData *)DuplicateBrushData(src->gsdata.bgcolor.pbrush);
		if (pbgbrush == NULL)
			return (FALSE);
	} else
		pbgbrush = NULL;
	dst->gsdata.bgcolor.pbrush = pbgbrush;

	return (TRUE);
}


bool get_page_size(char *paper, int *val)
{
	int		i;

	i = 0;
	while (strlen(paperSizeConvTable[i].paper_name)) {
		if (!strcmp(paperSizeConvTable[i].paper_name, paper)) {
			*val = paperSizeConvTable[i].paperSizeCode;
			return (TRUE);
		}
		i++;
	}

	errorno = opvpErrorNo = pdPARAMERROR;
	return (FALSE);
}

bool get_page_orientation(char *orient, int *val)
{
	if (!strcmp(orient, "portrait"))
		*val = page_Portrait;
	else if (!strcmp(orient, "landscape"))
		*val = page_Landscape;
	else {
		errorno = opvpErrorNo = pdPARAMERROR;
		return (FALSE);
	}

	return (TRUE);
}

bool get_media_type(char *media, int *val)
{
	int		i;

	if (!strlen(media))	{
		*val = media_None;
		return (TRUE);
	}

	i = 0;
	while (strlen(mediaTypeConvTable[i].media_name)) {
		if (!strcmp(mediaTypeConvTable[i].media_name, media)) {
			*val = mediaTypeConvTable[i].mediaTypeCode;
			return (TRUE);
		}
		i++;
	}


	*val = media_Panel;
	return (TRUE);

}

bool get_resolution(char *reso, int *val)
{
	int		resolution;

	if (sscanf(reso, "%d", &resolution) == 1)
		*val = resolution;
	else {
		errorno = opvpErrorNo = pdPARAMERROR;
		return (FALSE);
	}

	return (TRUE);
}

bool get_color_mode(char *color, int *val)
{
	if (!strcmp(color, "color"))
		*val = mode_Color;
	else if (!strcmp(color, "gray"))
		*val = mode_Mono;
	else
		return (FALSE);

	return (TRUE);
}

bool get_tonersave_mode(char *tonersave, int *val)
{
	if (!strlen(tonersave))	{
		*val = tonersave_None;
		return (TRUE);
	}

	if (!strcmp(tonersave, "Auto"))
		*val = tonersave_Auto;
	else if (!strcmp(tonersave, "True"))
		*val = tonersave_On;
	else if (!strcmp(tonersave, "False"))
		*val = tonersave_Off;
	else
		return (FALSE);

	return (TRUE);
}

bool get_bindedge_type(char *edge, int *val)
{
	int		i;

	if (!strlen(edge))	{
		*val = bindedge_None;
		return (TRUE);
	}

	i = 0;
	while (strlen(bindEdgeConvTable[i].bindedge_name)) {
		if (!strcmp(bindEdgeConvTable[i].bindedge_name, edge)) {
			*val = bindEdgeConvTable[i].bindedgeCode;
			return (TRUE);
		}
		i++;
	}

	errorno = opvpErrorNo = pdPARAMERROR;
	return (FALSE);
}

bool get_bindedge_shift(char *shift, int *val)
{
	int		edgeshift;

	if (sscanf(shift, "%d", &edgeshift) == 1)
		*val = edgeshift;
	else {
		errorno = opvpErrorNo = pdPARAMERROR;
		return (FALSE);
	}

	return (TRUE);
}

bool get_duplex_mode(char *duplex, int *val)
{
	if (!strcmp(duplex, "True"))
		*val = mode_Duplex;
	else if (!strcmp(duplex, "False"))
		*val = mode_Simplex;
	else
		return (FALSE);

	return (TRUE);
}

bool getFormParam(CnpkPageInfo *pageinfo, pdFormParam *formparam)
{
	int	value;

	if (!get_page_size (pageinfo->page_name, &value))
		return (FALSE);
	else {
		formparam->page_size_code = value;
		formparam->page_xsize = 0;
		formparam->page_ysize = 0;
	}

	if (!get_page_orientation(pageinfo->page_orient, &value))
		return (FALSE);
	else
		formparam->page_orientation = value;

	if (!get_media_type(pageinfo->media_type, &value))
		return (FALSE);
	else
		formparam->media_type = value;

	if (!get_resolution(pageinfo->resolution, &value))
		return (FALSE);
	else
		formparam->resolution = value;

	if (!get_color_mode(pageinfo->color_mode, &value))
		return (FALSE);
	else
		formparam->color_mode = value;

	if (!get_tonersave_mode(pageinfo->toner_save, &value))
		return (FALSE);
	else
		formparam->toner_save = value;

	if (!get_bindedge_type(pageinfo->bind_edge, &value))
		return (FALSE);
	else
		formparam->bind_edge = value;

	if (formparam->bind_edge != bindedge_None) {
		if (!get_bindedge_shift(pageinfo->bind_edge_shift, &value))
			return (FALSE);
		else
			formparam->bind_edge_shift = value;
	}

	if (!get_duplex_mode(pageinfo->duplex, &value))
		return (FALSE);
	else
		formparam->duplex_mode = value;

	return (TRUE);
}
