/*  Canon LIPS IV Driver for PDAPI Ver. 1.0.0
 *  Copyright CANON INC. 2004
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#include	<sys/types.h>
#include	<unistd.h>
#include	<stdio.h>
#include	<stdlib.h>
#include	<errno.h>
#include	<string.h>
#include 	<stdarg.h>

#include	"pddrv.h"
#include	"pdlips.h"
#include	"pdcommon.h"
#include	"cnpkliblips.h"

extern	int	errorno;
extern	int	opvpErrorNo;

extern	int	pageFormat[25][2];
extern	double marginSize[25];


bool pdWrite(printerContextData *pcd, char *buf, int datasize)
{
	if ((cnpkSendData(pcd->cnpk_context, buf, datasize) == 0))
		return (TRUE);
	else {
		errorno = opvpErrorNo = pdBADREQUEST;
		return (FALSE);
	}
}

bool pdWriteStr(printerContextData *pcd, char *buf)
{

	return (pdWrite(pcd, buf, strlen(buf)));
}



void pdConvParam(char *str, int data)
{

	u_char		sign;
	u_char		buf[8];
	int			i, j;
	u_int		value;


	sign = (data < 0)? 0x00:0x10;
	value = abs(data);

	buf[0] = (u_char)(value & 0x0000000f) | sign | 0x20;

	value = value >> 4;
	i = 1;
	while(value != 0) {
		buf[i] = (u_char)(value & 0x0000003f) | 0x40;
		value = value >> 6;
		i++;
	}

	for (j=0; j<i; j++)
		str[j] = buf[i-j-1];

	str[i] = '\0';
}

void pdConvParamAdd(char *str, int data)
{
	char	*ptr;


	ptr = str + strlen(str);
	pdConvParam(ptr, data);
}



int     searchContData (u_char *src, int start, int current, int total)
{
    if ((current == (total - 1))||((current - start) >= 127))
        return (current);

    if (src[current] == src[current + 1]) {
        current++;
        return (searchContData (src, start, current, total));
    } else
        return (current);
}

int     searchDescData (u_char *src, int start, int current, int total)
{
    if ((current == (total - 1))||((current - start) >= 127))
        return (current);

    if (src[current] != src[current + 1]) {
        current++;
        return (searchDescData (src, start, current, total));
    } else
        return (current - 1);
}

int  packbitsCompress(u_char *dst, u_char *src, int count)
{
    int             i;
    register u_char *p;
    int             cont_indx;
    int             desc_indx;
    int             n;

    i = 0;
    n = 0;

    while (1) {
        if (i >= count)
            break;

        cont_indx = searchContData (src, i, i, count);
        if (cont_indx > i) {
            *dst++ = (u_char)(i - cont_indx);
            *dst++ = src[i];
            i = cont_indx + 1;
            n += 2;
        } else {
            desc_indx = searchDescData (src, i, i, count);
            *dst++ = (u_char)(desc_indx - i);
            p = src + i;
            memcpy (dst, p, (desc_indx - i + 1));
            dst = dst + (desc_indx - i + 1);
            n += desc_indx - i + 2;
            i = desc_indx + 1;
        }
    }

    return (n);
}


bool pdlips_PrtInit(printerContextData *pcd)
{
	u_char	buf[10];
	int		count;


	count = snprintf(buf, sizeof(buf), "%c%%@", C_ESC);
	return (pdWrite(pcd, buf, count));
}

bool pdlips_PrtStart(printerContextData *pcd, pdFormParam *formdata)
{
	u_char	buf[256], param1[16], param2[16];
	int		size, count;
	u_char	*ptr;
	int		offset;
	char	colormode;


	ptr = buf;
	count = 0;

	size = snprintf(ptr, sizeof(buf), "%cP41;%d;1J%c\\", C_ESC, formdata->resolution, C_ESC);
	ptr += size;
	count += size;

	size = snprintf(ptr, sizeof(buf)-count, "%c[%d\"p", C_ESC, formdata->color_mode);
	ptr += size;
	count += size;

	if (formdata->toner_save != tonersave_None) {
		size = snprintf(ptr, sizeof(buf)-count, "%c[0;%d;0'v", C_ESC, formdata->toner_save);
		ptr += size;
		count += size;
	}

	size = snprintf(ptr, sizeof(buf)-count, "%c<", C_ESC);
	ptr += size;
	count += size;

	if (formdata->media_type != media_None) {
		size = snprintf(ptr, sizeof(buf)-count, "%c[%d't", C_ESC, formdata->media_type);
		ptr += size;
		count += size;
	}

	size = snprintf(ptr, sizeof(buf)-count, "%c[%dp", C_ESC, pageFormat[formdata->page_size_code][formdata->page_orientation]);
	ptr += size;
	count += size;

#if 0
	size = snprintf(ptr, sizeof(buf)-count, "%c[1v", C_ESC);
	ptr += size;
	count += size;
#endif

	if (formdata->duplex_mode == mode_Duplex) {
		size = snprintf(ptr, sizeof(buf)-count, "%c[2;0#x", C_ESC);
		ptr += size;
		count += size;
	} else {
		size = snprintf(ptr, sizeof(buf)-count, "%c[0#x", C_ESC);
		ptr += size;
		count += size;
	}

	if (formdata->bind_edge != bindedge_None) {
		offset = (int)((double)formdata->resolution * (double)formdata->bind_edge_shift / MM_PER_INCH);
		size = snprintf(ptr, sizeof(buf)-count, "%c[%d;%d#w", C_ESC, formdata->bind_edge, offset);
		ptr += size;
		count += size;
	}

	size = snprintf(ptr, sizeof(buf)-count, "%c[0&}", C_ESC);
	ptr += size;
	count += size;

	if (formdata->page_orientation == page_Landscape) {
		size = snprintf(ptr, sizeof(buf)-count, "%c@1%c", C_SP, C_IS2);
	}
	else {
		size = snprintf(ptr, sizeof(buf)-count, "%c@0%c", C_SP, C_IS2);
	}
	ptr += size;
	count += size;

	size = snprintf(ptr, sizeof(buf)-count, "#%c", C_IS2);
	ptr += size;
	count += size;

	pdConvParam(param1, formdata->resolution);
	pdConvParam(param2, 1);
	size =snprintf(ptr, sizeof(buf)-count, "!0%s%s%c", param1, param2, C_IS2);
	ptr += size;
	count += size;

	colormode = (formdata->color_mode == 0)? 3:1;
	pdConvParam(param1, colormode);
	size = snprintf(ptr, sizeof(buf)-count, "!1%s%c", param1, C_IS2);
	ptr += size;
	count += size;

	size = snprintf(ptr, sizeof(buf)-count, "$%c", C_IS2);
	ptr += size;
	count += size;

	offset = double2int((double)formdata->resolution * marginSize[formdata->page_size_code] / MM_PER_INCH);
	pdConvParam(param1, -offset);
	pdConvParamAdd(param1, -offset);
	size = snprintf(ptr, sizeof(buf)-count, "}\"%s%c", param1, C_IS2);
	count += size;

	return (pdWrite(pcd, buf, count));
}

bool pdlips_PrtEnd(printerContextData *pcd)
{
	u_char	buf[64];
	u_char	*ptr;
	int		count, size;


	ptr = buf;
	count = 0;

	size = snprintf(ptr, sizeof(buf)-count, "%%%c", C_IS2);
	ptr += size;
	count += size;

	size = snprintf(ptr, sizeof(buf)-count, "}p%c", C_IS2);
	ptr += size;
	count += size;

	size = snprintf(ptr, sizeof(buf)-count, "%cP0J%c\\", C_ESC, C_ESC);
	count += size;

	return (pdWrite(pcd, buf, count));
}


bool pdlips_Out_FF(printerContextData *pcd)
{
	u_char	buf[64];
	int		size;


	size = snprintf(buf, sizeof(buf), "%c", C_FF);
	return (pdWrite(pcd, buf, size));
}


#if 0
bool pdlips_Set_ROP(printerContextData *pcd)
{
	GraphicsStateData	*gsd;
	u_char				rop_mode;
	u_char				buf[10];
	int					size;


	gsd = &pcd->gslist.gsdata;

	if (gsd->rop == SRCCOPY)
		rop_mode = 0x30;
	else if (gsd->rop == SRCPAINT)
		rop_mode = 0x31;
	else
		rop_mode = 0x30;

	size = snprintf(buf, sizeof(buf), "}H%c%c", rop_mode, C_IS2);
	return (pdWrite(pcd, buf, size));
}

#else
bool pdlips_Set_ROP(printerContextData *pcd)
{
	GraphicsStateData	*gsd;
	u_char				rop_mode;
	u_char				buf[10];
	int					size;


	gsd = &pcd->gslist.gsdata;

	if (pcd->vdapi_ver == 0) {
		if (gsd->rop == SRCCOPY)
			rop_mode = 0x30;
		else if (gsd->rop == SRCPAINT)
			rop_mode = 0x31;
		else
			rop_mode = 0x30;
	} else {
		if (gsd->paintmode == pdpaintModeOpaque )
			rop_mode = 0x30;
		else if (gsd->paintmode == pdpaintModeTransparent)
			rop_mode = 0x31;
		else
			rop_mode = 0x30;
	}

	size = snprintf(buf, sizeof(buf), "}H%c%c", rop_mode, C_IS2);
	return (pdWrite(pcd, buf, size));
}

#endif

bool pdlips_Set_FillMode(printerContextData *pcd, pdFillMode fillmode)
{
	u_char				fill_mode;
	u_char				buf[10];
	int					size;


	if (fillmode == pdFillModeEvenOdd)
		fill_mode = 0x30;
	else if (fillmode == pdFillModeWinding)
		fill_mode = 0x31;
	else {
		errorno = opvpErrorNo = pdPARAMERROR;
		return (FALSE);
	}

	size = snprintf(buf, sizeof(buf), "}I%c%c", fill_mode, C_IS2);
	return (pdWrite(pcd, buf, size));
}

bool pdlips_Set_LineWidth(printerContextData *pcd)
{
	GraphicsStateData	*gsd;
	u_char				buf[16], param[6];
	int					width, size;


	gsd = &pcd->gslist.gsdata;

	width = fix2int(gsd->linewidth);
	if (width <= 0)
		width = 1;

	pdConvParam(param, width);
	size = snprintf(buf, sizeof(buf), "F1%s%c", param, C_IS2);
	return (pdWrite(pcd, buf, size));
}

bool pdlips_Set_LineStyle(printerContextData *pcd)
{
	GraphicsStateData	*gsd;
	u_char				buf[128], param1[6], param2[80];
	int					size, count, i;
	u_char		*ptr;


	gsd = &pcd->gslist.gsdata;

	if (gsd->linestyle == pdlineStyleSolid) {
		size = snprintf(buf, sizeof(buf), "E1%c%c", 0x30, C_IS2);
		return (pdWrite(pcd, buf, size));

	} else if (gsd->linestyle == pdlineStyleDash) {
		pdConvParam(param1, fix2int(gsd->dashoffset));

		count = 0;
		ptr = buf;
		param2[0]='\0';
		for (i=0; i<gsd->dashnum; i++)
			pdConvParamAdd(param2, fix2int(gsd->pDash[i]));
		size = snprintf(ptr, sizeof(buf), "}d%c%c%s%s%c", 0x2c, 0x31, param1, param2, C_IS2);
		ptr += size;
		count += size;

		size = snprintf(ptr, sizeof(buf)-count, "E1%c%c", 0x2c, C_IS2);
		count += size;
		return (pdWrite(pcd, buf, count));

	} else {
		errorno = opvpErrorNo = pdPARAMERROR;
		return (FALSE);
	}
}

bool pdlips_Set_LineCap(printerContextData *pcd)
{
	GraphicsStateData	*gsd;
	u_char				linecap, buf[16];
	int					size;


	gsd = &pcd->gslist.gsdata;

	if (gsd->linecap == pdlineCapButt)
		linecap = 0x30;
	else if (gsd->linecap == pdlineCapRound)
		linecap = 0x31;
	else if (gsd->linecap == pdlineCapSquare)
		linecap = 0x32;
	else {
		errorno = opvpErrorNo = pdPARAMERROR;
		return (FALSE);
	}
	size = snprintf(buf, sizeof(buf), "}E%c%c", linecap, C_IS2);
	return (pdWrite(pcd, buf, size));
}

bool pdlips_Set_LineJoin(printerContextData *pcd)
{
	GraphicsStateData	*gsd;
	u_char				linejoin, buf[16];
	int			size;


	gsd = &pcd->gslist.gsdata;

	if (gsd->linejoin == pdlineJoinMiter)
		linejoin = 0x32;
	else if (gsd->linejoin == pdlineJoinRound)
		linejoin = 0x31;
	else if (gsd->linejoin == pdlineJoinBevel)
		linejoin = 0x33;
	else {
		errorno = opvpErrorNo = pdPARAMERROR;
		return (FALSE);
	}

	size = snprintf(buf, sizeof(buf), "}F%c%c", linejoin, C_IS2);
	return (pdWrite(pcd, buf, size));
}

bool pdlips_Set_MiterLimit(printerContextData *pcd)
{
	GraphicsStateData	*gsd;
	int					miterlimit, size;
	u_char				buf[16], param[6];


	gsd = &pcd->gslist.gsdata;

	if (fix2int(gsd->miterlimit) != 0)
		miterlimit = double2int((16383.0 * 2.0) / fix2float(gsd->miterlimit));
	else
		miterlimit = double2int(16383.0 * 2.0);

	pdConvParam(param, miterlimit);

	size = snprintf(buf, sizeof(buf), "}M%s%c", param, C_IS2);
	return (pdWrite(pcd, buf, size));
}

bool pdlips_Set_PaintMode(printerContextData *pcd)
{
	errorno = opvpErrorNo = pdNOTSUPPORTED;
	return (TRUE);
}

u_int pdConvColorValue(u_char color)
{

	return ((u_int)color * 1000 / 255);
}

u_int pdConvGrayValue(u_char R, u_char G, u_char B)
{
	u_int	Y, val;

	Y = double2int(0.2990 * (double)R + 0.5870 * (double)G + 0.1140 * (double)B);
	val = ((1000 - pdConvColorValue(Y)) < 0)? 0 : (1000 - pdConvColorValue(Y));
	return (val);
}


bool pdRegisterBrush(printerContextData *pcd, pdBrush *pb, int regnum)
{
	u_char		buf[64], param[6];
	int			size;
	pdBrushData	*pbd;


	pbd = pb->pbrush;
	if ((pbd->width != 32) || (pbd->height != 32)) {
		errorno = opvpErrorNo = pdNOTSUPPORTED;
		return (FALSE);
	}

	pdConvParam(param, regnum);
	size = snprintf(buf, sizeof(buf), "}O1%s%c", param, C_IS2);
	if (!pdWrite(pcd, buf, size))
		return (FALSE);

	return (pdWrite(pcd, pbd->data, pbd->pitch * pbd->height / 8));
}

bool pdlips_Set_StrokeColor(printerContextData *pcd)
{
	GraphicsStateData	*gsd;
	u_char				buf[64], param[32];
	int					size;


	gsd = &pcd->gslist.gsdata;

	if (gsd->strokecolor.pbrush == NULL) {
		size = snprintf(buf, sizeof(buf), "}G%c%c", 0x30, C_IS2);
		if (!pdWrite(pcd, buf, size))
			return (FALSE);
	} else {
		if (!pdRegisterBrush(pcd, &gsd->strokecolor, -10))
			return (FALSE);

		size = snprintf(buf, sizeof(buf), "}G%c%c%c", 0x33, 0x2a, C_IS2);
		if (!pdWrite(pcd, buf, size))
			return (FALSE);
	}

	if (pcd->pageForm.color_mode) {
		pdConvParam(param, pdConvColorValue(gsd->strokecolor.color[2]));
		pdConvParamAdd(param, pdConvColorValue(gsd->strokecolor.color[1]));
		pdConvParamAdd(param, pdConvColorValue(gsd->strokecolor.color[0]));
	} else {
		pdConvParam(param, pdConvGrayValue(gsd->strokecolor.color[2],
						gsd->strokecolor.color[1], gsd->strokecolor.color[0]));
	}
	size = snprintf(buf, sizeof(buf), "G%s%c", param, C_IS2);
	return (pdWrite(pcd, buf, size));
}


bool pdlips_Set_FillColor(printerContextData *pcd)
{
	GraphicsStateData	*gsd;
	u_char				buf[64], param[32];
	int					size;


	gsd = &pcd->gslist.gsdata;

	if (gsd->fillcolor.pbrush == NULL) {
		size = snprintf(buf, sizeof(buf), "I1%c", C_IS2);
		if (!pdWrite(pcd, buf, size))
			return (FALSE);
	} else {
		if (!pdRegisterBrush(pcd, &gsd->fillcolor, -11))
			return (FALSE);

		size = snprintf(buf, sizeof(buf), "I%c%c", 0x2b, C_IS2);
		if (!pdWrite(pcd, buf, size))
			return (FALSE);
	}

	if (pcd->pageForm.color_mode) {
		pdConvParam(param, pdConvColorValue(gsd->fillcolor.color[2]));
		pdConvParamAdd(param, pdConvColorValue(gsd->fillcolor.color[1]));
		pdConvParamAdd(param, pdConvColorValue(gsd->fillcolor.color[0]));
	} else {
		pdConvParam(param, pdConvGrayValue(gsd->fillcolor.color[2],
						gsd->fillcolor.color[1], gsd->fillcolor.color[0]));
	}
	size = snprintf(buf, sizeof(buf), "J%s%c", param, C_IS2);
	return (pdWrite(pcd, buf, size));
}


bool pdlips_Set_Clip(printerContextData *pcd, int pid)
{
	int		size;
	u_char	buf[32], param[6];

	pdConvParam(param, pid);
	size = snprintf (buf, sizeof(buf), "PC%s0%c", param, C_IS2);
	if (!pdWrite(pcd, buf, size))
		return (FALSE);

	size = snprintf (buf, sizeof(buf), "*0%c", C_IS2);
	if (!pdWrite(pcd, buf, size))
		return (FALSE);

	return (TRUE);
}


bool pdlips_Reset_Clip(printerContextData *pcd, int pid)
{
	int		size;
	u_char	buf[32];

	size = snprintf (buf, sizeof(buf), "*1%c", C_IS2);
	if (!pdWrite(pcd, buf, size))
		return (FALSE);

	return (TRUE);
}


bool pdlips_Set_AllAttr(printerContextData *pcd)
{
	GraphicsStateData	*gsd;


	gsd = &pcd->gslist.gsdata;

	if (!pdlips_Set_ROP(pcd))
		return (FALSE);
	if (!pdlips_Set_FillMode(pcd, gsd->fillmode))
		return (FALSE);
	if (!pdlips_Set_LineWidth(pcd))
		return (FALSE);
	if (!pdlips_Set_LineStyle(pcd))
		return (FALSE);
	if (!pdlips_Set_LineCap(pcd))
		return (FALSE);
	if (!pdlips_Set_LineJoin(pcd))
		return (FALSE);
	if (!pdlips_Set_MiterLimit(pcd))
		return (FALSE);
	if (!pdlips_Set_StrokeColor(pcd))
		return (FALSE);
	if (!pdlips_Set_FillColor(pcd))
		return (FALSE);

	return (TRUE);
}

