/*  Canon LIPS IV Driver for PDAPI Ver. 1.0.0
 *  Copyright CANON INC. 2004
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#ifndef	_PDDRV_H_
#define	_PDDRV_H_

#include	<sys/types.h>
#include	"cnpkliblips.h"

#define	pdOK			0
#define	pdERROR			-1

#define	pdFATALERROR	-1
#define	pdBADREQUEST	-2
#define	pdBADCONTEXT	-3
#define	pdNOTSUPPORTED	-4
#define	pdJOBCANCELED	-5
#define	pdPARAMERROR	-6



typedef	int		bool;

#ifndef TRUE
#define	TRUE	1
#endif

#ifndef FALSE
#define	FALSE	0
#endif

#ifndef NULL
#define	NULL	0
#endif

#ifndef NOP
#define NOP	(void *)0
#endif

#define	MM_PER_INCH		25.4

#if 0
#define	SRCCOPY		0x00cc0020
#define	SRCPAINT	0x00ee0086
#else
#define	SRCCOPY		0xcc
#define	SRCPAINT	0xb8
#endif

#if 1

#define	FIX_FRAC_WIDTH	8
#define	FIX_INT_WIDTH	24
#define	FIX_FRAC_DENOM	(1<<FIX_FRAC_WIDTH)
#define	FIX_FRAC_RMASK	(1<<(FIX_FRAC_WIDTH-1))
typedef	int		pdFix;

#else

#define	FIX_FRACT_WIDTH		8
#define	FIX_FRACT_MASK		0x000000FF
#define	FIX_FRACT_DENOM		(1<<FIX_FRACT_WIDTH)
#define	FIX_FLOOR_WIDTH		(sizeof(int)*8-FIX_FRACT_WIDTH)
#define	FIX_FRACT_RMASK		(1<<(FIX_FRACT_WIDTH-1))

typedef	int	pdFix;

#define	i2Fix(i)		(i << FIX_FRACT_WIDTH)
#define	Fix2f(fix)		((double)(fix >> FIX_FRACT_WIDTH)\
							 + (double)(fix & FIX_FRACT_MASK)/FIX_FRACT_DENOM)
#define	f2Fix(f)		((int)floor(f) << FIX_FRACT_WIDTH |\
							(int)((f - floor(f)) * FIX_FRACT_DENOM))

#define	Fix2i(fix)	((fix >> FIX_FRACT_WIDTH) + ((fix & FIX_FRACT_RMASK)? 1:0))
#define	Fix2d(fix)	((double)(fix >> FIX_FRACT_WIDTH) +\
							(double)(fix & FIX_FRACT_MASK)/FIX_FRACT_DENOM)
#endif


typedef	struct _pdPoint {
	pdFix	x;
	pdFix	y;
} pdPoint;

typedef struct _pdRectangle {
	pdPoint	p0;
	pdPoint	p1;
} pdRectangle;

typedef struct _pdRoundRectangle {
	pdPoint	p0;
	pdPoint	p1;
	pdFix	xellipse;
	pdFix	yellipse;
} pdRoundRectangle;

#if 0
typedef enum _pdImageFormat {
	pdiformatRaw = 0,
	pdiformatRLE = 1,
	pdiformatJPEG = 2,
	pdiformatPNG = 3
} pdImageFormat;
#else
typedef enum _pdImageFormat {
	pdiformatRaw = 0,
	pdiformatMask = 1,
	pdiformatRLE = 2,
	pdiformatJPEG = 3,
	pdiformatPNG = 4
} pdImageFormat;
#endif

typedef enum _pdColorMapping {
	pdcmapDirect = 0,
	pdcmapIndexed = 1
} pdColorMapping;

#if 0
typedef enum _pdColorSpace {
	pdcspaceBW = 0,
	pdcspaceDeviceGray = 1,
	pdcspaceDeviceCMY=2,
	pdcspaceDeviceCMYK = 3,
	pdcspaceDeviceRGB = 4,
	pdcspaceStandardRGB = 5,
	pdcspaceStandardRGB64 = 6
} pdColorSpace;
#else
typedef enum _pdColorSpace {
	pdcspaceBW 				= 0,
	pdcspaceDeviceGray 		= 1,
	pdcspaceDeviceCMY		= 2,
	pdcspaceDeviceCMYK 		= 3,
	pdcspaceDeviceRGB 		= 4,
	pdcspaceDeviceKRGB		= 5,
	pdcspaceStandardRGB 	= 6,
	pdcspaceStandardRGB64 	= 7
} pdColorSpace;
#endif

typedef enum _pdFillMode {
	pdFillModeEvenOdd = 0,
	pdFillModeWinding = 1
} pdFillMode;

typedef enum _pdPaintMode {
	pdpaintModeOpaque = 0,
	pdpaintModeTransparent
} pdPaintMode;

typedef enum _pdClipRule {
	pdclipRuleEvenOdd = 0,
	pdclipRuleWinding = 1
} pdClipRule;

typedef enum _pdLineStyle {
	pdlineStyleSolid = 0,
	pdlineStyleDash = 1
} pdLineStyle;

typedef enum _pdLineCap {
	pdlineCapButt = 0,
	pdlineCapRound = 1,
	pdlineCapSquare = 2
} pdLineCap;

typedef enum _pdLineJoin {
	pdlineJoinMiter = 0,
	pdlineJoinRound = 1,
	pdlineJoinBevel = 2
} pdLineJoin;

typedef enum _pdBrushDataType {
	pdbdtypeNormal = 0
} pdBrushDataType;

#if 0
typedef struct _pdBrushData {
	pdBrushDataType	type;
	int				width;
	int				height;
	int				pitch;
	char			data[1];
} pdBrushData;

typedef struct _pdBrush {
	pdColorSpace	colorSpace;
	int				color[4];
#if 0
	pdBrushData		*pbrush;
	int				xorg;
	int				yorg;
#else
	int				xorg;
	int				yorg;
	pdBrushData		*pbrush;
#endif
} pdBrush;
#else
typedef struct _pdBrushData {
	pdBrushDataType	type;
	int				width;
	int				height;
	int				pitch;
#if defined(__GNUC__) && __GNUC__ <= 2
	char			data[1];
#else
	char			 data[];
#endif
} pdBrushData;

typedef struct _pdBrush {
	pdColorSpace	colorSpace;
	int				color[4];
	int				xorg;
	int				yorg;
	pdBrushData		*pbrush;
} pdBrush;
#endif

#if 0
#define		pdArc		0
#define		pdChord		1
#define		pdPie		2
#define		pdClockwise
#define		pdCounterclockwise
#define		pdPathClose	0
#define		pdPathOpen	1
#else
typedef enum _pdArcMode {
	pdArc			= 0,
	pdChord			= 1,
	pdPie			= 2
} pdArcMode;

typedef enum _pdArcDir {
	pdClockwise			= 0,
	pdCounterclockwise	= 1
} pdArcDir;

typedef enum _pdPathMode {
	pdPathClose			= 0,
	pdPathOpen			= 1
} pdPathMode;
#endif

typedef	struct _pdCTM {
	float	a, b, c, d, e, f;
} pdCTM;

typedef	struct _opvp_api_procs {
	int (*opvpOpenPrinter)(int, char*, int*, int*, struct _opvp_api_procs**);
	int (*opvpClosePrinter)(int);
	int (*opvpStartJob)(int, char*);
	int (*opvpEndJob)(int);
	int (*opvpAbortJob)(int);
	int (*opvpStartDoc)(int, char*);
	int (*opvpEndDoc)(int);
	int (*opvpStartPage)(int, char*);
	int (*opvpEndPage)(int);
	int (*opvpQueryDeviceCapability)(int, unsigned int, int, unsigned char*);
	int (*opvpQueryDeviceInfo)(int, unsigned int, int, char*);
	int (*opvpResetCTM)(int);
	int (*opvpSetCTM)(int, pdCTM*);
	int (*opvpGetCTM)(int, pdCTM*);
	int (*opvpInitGS)(int);
	int (*opvpSaveGS)(int);
	int (*opvpRestoreGS)(int);
	int (*opvpQueryColorSpace)(int,pdColorSpace*,int*);
	int (*opvpSetColorSpace)(int,pdColorSpace);
	int (*opvpGetColorSpace)(int,pdColorSpace*);
#if 0
	int (*opvpQueryROP)(int,int*,int*);
	int (*opvpSetROP)(int,int);
	int (*opvpGetROP)(int,int*);
#endif
	int (*opvpSetFillMode)(int,pdFillMode);
	int (*opvpGetFillMode)(int,pdFillMode*);
	int (*opvpSetAlphaConstant)(int,float);
	int (*opvpGetAlphaConstant)(int,float*);
	int (*opvpSetLineWidth)(int,pdFix);
	int (*opvpGetLineWidth)(int,pdFix*);
	int (*opvpSetLineDash)(int,pdFix*,int);
	int (*opvpGetLineDash)(int,pdFix*,int*);
	int (*opvpSetLineDashOffset)(int,pdFix);
	int (*opvpGetLineDashOffset)(int,pdFix*);
	int (*opvpSetLineStyle)(int,pdLineStyle);
	int (*opvpGetLineStyle)(int,pdLineStyle*);
	int (*opvpSetLineCap)(int,pdLineCap);
	int (*opvpGetLineCap)(int,pdLineCap*);
	int (*opvpSetLineJoin)(int,pdLineJoin);
	int (*opvpGetLineJoin)(int,pdLineJoin*);
	int (*opvpSetMiterLimit)(int,pdFix);
	int (*opvpGetMiterLimit)(int,pdFix*);
	int (*opvpSetPaintMode)(int,pdPaintMode);
	int (*opvpGetPaintMode)(int,pdPaintMode*);
	int (*opvpSetStrokeColor)(int,pdBrush*);
	int (*opvpSetFillColor)(int,pdBrush*);
	int (*opvpSetBgColor)(int,pdBrush*);
	int (*opvpNewPath)(int);
	int (*opvpEndPath)(int);
	int (*opvpStrokePath)(int);
	int (*opvpFillPath)(int);
	int (*opvpStrokeFillPath)(int);
	int (*opvpSetClipPath)(int,pdClipRule);
	int (*opvpResetClipPath)(int);
	int (*opvpSetCurrentPoint)(int,pdFix,pdFix);
	int (*opvpLinePath)(int,pdPathMode,int,pdPoint*);
	int (*opvpPolygonPath)(int,int,int*,pdPoint*);
	int (*opvpRectanglePath)(int,int,pdRectangle*);
	int (*opvpRoundRectanglePath)(int,int,pdRoundRectangle*);
	int (*opvpBezierPath)(int,int,pdPoint*);
	int (*opvpArcPath)(int,pdArcMode,pdArcDir,pdFix,pdFix,pdFix,pdFix,pdFix,pdFix,pdFix,pdFix);
	int (*opvpDrawImage)(int,int,int,int,int,pdImageFormat,pdRectangle,void*);
	int (*opvpStartDrawImage)(int,int,int,int,int,pdImageFormat,pdRectangle);
	int (*opvpTransferDrawImage)(int,int,void*);
	int (*opvpEndDrawImage)(int);
	int (*opvpStartScanline)(int,int);
	int (*opvpScanline)(int,int,int*);
	int (*opvpEndScanline)(int);
	int (*opvpStartRaster)(int,int);
	int (*opvpTransferRasterData)(int,int,unsigned char*);
	int (*opvpSkipRaster)(int,int);
	int (*opvpEndRaster)(int);
	int (*opvpStartStream)(int);
	int (*opvpTransferStreamData)(int,int,void*);
	int (*opvpEndStream)(int);
} opvp_api_procs_t;

int	OpenPrinter (int outputFD, char *printerModel, int *nApiEntry, int (**apiEntry[])());
int	ClosePrinter (int printerContext);

int	opvpOpenPrinter (int outputFD, char *printerModel, int apiVersion[2], opvp_api_procs_t **apiEntry );
int	opvpClosePrinter (int printerContext);



#define	LINE_DASH_MAX	16

typedef struct GraphicsStateData_t {
	pdCTM			ctm;
	pdColorSpace	colorspace;
	int				rop;
	pdFillMode		fillmode;
	pdFix			linewidth;

	pdLineStyle		linestyle;
	int				dashnum;
	pdFix			pDash[LINE_DASH_MAX];
	pdFix			dashoffset;

	pdLineCap		linecap;
	pdLineJoin		linejoin;
	pdFix			miterlimit;
	pdPaintMode		paintmode;

	pdBrush			strokecolor;
	pdBrush			fillcolor;
	pdBrush			bgcolor;

	bool			clipPath_registered;
	int				clipPath_id;
	bool			drawPath_registered;
	int				drawPath_id;
} GraphicsStateData;

typedef struct GraphicsStateDataList_t {
	struct GraphicsStateDataList_t	*next;
	GraphicsStateData				gsdata;
} GraphicsStateDataList;



typedef struct pdFormParam_t {
	int		page_size_code;
	int		page_xsize;
	int		page_ysize;
	int		page_orientation;
	int		media_type;
	int		resolution;
	int		color_mode;
	int		toner_save;
	int		bind_edge;
	int		bind_edge_shift;
	int		duplex_mode;
} pdFormParam;

typedef enum _pdPaperSize {
	paperSize_User,
	paperSize_A3,
	paperSize_A4,
	paperSize_A5,
	paperSize_PS,
	paperSize_B4,
	paperSize_B5,
	paperSize_B6,
	paperSize_Letter,
	paperSize_Legal,
	paperSize_Ledger,
	paperSize_Executive,
	paperSize_Envelope_you4,
	paperSize_Envelope_kaku2,
	paperSize_4xPS,
	paperSize_2xPS,
	paperSize_Envelope_chou3,
	paperSize_SRA3,
	paperSize_12x18,
	paperSize_13x19,
	paperSize_Envelope_you2,
	paperSize_Envelope_you_chou3,

	paperSize_A6,
	paperSize_Envelope_chou4,
	paperSize_Envelope_chou40
} pdPaperSize;

typedef enum _pdPageOrientation {
	page_Portrait,
	page_Landscape
} pdPageOrientation;

typedef enum _pdMediaType {
	media_None			= -1,
	media_Panel			= 0,
	media_ThinPaper		= 10,
	media_PlainPaperL	= 15,
	media_PlainPaper	= 20,
	media_HEAVYL		= 27,
	media_HEAVY			= 30,
	media_HEAVY2		= 35,
	media_OHP			= 40,
	media_GLOSSY		= 41
} pdMediaType;

typedef enum _pdColorMode {
	mode_Mono,
	mode_Color
} pdColorMode;

typedef enum _pdTonesaveMode {
	tonersave_None		= -1,
	tonersave_Auto		= 0,
	tonersave_Off		= 1,
	tonersave_On		= 2
} pdTonersaveMode;

typedef enum _pdBindEdge {
	bindedge_None		= -1,
	bindedge_Left		= 0,
	bindedge_Right		= 1,
	bindedge_Top		= 2,
	bindedge_Bottom		= 3
} pdBindEdge;

typedef enum _pdDuplexMode {
	mode_Simplex,
	mode_Duplex
} pdDuplexMode;


typedef struct printerCntextData_t {
	int			context_number;
	int			fd;
	CnpkCtx		*cnpk_context;
	int			vdapi_ver;

	pdFormParam	curForm;
	pdFormParam	defForm;
	pdFormParam	jobForm;
	pdFormParam	docForm;
	pdFormParam	pageForm;

	bool		job_process_flag;
	bool		doc_process_flag;
	bool		page_process_flag;

	bool		path_process_flag;
	int			path_id;

	bool		drawimage_process_flag;
	int			imageWidth;
	int			imageDepth;
	int		imageHeight;
	int		currentImageHeight;

	bool		scanline_process_flag;
	bool		raster_process_flag;
	bool		stream_process_flag;

	bool		first_page_flag;

	int		org_pagesize_index;

	pdFix		currentPoint_x;
	pdFix		currentPoint_y;

	GraphicsStateDataList	gslist;

} printerContextData;

typedef struct printerContextDataList_t {
	struct printerContextDataList_t	*next;
	printerContextData				pcdata;
} printerContextDataList;


typedef struct {
	char		*paper_name;
	pdPaperSize	paperSizeCode;
} paperSize;

typedef struct {
	char		*media_name;
	pdMediaType	mediaTypeCode;
} mediaType;

typedef struct {
	char		*bindedge_name;
	pdBindEdge	bindedgeCode;
} bindType;

#endif
