/*
 *  CUPS add-on module for Canon LIPS IV printer.
 *  Copyright CANON INC. 2004
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <cups/cups.h>
#include <cups/ppd.h>
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <ctype.h>
#include <errno.h>
#include <stdbool.h>
#include <sys/stat.h>
#include <string.h>
#include <strings.h>
#include <zlib.h>

#include "paramlist.h"
#include "buflist.h"
#include "functions.h"



#define	INDENT_NONE	0
#define	INDENT_RIGHT	1
#define	INDENT_LEFT	2
#define JOBINFO_STR_MAX 2046

#if HAVE_CONFIG_H
#include <config.h>
#endif

#define	EXEC_PATH		PROG_PATH
#define	EXEC_GS			"gs"

#define	TABLE_BUF_SIZE	1024
#define	DATA_BUF_SIZE	(1024 * 256)
#define	PARAM_BUF_SIZE	256

#define IS_BLANK(c)	(c == ' '  || c == '\t')
#define IS_RETURN(c)	(c == '\r' || c == '\n')

#define ACRO_UTIL "%%BeginResource: procset Adobe_CoolType_Utility_MAKEOCF"
#define ACRO_UTIL_LEN 55

#define	BANNER_DESC "%%BeginResource procset bannerprint"
#define	BANNER_DESC_LEN 35

#define	DEF_NO_ERR	0
#define	DEF_ERR		-1
#define	error_proc(src,line,err)	{\
	if(err == DEF_NO_ERR){err = DEF_ERR;}}
#define CN_JOB_INFO_HEADER  "CNCompJobInfo="
#define PS_HEADER	"ps:"

#define	NUMOF( array )	( sizeof( array ) / sizeof( ( array )[ 0 ] ) )
#define OPTIONLIST_VALUE_DELIMITER ';'
#define EQUAL_AND_SEMICOLON_SIZE 2u
#define LEN_KEY_INFO_VALUE_LENGTH 16
#define LEN_KEY_OPT_LIST_INFO_NUM 2

#define PS_NAME_SPACE_SIZE		3u

typedef struct
{
	const char *key1;
	const char *value1;
	const char *key2;
	const char *value2;
	const char *key3;
	const char *value3;
	const char *change_key;
	const char *change_value;
} ChangeOptListTable;

static ChangeOptListTable *get_change_option_list_table( const int id );
static ParamList *get_p_job_attr_list( const char * const p_job_attr );
static void change_p_job_attr_list( ParamList **p_list );

enum {
	ARGV_GS,
	ARGV_RESO,
	ARGV_WIDTH,
	ARGV_HEIGHT,
	ARGV_DEVICE,
	ARGV_DRIVER,
	ARGV_MODEL,
	ARGV_JOB,
};

char *org_cmd_arg[] = {
NULL,
"-r",
"-dDEVICEWIDTH=",
"-dDEVICEHEIGHT=",
"-sDEVICE=",
"-sDriver=",
"-sModel=",
"-sJobInfo=",
"-q",
"-dBATCH",
"-dSAFER",
"-dQUIET",
"-dNOPAUSE",
"-sFastImage=All",
"-sOutputFile=-",
"-",
NULL
};

#define	ARGV_NUM	(sizeof(org_cmd_arg)/sizeof(char*))

int g_filter_pid = -1;
int g_signal_received = 0;

#ifdef	DEBUG_IN_PS
int g_in_log_fd = -1;
#endif

int g_in_debug_fd = -1;

const char *ppd_opt_name[] =
{
	"PageSize",
	"MediaType",
	"InputSlot",
	"CNInserter",
	"Duplex",
	"BookLet",
	"BindEdge",
	"Collate",
	"StapleLocation",
	"CNPunch",
	"CNPunchLocation",
	"CNOutputPartition",
	"CNPureBlackProcess",
	"CNJobExecMode",
	"CNSaddleStitch",
	"CNZfolding",
	"CNTrimming",
	"CNTrimmingValue",
	"CNRotatePrint",
	"CNShiftPrint",
	"OutputBin",
	"Resolution",
	"CNTonerSaving",
	"CNColorMode",
	"CNBlackSubstitute",
	"CNColorMatchingMode",
	"CNOutputFace",
	"CNImageSpecialProcess",
	"CNCopySetNumbering",
	"CNDraftMode",
	"CNInterleafSheet",
	"CNInterleafMediaType",
	"CNInterleafPrint",
	"CNImageRefinement",
	"CNGradationSmoothing",
	"CNLineRefinement",
	"CNSpecialPrintMode",
	"CNSetFrontSheet",
	"CNFrontPrintOn",
	"CNFrontPaperSrc",
	"CNSetBackSheet",
	"CNBackPrintOn",
	"CNBackPaperSrc",
	"CNAdvancedSmoothing",
	"CNOHPPrintMode",
	"CNFinisher",
	"CNSmallSizeCurl",
	"CNShiftStartPrintPosition",
	"CNSharpness",
	"CNCreep",
	"CNDisplacementCorrection",
	"CNAdjustTrim",
	"CNCfolding",
	"CNCfoldSetting",
	"CNTrustPrint",
	"CNLineControl",
	"CNTonerVolumeAdjustment",
	"CNSpecialPrintAdjustmentA",
	"CNSpecialPrintAdjustmentB",
	"CNHalfFolding",
	"CNAccordionZfolding",
	"CNDoubleParallelFolding",
	"CNHalfFoldSetting",
	"CNAccordionZfoldSetting",
	"CNDoubleParallelFoldSetting",
	"CNVfolding",
	"CNSpecialSmooth",
	"CNTonerDensity",
	"CNGlossyPlainPaperProc",
	"CNMultiPunch",
	"CNBarCodeMode",
	"CNHyperSharpness",
	"CNTonerVolumeAdjustment",
	"CNUseCSModeJobAccount",
	"CNUseCSModeSecured",
	"CNWidthAdjustment",
	"CNOptimizeImageComp",
	"CNColorTonerVolumeAdjustment",
	"CNImprovePrintSpeed",
	"CNSpecialPrintAdjustmentC",
	"CNSpecifyNumOfCopiesStack",
	"CNTimingOfAutoEjectStacker",
	"CNPlainPaperCurlCorrect",
	"CNSaddleFolding",
	"CNSaddleFoldSetting",
	"CNSaddleFoldSheetsNum",
	"CNSaddleCfolding",
	"CNSaddleCfoldSetting",
	"CNHighBlackDensity"
};

char *cmd_opt_name[] =
{
	"CNBindEdgeShift",
	"CNJobAccount",
	"CNSecuredPrint",
	"CNMailBox",
	"CNBookletShift",
	"CNDocName",
	"CNUsrName",
	"CNProofPrint",
	"number-up",
	"CNStartingNumber",
	"CNDisableJobAccountingBW",
	"CNOffsetNum",
	"CNShiftFrShortEdge",
	"CNShiftFrLongEdge",
	"CNShiftBkShortEdge",
	"CNShiftBkLongEdge",
	"CNGutterShiftNum",
	"CNAdjustTrimNum",
	"CNStackCopiesNum"
};

char *drv_tbl_name[] =
{
	"ColorDevice",
	"PCFileName",
	"CNTblFormat",
	"CNTblInputSlot",
	"CNTblDuplex",
	"CNTblOutputBin",
	"CNTblSubstitute",
	"CNTblMediaType",
	"CNModelMethod",
	"CNModelMethod2",
	"CNMaxBoxNum",
	"CNMaxDocStr",
	"CNTblOutputPartition",
	"CNMaxCopies",
	"CNMaxGutter",
	"CNExecuteMethod",
	"CNUIOffsetMax",
	"CNShiftStartPrintPosType",
	"CNShiftPositionMin",
	"CNShiftPositionMax",
	"CNMINGutterShiftNum",
	"CNMAXGutterShiftNum",
	"CNDefaultGutterShiftNum",
	"CNUITrimValMin",
	"CNUITrimValMax",
	"CNUIAdjustTrimNumDefault",
	"CN_Filter_DebugMode",
	"CN_Cnpkmodule_DebugMode",
	"CNTblTrayAlias",
	"CNChangeOptList",
	"CNSaddleFoldMaxSheets",
	"CNSaddleCfoldMaxSheets"
};

char *org_page_size_table[] =
{
	"jenv_you_chou3",
};

enum {
	OPVP_MODEL,
	OPVP_DRIVER,
	OPVP_DEVICE,
};

char *mdl_tbl_name[] =
{
	"opvpModel",
	"opvpDriver",
	"opvpDevice",
};

char *device_type_key_name[] = {
#if DEBUGLOG
  "CNDeviceType",
#endif
  "CNTblMediaType",
  "CNModelMethod",
  "CNModelMethod2",
  "CNExecuteMethod",
  "CNTblInputSlot",
  "CNTblDuplex",
  "CNTblOutputBin",
  "CNTblOutputPartition",
  "CNMaxBoxNum",
  "CNMaxDocStr",
  "CNTblTrayAlias",
  "CNChangeOptList"
};

typedef struct device_type_param_list DeviceTypeParamList;
struct device_type_param_list {
  const char	*pCNDeviceType;
  const char	*pCNTblMediaType;
  const char	*pCNModelMethod;
  const char	*pCNModelMethod2;
  const char	*pCNExecuteMethod;
  const char	*pCNTblInputSlot;
  const char	*pCNTblDuplex;
  const char	*pCNTblOutputBin;
  const char	*pCNTblOutputPartition;
  const char	*pCNMaxBoxNum;
  const char	*pCNMaxDocStr;
  const char	*pCNTblTrayAlias;
  const char	*pCNChangeOptList;
};

DeviceTypeParamList device_type_param_list_000[] =
{
	{ "1",	"36",	"16777532",		"0",	"0",	"4",	"1",	"5",	"0",	"99",	"32",	"0",	"0"},
	{ "2",	"1",	"16777532",		"0",	"0",	"1",	"1",	"3",	"0",	"99",	"32",	"0",	"0"},
	{ "3",	"2",	"16777516",		"0",	"0",	"2",	"1",	"0",	"0",	"99",	"32",	"0",	"0"},
	{ "4",	"1",	"16777476",		"0",	"0",	"1",	"1",	"0",	"0",	"99",	"32",	"0",	"0"},
	{ "5",	"2",	"16777216",		"0",	"0",	"2",	"1",	"0",	"0",	"99",	"32",	"0",	"0"},
	{ "6",	"0",	"16777216",		"0",	"0",	"1",	"1",	"0",	"0",	"99",	"32",	"0",	"0"},
	{ "7",	"88",	"16777516",		"0",	"0",	"1",	"1",	"4",	"0",	"99",	"32",	"0",	"0"},
	{ "8",	"85",	"16777516",		"0",	"0",	"1",	"1",	"4",	"0",	"99",	"32",	"0",	"0"},
	{ "9",	"2",	"16777516",		"0",	"0",	"2",	"1",	"7",	"0",	"99",	"32",	"0",	"0"},
	{ "10",	"0",	"16777516",		"0",	"0",	"1",	"1",	"0",	"0",	"99",	"32",	"0",	"0"},
	{ "11",	"12294","25174972",		"0",	"0",	"2",	"1",	"10",	"2",	"99",	"32",	"1",	"0"},
	{ "12",	"12294","1099236284",		"0",	"0",	"1",	"1",	"10",	"2",	"99",	"32",	"2",	"0"},
	{ "13",	"77830","1099366844",		"0",	"0",	"1",	"1",	"10",	"2",	"99",	"32",	"2",	"0"},
	{ "14",	"65542","4661564",		"0",	"0",	"3",	"1",	"10",	"0",	"0",	"32",	"0",	"0"},
	{ "15",	"339968","897917884",	"1",	"0",	"6",	"1",	"11",	"0",	"99",	"32",	"4",	"1"},
	{ "None","0",	"16777216",		"0",	"0",	"0",	"1",	"0",	"0",	"99",	"32",	"0",	"0"},
	{ NULL,	NULL,	NULL,			NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL}
};
DeviceTypeParamList device_type_param_list_001[] =
{
	{ "1",	"0",	"16",			"0",	"0",	"2",	"1",	"0",	"0",	"99",	"32",	"0",	"0"	},
	{ "2",	"0",	"0",			"0",	"0",	"1",	"1",	"0",	"0",	"99",	"32",	"0",	"0"	},
	{ "3",	"0",	"3",			"0",	"0",	"0",	"1",	"0",	"0",	"63",	"15",	"0",	"0"	},
	{ "4",	"0",	"8860",			"0",	"0",	"2",	"1",	"2",	"0",	"99",	"32",	"0",	"0"	},
	{ "5",	"0",	"12",			"0",	"0",	"1",	"1",	"0",	"0",	"99",	"32",	"0",	"0"	},
	{ "6",	"0",	"12",			"0",	"0",	"1",	"1",	"6",	"0",	"99",	"32",	"0",	"0"	},
	{ "7",	"0",	"12",			"0",	"0",	"1",	"1",	"6",	"2",	"99",	"32",	"0",	"0"	},
	{ "8",	"0",	"28",			"0",	"0",	"2",	"1",	"0",	"0",	"99",	"32",	"0",	"0"	},
	{ "9",	"0",	"156",			"0",	"0",	"2",	"1",	"5",	"0",	"99",	"32",	"0",	"0"	},
	{ "10",	"6",	"8659612",		"0",	"0",	"2",	"1",	"10",	"2",	"99",	"32",	"1",	"0"	},
	{ "11",	"6",	"8650780",		"0",	"0",	"1",	"1",	"10",	"2",	"99",	"32",	"0",	"0"	},
	{ "12",	"45088","8650908",		"0",	"0",	"5",	"1",	"11",	"0",	"99",	"32",	"3",	"0"	},
	{ "13",	"6",	"12845084",		"0",	"0",	"1",	"1",	"10",	"2",	"99",	"32",	"0",	"0"	},
	{ "14",	"262144","1220813468",	"8",	"0",	"2",	"1",	"11",	"2",	"99",	"32",	"1",	"0"	},
	{ "None","0",	"0",			"0",	"0",	"0",	"1",	"0",	"0",	"63",	"15",	"0",	"0"	},
	{ NULL,	NULL,	NULL,			NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL}
};
DeviceTypeParamList device_type_param_list_002[] =
{
	{ "1",	"0",	"3",	"0",	"0",	"0",	"0",	"0",	"0",	"0",	"15",	"0",	"0"	},
	{ "2",	"12",	"268",	"0",	"0",	"0",	"1",	"0",	"0",	"0",	"15",	"0",	"0"	},
	{ "3",	"144",	"12",	"0",	"0",	"3",	"1",	"0",	"0",	"0",	"15",	"0",	"0"	},
	{ "4",	"4118",	"12",	"0",	"2144",	"3",	"1",	"0",	"0",	"99",	"15",	"0",	"0"	},
	{ "None","0",	"0",	"0",	"0",	"0",	"1",	"0",	"0",	"0",	"15",	"0",	"0"	},
	{ NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL}
};
DeviceTypeParamList device_type_param_list_003[] =
{
	{ "1",	"0",	"3",			"0",	"0",	"0",	"0",	"0",	"1",	"0",	"15",	"0",	"0"	},
	{ "2",	"0",	"3",			"0",	"0",	"0",	"0",	"0",	"0",	"0",	"15",	"0",	"0"	},
	{ "3",	"1542",	"12",			"0",	"6162",	"3",	"1",	"0",	"0",	"0",	"0",	"0",	"0"	},
	{ "4",	"0",	"12",			"0",	"0",	"1",	"1",	"6",	"2",	"99",	"32",	"0",	"0"	},
	{ "5",	"4",	"12",			"0",	"12",	"3",	"0",	"0",	"0",	"0",	"15",	"0",	"0"	},
	{ "6",	"1540",	"12",			"0",	"2",	"3",	"0",	"0",	"0",	"0",	"15",	"0",	"0"	},
	{ "7",	"1540",	"12",			"0",	"2",	"3",	"1",	"0",	"0",	"0",	"15",	"0",	"0"	},
	{ "8",	"6",	"1077936140",	"8",	"2064",	"7",	"1",	"0",	"0",	"0",	"0",	"0",	"2"	},
	{ "None","0",	"0",			"0",	"0",	"0",	"1",	"0",	"0",	"0",	"0",	"0",	"0"	},
	{ NULL,	NULL,	NULL,			NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL}
};

static ChangeOptListTable change_option_list_table_0 [] =
{
	{NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL}
};

static ChangeOptListTable change_option_list_table_1 [] =
{
	{"Collate", "StapleCollate", "CNOutputPartition", "offset",		 NULL, NULL, "CNOffsetNum", "1"},
	{"Collate", "StapleGroup",   "CNOutputPartition", "offset",		 NULL, NULL, "CNOffsetNum", "1"},
	{"Collate", "Group",         "CNOutputPartition", "offset",		 NULL, NULL, "CNOffsetNum", "1"},
	{NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL}
};

static ChangeOptListTable change_option_list_table_2[] =
{
	{"InputSlot", "Env",	"PageSize", "jenv_you_chou3",	NULL,	NULL,	"MediaType", "Envelope"},
	{"InputSlot", "Env",	"PageSize", "jenv_chou3",	NULL,	NULL,	"MediaType", "Envelope"},
	{"InputSlot", "Env",	"PageSize", "Postcard",		NULL,	NULL,	"MediaType", "Postcard"},
	{"InputSlot", "Env",	"PageSize", "dbl_postcard",	NULL,	NULL,	"MediaType", "Postcard"},
	{NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL}
};

static ChangeOptListTable* change_option_list_table_list[] =
{
	change_option_list_table_0,
	change_option_list_table_1,
	change_option_list_table_2
};


#ifdef DEBUGLOG
	#include <stdarg.h>

	static void debuglog(int indent, const char *message, ...)
	{
		static int debug_indent;
		va_list	ap;
		int	i;
		FILE *fp = fopen("/tmp/pstolipscpcar.log", "a");

		if (indent == INDENT_LEFT)
		{
			debug_indent--;
			if (debug_indent < 0){
				debug_indent = 0;
			}
		}

		for (i = 0; i < debug_indent; i++)
		{
			fprintf(fp, "\t");
		}

		va_start(ap, message);
		vfprintf(fp, message, ap);

		fclose(fp);

		if (indent == INDENT_RIGHT)
		{
			debug_indent++;
		}
	}
#else
	#define	debuglog(a, b, ...)
#endif

static char z_Val2Code(
	int nVal
){
	char cVal = 0x00;

	if((0x00 <= nVal)&&(nVal <= 0x19)){
		cVal = ('A' + nVal);
	}
	else if((0x1a <= nVal)&&(nVal <= 0x33)){
		cVal = ('a' + (nVal - 0x1a));
	}
	else if((0x34 <= nVal)&&(nVal <= 0x3d)){
		cVal = ('0' + (nVal - 0x34));
	}
	else if(nVal == 0x3e){
		cVal = '+';
	}
	else if(nVal == 0x3f){
		cVal = '/';
	}
	else{
		cVal = 0x00;
	}

	return cVal;
}

static int z_GetVal(
	const unsigned char *pSrc,
	long lSrcByte,
	long nChar
){
	long lStartByte = 0;
	unsigned int nSrcData = 0;
	int nTmp = 0;
	bool bSingleByte = false;

	if(pSrc){

		lStartByte = (nChar * 6) / 8;

		bSingleByte = ((nChar * 6) % 8) ? false : true;

		if(bSingleByte){
			nSrcData = pSrc[lStartByte];
		}
		else{
			nSrcData = pSrc[lStartByte] << 8;
			if(lStartByte + 1 < lSrcByte){
				nSrcData |= pSrc[lStartByte + 1];
			}
		}

		nTmp = (nChar * 6) - (lStartByte * 8);
		nSrcData <<= nTmp;
		nSrcData &= 0x0000FFFF;
		nSrcData >>= nTmp;

		if(bSingleByte){
			nTmp = ((lStartByte + 1) * 8) - ((nChar + 1) * 6);
		}
		else{
			nTmp = ((lStartByte + 2) * 8) - ((nChar + 1) * 6);
		}
		nSrcData >>= nTmp;
	}

	return nSrcData;
}

static int util_encodeBase64(
	const unsigned char *pSrc,
	long lSrcLenInByte,
	char **pDst,
	long *lDstByte
){
	int err = DEF_NO_ERR;

	long lSrcByte = 0;
	char *pTmpDst = NULL;
	long lTmpDstEnableByte = 0;
	long lTmpDstAddByte = 0;

	if((pSrc == NULL)||(pDst == NULL)){
		error_proc(__FILE__,__LINE__,err);
		return err;
	}

	*pDst = NULL;
	if(lDstByte) *lDstByte = 0;

	lSrcByte = lSrcLenInByte;

	if(err == DEF_NO_ERR){
		lTmpDstEnableByte = ((lSrcByte * 8) / 6);
		if((lSrcByte * 8) % 6){
			lTmpDstEnableByte++;
		}
		if(lTmpDstEnableByte % 4){
			lTmpDstAddByte = (4 - (lTmpDstEnableByte % 4));
		}

		*pDst = (char *)calloc(1, lTmpDstEnableByte + lTmpDstAddByte + 1);
		if(*pDst == NULL){
			error_proc(__FILE__,__LINE__,err);
		}
		else{
			pTmpDst = *pDst;
		}
	}

	if(err == DEF_NO_ERR){
		long nCharNo = 0;
		int nSrcData = 0;
		for(nCharNo = 0; nCharNo < lTmpDstEnableByte; nCharNo++){
			nSrcData = z_GetVal(pSrc, lSrcByte, nCharNo);
			*(pTmpDst++) = z_Val2Code(nSrcData);
		}
		for(nCharNo = 0; nCharNo < lTmpDstAddByte; nCharNo++){
			*(pTmpDst++) = '=';
		}
	}

	if(err != DEF_NO_ERR){
		if(*pDst){
			free(*pDst);
			*pDst = NULL;
		}
		error_proc(__FILE__,__LINE__,err);
	}
	else{
        if(lDstByte){
            *lDstByte = lTmpDstEnableByte + lTmpDstAddByte;
        }
	}

	return err;
}


static
bool is_cmd_opt_name(char *name)
{
	int num = sizeof(cmd_opt_name) / sizeof(char *);
	int i;

	for( i = 0; i < num; i++ )
	{
		if( strcmp(name, cmd_opt_name[i]) == 0 )
		{
			return true;
		}
	}
	return false;
}

static
bool is_drv_tbl_name(char *name)
{
	int num = sizeof(drv_tbl_name) / sizeof(char *);
	int i;

	for( i = 0; i < num; i++ )
	{
		if( strcmp(name, drv_tbl_name[i]) == 0 )
		{
			return true;
		}
	}
	return false;
}

static
bool is_org_page_size_table(ppd_file_t *p_ppd)
{
	int num = sizeof(org_page_size_table) / sizeof(char *);
	ppd_choice_t *p_choice = NULL;
	int i;

	p_choice = ppdFindMarkedChoice(p_ppd, "PageSize");

	if(p_choice != NULL)
	{
		for( i = 0; i < num; i++ )
		{
			if( strncmp(p_choice->choice, org_page_size_table[i], strlen(org_page_size_table[i]) ) == 0 )
			{
				return true;
			}
		}
	}
	return false;
}

static
int get_CNCopies_value(cups_option_t *p_cups_opt, int num_opt, int copies, char *p_value_buf, int value_len)
{
	cups_option_t *p_opt = p_cups_opt;
	char *cncopies = NULL;
	int cncopies_len = 0;
	int i;

	for( i = 0; i < num_opt; i++)
	{
		if(strcmp(p_opt->name, "CNCopies") == 0)
		{
			cncopies = p_opt->value;
			break;
		}
		p_opt++;
	}

	if(copies > 1)
	{
		snprintf(p_value_buf, value_len, "%d", copies);
		cncopies_len = strlen(p_value_buf);
	}
	else if(copies == 1)
	{
		if(cncopies == NULL)
		{
			snprintf(p_value_buf, value_len, "%d", copies);
			cncopies_len = strlen(p_value_buf);
		}
		else
		{
			strncpy(p_value_buf, cncopies, value_len);
			cncopies_len = strlen(p_value_buf);
		}
	}
	else if(copies == -1 && cncopies != NULL)
	{
		strncpy(p_value_buf, cncopies, value_len);
		cncopies_len = strlen(p_value_buf);
	}
	else if(copies == -2)
	{
		strcpy(p_value_buf, "1");
		cncopies_len = strlen(p_value_buf);
	}

	return cncopies_len;
}

static DeviceTypeParamList *find_device_type_param_list( ppd_file_t *p_ppd, char* p_table )
{
	char *device_type = NULL;
	char *file_name = NULL;
	ppd_choice_t *p_choice = NULL;
	DeviceTypeParamList *device_type_param_list = NULL;

	p_choice = ppdFindMarkedChoice( p_ppd, "CNDeviceType" );
	if( p_choice != NULL )
	{
		device_type = p_choice->choice;
	}

	if( device_type != NULL )
	{
		file_name = strstr( p_table, "PCFileName" );
		file_name = strchr( file_name, '=' );
		file_name++;
	}

	if( file_name != NULL )
	{
		if( !strncasecmp( file_name, "CNMFPCLJ.PPD", strlen("CNMFPCLJ.PPD") ) )
				device_type_param_list = device_type_param_list_000;
		else if( !strncasecmp( file_name, "CNMFPBWLJ.PPD", strlen("CNMFPBWLJ.PPD") ) )
				device_type_param_list = device_type_param_list_001;
		else if( !strncasecmp( file_name, "CNSFPCLJ.PPD", strlen("CNSFPCLJ.PPD") ) )
				device_type_param_list = device_type_param_list_002;
		else if( !strncasecmp( file_name, "CNSFPBWLJ.PPD", strlen("CNSFPBWLJ.PPD") ) )
				device_type_param_list = device_type_param_list_003;
	}

	if( device_type_param_list != NULL )
	{
		while( device_type_param_list->pCNDeviceType != NULL )
		{
			if( !strcasecmp( device_type_param_list->pCNDeviceType, device_type ) )
				return device_type_param_list;
			device_type_param_list++;
		}
	}

	return NULL;
}

static ChangeOptListTable* get_change_option_list_table( const int id )
{
	int table_id = id;
	const int num = sizeof(change_option_list_table_list) / sizeof(ChangeOptListTable *);

	if ((0 > table_id) || (table_id > num))
	{
		table_id = 0;
	}
	return change_option_list_table_list[table_id];
}

static ParamList *get_p_job_attr_list( const char * const p_job_attr )
{
	ParamList *p_list = NULL;
	char *pName = NULL;
	char *pValue = NULL;
	char *pTemp = NULL;
	char *pOptList = NULL;
	unsigned int idx = 0;
	int add_value_size = 0;

	pTemp = strdup(p_job_attr);
	pOptList = pTemp;

	struct _OptListInfo {
		const char	*pKeyName;
		const char	*pKeyNameValueLen;
	}
	const OptListInfo[LEN_KEY_OPT_LIST_INFO_NUM] = {
		{ "CNUsrName",		"CNUsrNameLen"		},
		{ "CNDocName",		"CNDocNameLen"		}
	};

	if( pOptList != NULL ) {
		while( *pOptList != '\0') {
			while( isspace(*pOptList) != 0 ) {
				pOptList++;
			}

			if( strncasecmp(pOptList, "ps:", PS_NAME_SPACE_SIZE) == 0 ) {
				pOptList += PS_NAME_SPACE_SIZE;
			}
			pName = pOptList;

			while( (*pOptList != '=')  && (*pOptList != '\0') ) {
				pOptList++;
			}
			*pOptList = '\0';
			pOptList++;

			while( isspace(*pOptList) != 0 ) {
				pOptList++;
			}
			pValue = pOptList;

			for( idx = 0; idx < NUMOF(OptListInfo); idx++ ) {
				if( strcasecmp( pName, OptListInfo[ idx ].pKeyName ) == 0 ) {
					if( strstr( p_job_attr, OptListInfo[ idx ].pKeyNameValueLen ) != NULL) {
						break;
					}
				}
			}

			if( idx < NUMOF(OptListInfo) ) {
				char *pLenValue = NULL;
				pLenValue = strstr( p_job_attr, OptListInfo[ idx ].pKeyNameValueLen );
				pLenValue += ( strlen( OptListInfo[ idx ].pKeyNameValueLen ) + 1u );

				pOptList += atoi( pLenValue );
				if( pOptList != NULL ) {
					if( *pOptList != '\0' ) {
						*pOptList = '\0';
						pOptList++;
					}
				}
			}
			else {
				while( (*pOptList != ';') && (*pOptList !='\0') ) {
					pOptList++;
				}
				if( *pOptList != '\0' ) {
					*pOptList = '\0';
					pOptList++;
				}
			}
			add_value_size = (int)(strlen(pValue)+1u) ;
			param_list_add( &p_list, pName, pValue, add_value_size );
		}
	}
	free(pTemp);
	return p_list;
}

static char* get_compress_job_attr(char* p_job_attr)
{
	unsigned long compBondSize = 0;
	long base64Len = 0;
	char* comp_job_attr = NULL;
	unsigned char* dst_comp_str = NULL;
	char* dst_base64_str = NULL;
	int size = 0;
	int errorNo = 0;
	char* temp_job_attr = p_job_attr;
	int	 ps_header_size = 0;

	if (p_job_attr == NULL) {
		return NULL;
	}

	if(strncmp(p_job_attr, PS_HEADER, strlen(PS_HEADER)) == 0 )
	{
		ps_header_size = strlen(PS_HEADER);
		temp_job_attr += ps_header_size;
	}

	compBondSize = compressBound(strlen(temp_job_attr));
	if (compBondSize <= 0) {
		debuglog(INDENT_NONE, "get_compress_job_attr(): Fail to call the compressBound()\n");
		return NULL;
	}

	dst_comp_str = (unsigned char*)malloc(compBondSize);
	if (dst_comp_str == NULL) {
		debuglog(INDENT_NONE, "get_compress_job_attr(): Fail to call the malloc()\n");
		return NULL;
	}

	memset(dst_comp_str, 0, compBondSize);
	errorNo = compress((Bytef*)dst_comp_str, &compBondSize, (Bytef*)temp_job_attr, strlen(temp_job_attr));

	if (errorNo == Z_OK)
	{
		if (util_encodeBase64(dst_comp_str, compBondSize, &dst_base64_str, &base64Len) == DEF_NO_ERR)
		{
			size = ps_header_size + base64Len + strlen(CN_JOB_INFO_HEADER) + 1;
			comp_job_attr = (char*)malloc(size);

			if (comp_job_attr != NULL)
			{
				memset(comp_job_attr, 0, size);
				if (ps_header_size != 0)
				{
					strncpy(comp_job_attr, PS_HEADER, size - 1);
					strncat(comp_job_attr, CN_JOB_INFO_HEADER, size - strlen(comp_job_attr) - 1);
					strncat(comp_job_attr, dst_base64_str, size - strlen(comp_job_attr) - 1);
				}
				else
				{
					strncpy(comp_job_attr, CN_JOB_INFO_HEADER, size - 1);
					strncat(comp_job_attr, dst_base64_str, size - strlen(comp_job_attr) - 1);
				}
			}
			else
			{
				debuglog(INDENT_NONE, "get_compress_job_attr(): Fail to call the malloc()\n");
			}
		}
		else
		{
			debuglog(INDENT_NONE, "get_compress_job_attr(): Fail to call the util_encodeBase64()\n");
		}
	}
	else
	{
		debuglog(INDENT_NONE, "get_compress_job_attr(): Fail to call the compress()\n");
	}


	if (dst_comp_str) {
		free(dst_comp_str);
	}

	if (dst_base64_str) {
		free(dst_base64_str);
	}

	debuglog(INDENT_NONE, "get_compress_job_attr():The compress job info is [%s]\n", comp_job_attr);
	return comp_job_attr;

}

static void change_p_job_attr_list( ParamList ** const p_list )
{
	int			table_id = 0;
	int			change_value_size = 0;
	ParamList*	curs = NULL;
	ParamList*	curs2 = NULL;
	ParamList*	curs3 = NULL;
	ChangeOptListTable*	change_option_list_table	= NULL;
	ChangeOptListTable*	p_table						= NULL;

	curs = param_list_find(*p_list, "CNChangeOptList");
	if (curs != NULL)
	{
		table_id = atoi(curs->value);
	}

	change_option_list_table = get_change_option_list_table(table_id);
	if (change_option_list_table != NULL)
	{
		p_table = change_option_list_table;

		while (p_table->key1 != NULL)
		{
			curs = param_list_find(*p_list, p_table->key1);
			if (curs != NULL)
			{
				if (strcasecmp(p_table->value1, curs->value) == 0)
				{
					if (p_table->key2 == NULL)
					{
						change_value_size = (int)(strlen(p_table->change_value)+1u);
						param_list_add(p_list, p_table->change_key, p_table->change_value, change_value_size);
					}
					else
					{
						curs2 = param_list_find(*p_list, p_table->key2);
						if (curs2 != NULL)
						{
							if (strcasecmp(p_table->value2, curs2->value) == 0)
							{
								if (p_table->key3 == NULL)
								{
									change_value_size = (int)(strlen(p_table->change_value)+1u);
									param_list_add(p_list, p_table->change_key, p_table->change_value, change_value_size);
								}
								else
								{
									curs3 = param_list_find(*p_list, p_table->key3);
									if (curs3 != NULL)
									{
										if (strcasecmp(p_table->value3, curs3->value) == 0)
										{
											change_value_size = (int)(strlen(p_table->change_value)+1u);
											param_list_add(p_list, p_table->change_key, p_table->change_value, change_value_size);
										}
									}
								}
							}
						}
					}
				}
			}
			p_table++;
		}
	}

	return;
}

static
char *ppd_mark_to_job_attr(ppd_file_t *p_ppd, cups_option_t *p_cups_opt,
	 int num_opt, char *p_table, int copies, char *user, char *title)
{
	int ppd_opt_num = sizeof(ppd_opt_name) / sizeof(char*);
	unsigned int opt_len = strlen(p_table);
	char *p_job_attr = NULL;
	ppd_choice_t *p_choice;
	cups_option_t *p_opt;
	DeviceTypeParamList *device_type_param_list = NULL;
	char value[MAX_VALUE_LEN + 1];
	int cncopies_len = 0;
	int is_docname = 0;
	int is_usrname = 0;

	char	*p_tmp = NULL;
	unsigned int	docname_len = 0;
	unsigned int	usrname_len = 0;
	char		cBuf[LEN_KEY_INFO_VALUE_LENGTH];
	unsigned int	idx = 0;

	struct _LENKEYINFO {
		const char	*pKeyName;
		const char	*pKeyNameLen;
		unsigned int	*pDataSize;
		int		*pDataFlag;
	} len_key_info[LEN_KEY_OPT_LIST_INFO_NUM] = {
		{ "CNUsrName",		"CNUsrNameLen",		&usrname_len,		&is_usrname	},
		{ "CNDocName",		"CNDocNameLen",		&docname_len,		&is_docname	},
	};

	int i;

	ParamList*	p_list = NULL;
	ParamList*	curs = NULL;

	memset(value, 0, MAX_VALUE_LEN + 1);

	cncopies_len
	 = get_CNCopies_value(p_cups_opt, num_opt, copies, value, MAX_VALUE_LEN);
	if(cncopies_len > 0) {
		opt_len += strlen("CNCopies") + (unsigned int)cncopies_len + EQUAL_AND_SEMICOLON_SIZE;
	}

	for( i = 0 ; i < ppd_opt_num ; i++ )
	{
		p_choice = ppdFindMarkedChoice(p_ppd, ppd_opt_name[i]);
		if( p_choice != NULL )
		{
			opt_len += strlen(ppd_opt_name[i])
					+ strlen(p_choice->choice)
					+ EQUAL_AND_SEMICOLON_SIZE;
		}
	}
	for( p_opt = p_cups_opt, i = 0 ; i < num_opt ; i++ )
	{
		if( is_cmd_opt_name(p_opt->name) )
		{
			opt_len += strlen(p_opt->name)
					+ strlen(p_opt->value)
					+ EQUAL_AND_SEMICOLON_SIZE;

			for( idx = 0; idx < NUMOF( len_key_info ); idx++ ) {
				if( strcasecmp( p_opt->name, len_key_info[idx].pKeyName ) == 0 ) {
					if( len_key_info[idx].pDataFlag != NULL ) {
						*len_key_info[idx].pDataFlag = 1;
					}
		        		p_tmp = strchr( p_opt->value, OPTIONLIST_VALUE_DELIMITER );
					if( p_tmp != NULL ) {
						*len_key_info[idx].pDataSize = strlen( p_opt->value );
						sprintf( cBuf, "%d",  *len_key_info[idx].pDataSize );
						opt_len += strlen( len_key_info[idx].pKeyNameLen );
						opt_len += strlen( cBuf );
						opt_len += EQUAL_AND_SEMICOLON_SIZE;
					}
					break;
				}
			}
		}
		p_opt++;
	}

	if( (is_docname == 0) && (title != NULL) ){
		opt_len += strlen("CNDocName") + strlen(title) + EQUAL_AND_SEMICOLON_SIZE;
		p_tmp = strchr( title, OPTIONLIST_VALUE_DELIMITER );
		if( p_tmp != NULL ) {
			docname_len = strlen( title );
			sprintf( cBuf, "%d", docname_len );
			opt_len += strlen( "CNDocNameLen" );
			opt_len += strlen( cBuf );
			opt_len += EQUAL_AND_SEMICOLON_SIZE;
		}
	}
	if( (is_usrname == 0) && (user != NULL) ){
		opt_len += strlen("CNUsrName") + strlen(user) + EQUAL_AND_SEMICOLON_SIZE;
		p_tmp = strchr( user, OPTIONLIST_VALUE_DELIMITER );
		if( p_tmp != NULL ) {
			usrname_len = strlen( user );
			sprintf( cBuf, "%d", usrname_len );
			opt_len += strlen( "CNUsrNameLen" );
			opt_len += strlen( cBuf );
			opt_len += EQUAL_AND_SEMICOLON_SIZE;
		}
	}
	if(user) {
		opt_len += strlen("CNJobUser") + strlen(user) + EQUAL_AND_SEMICOLON_SIZE;
	}

	device_type_param_list = find_device_type_param_list( p_ppd, p_table );
	if( device_type_param_list != NULL )
	{
		for( i = 0; i < ( sizeof(device_type_key_name) / sizeof(char*) ); i++ )
		{
			opt_len += strlen( device_type_key_name[i] );
			opt_len += EQUAL_AND_SEMICOLON_SIZE;
		}
#if DEBUGLOG
		opt_len += strlen( device_type_param_list->pCNDeviceType );
#endif
		opt_len += strlen( device_type_param_list->pCNTblMediaType );
		opt_len += strlen( device_type_param_list->pCNModelMethod );
		opt_len += strlen( device_type_param_list->pCNModelMethod2 );
		opt_len += strlen( device_type_param_list->pCNExecuteMethod );
		opt_len += strlen( device_type_param_list->pCNTblInputSlot );
		opt_len += strlen( device_type_param_list->pCNTblDuplex );
		opt_len += strlen( device_type_param_list->pCNTblOutputBin );
		opt_len += strlen( device_type_param_list->pCNTblOutputPartition );
		opt_len += strlen( device_type_param_list->pCNMaxBoxNum );
		opt_len += strlen( device_type_param_list->pCNMaxDocStr );
		if( strcasecmp( device_type_param_list->pCNTblTrayAlias, "0" ) != 0 )
		{
			opt_len += strlen( device_type_param_list->pCNTblTrayAlias );
		}
		else
		{
			opt_len -= ( strlen("CNTblTrayAlias") + EQUAL_AND_SEMICOLON_SIZE );
		}
		if( strcasecmp( device_type_param_list->pCNChangeOptList, "0" ) != 0 )
		{
			opt_len += strlen( device_type_param_list->pCNChangeOptList );
		}
		else
		{
			opt_len -= ( strlen("CNChangeOptList") + EQUAL_AND_SEMICOLON_SIZE );
		}
	}

        if(is_org_page_size_table(p_ppd))
        {
		p_choice = ppdFindMarkedChoice(p_ppd, "PageSize");
		if( p_choice != NULL )
		{
			opt_len += strlen("CNOrgPageSize") + strlen(p_choice->choice) + EQUAL_AND_SEMICOLON_SIZE;
		}
        }

	if( opt_len > 0 )
	{
		opt_len += 3;

		if( (p_job_attr = malloc(opt_len + 1)) != NULL )
		{
			strcpy(p_job_attr, "ps:");
			strcat(p_job_attr, p_table);

			for( i = 0 ; i < ppd_opt_num ; i++ )
			{
				p_choice = ppdFindMarkedChoice(p_ppd, ppd_opt_name[i]);
				if( p_choice != NULL )
				{
					strcat(p_job_attr, ppd_opt_name[i]);
					strcat(p_job_attr, "=");
					strcat(p_job_attr, p_choice->choice);
					strcat(p_job_attr, ";");
				}
			}

			if(cncopies_len > 0){
				strcat(p_job_attr, "CNCopies");
				strcat(p_job_attr, "=");
				strcat(p_job_attr, value);
				strcat(p_job_attr, ";");
			}

			for( p_opt = p_cups_opt, i = 0 ; i < num_opt ; i++ )
			{
				if( is_cmd_opt_name(p_opt->name) )
				{
					strcat(p_job_attr, p_opt->name);
					strcat(p_job_attr, "=");
					strcat(p_job_attr, p_opt->value);
					strcat(p_job_attr, ";");
				}
				p_opt++;
			}
			if( docname_len > 0 )
			{
				sprintf( cBuf, "%d", docname_len );

				strcat(p_job_attr, "CNDocNameLen");
				strcat(p_job_attr, "=");
				strcat(p_job_attr, cBuf);
				strcat(p_job_attr, ";");
			}
			if( usrname_len > 0 )
			{
				sprintf( cBuf, "%d", usrname_len );

				strcat(p_job_attr, "CNUsrNameLen");
				strcat(p_job_attr, "=");
				strcat(p_job_attr, cBuf);
				strcat(p_job_attr, ";");
			}

			if(is_usrname == 0 && user != NULL){
				strcat(p_job_attr, "CNUsrName");
				strcat(p_job_attr, "=");
				strcat(p_job_attr, user);
				strcat(p_job_attr, ";");
			}

			if(is_docname == 0 && title != NULL){
				strcat(p_job_attr, "CNDocName");
				strcat(p_job_attr, "=");
				strcat(p_job_attr, title);
				strcat(p_job_attr, ";");
			}

			if(user){
				strcat(p_job_attr, "CNJobUser");
				strcat(p_job_attr, "=");
				strcat(p_job_attr, user);
				strcat(p_job_attr, ";");
			}

        		if(is_org_page_size_table(p_ppd))
			{
				p_choice = ppdFindMarkedChoice(p_ppd, "PageSize");
				if( p_choice != NULL )
				{
					strcat(p_job_attr, "CNOrgPageSize");
					strcat(p_job_attr, "=");
					strcat(p_job_attr, p_choice->choice);
					strcat(p_job_attr, ";");
				}
			}

			if( device_type_param_list != NULL )
			{
#if DEBUGLOG
				strcat(p_job_attr, "CNDeviceType");
				strcat(p_job_attr, "=");
				strcat(p_job_attr, device_type_param_list->pCNDeviceType);
				strcat(p_job_attr, ";");
#endif
				strcat(p_job_attr, "CNTblMediaType");
				strcat(p_job_attr, "=");
				strcat(p_job_attr, device_type_param_list->pCNTblMediaType);
				strcat(p_job_attr, ";");

				strcat(p_job_attr, "CNModelMethod");
				strcat(p_job_attr, "=");
				strcat(p_job_attr, device_type_param_list->pCNModelMethod);
				strcat(p_job_attr, ";");

				strcat(p_job_attr, "CNModelMethod2");
				strcat(p_job_attr, "=");
				strcat(p_job_attr, device_type_param_list->pCNModelMethod2);
				strcat(p_job_attr, ";");

				strcat(p_job_attr, "CNExecuteMethod");
				strcat(p_job_attr, "=");
				strcat(p_job_attr, device_type_param_list->pCNExecuteMethod);
				strcat(p_job_attr, ";");

				strcat(p_job_attr, "CNTblInputSlot");
				strcat(p_job_attr, "=");
				strcat(p_job_attr, device_type_param_list->pCNTblInputSlot);
				strcat(p_job_attr, ";");

				strcat(p_job_attr, "CNTblDuplex");
				strcat(p_job_attr, "=");
				strcat(p_job_attr, device_type_param_list->pCNTblDuplex);
				strcat(p_job_attr, ";");

				strcat(p_job_attr, "CNTblOutputBin");
				strcat(p_job_attr, "=");
				strcat(p_job_attr, device_type_param_list->pCNTblOutputBin);
				strcat(p_job_attr, ";");

				strcat(p_job_attr, "CNTblOutputPartition");
				strcat(p_job_attr, "=");
				strcat(p_job_attr, device_type_param_list->pCNTblOutputPartition);
				strcat(p_job_attr, ";");

				strcat(p_job_attr, "CNMaxBoxNum");
				strcat(p_job_attr, "=");
				strcat(p_job_attr, device_type_param_list->pCNMaxBoxNum);
				strcat(p_job_attr, ";");

				strcat(p_job_attr, "CNMaxDocStr");
				strcat(p_job_attr, "=");
				strcat(p_job_attr, device_type_param_list->pCNMaxDocStr);
				strcat(p_job_attr, ";");

				if( strcasecmp( device_type_param_list->pCNTblTrayAlias, "0" ) != 0 )
				{
					strcat(p_job_attr, "CNTblTrayAlias");
					strcat(p_job_attr, "=");
					strcat(p_job_attr, device_type_param_list->pCNTblTrayAlias);
					strcat(p_job_attr, ";");
				}

				if( strcasecmp( device_type_param_list->pCNChangeOptList, "0" ) != 0 )
				{
					strcat(p_job_attr, "CNChangeOptList");
					strcat(p_job_attr, "=");
					strcat(p_job_attr, device_type_param_list->pCNChangeOptList);
					strcat(p_job_attr, ";");
				}

				debuglog(INDENT_NONE, "p_job_attr = [%s]\n", p_job_attr);
			}

			p_job_attr[opt_len - 1] = '\0';

			debuglog(INDENT_NONE, "p_job_attr = [%s]\n", p_job_attr);

			if( strstr( p_job_attr, "CNChangeOptList" ) != NULL )
			{
				p_list = get_p_job_attr_list( p_job_attr );

				change_p_job_attr_list( &p_list );

				free( p_job_attr );

				opt_len = PS_NAME_SPACE_SIZE;
				for( curs = p_list; curs != NULL; curs = curs->next )
				{
					opt_len += (strlen(curs->key) + strlen(curs->value) + EQUAL_AND_SEMICOLON_SIZE);
				}

				p_job_attr = malloc(opt_len + 1u);
				if( p_job_attr != NULL )
				{
					strcpy(p_job_attr, "ps:");

					for (curs = p_list; curs != NULL; curs = curs->next)
					{
						strcat(p_job_attr, curs->key);
						strcat(p_job_attr, "=");
						strcat(p_job_attr, curs->value);
						strcat(p_job_attr, ";");
					}
					p_job_attr[opt_len - 1u] = '\0';
					param_list_free( p_list );

					debuglog(INDENT_NONE, "p_job_attr(change) = [%s]\n", p_job_attr);
				}
			}

			if( strstr( p_job_attr, "CN_Filter_DebugMode" ) != NULL ) {
				g_in_debug_fd = open( "/tmp/debug_image.ps", O_WRONLY|O_CREAT|O_TRUNC, S_IWUSR );
			}
		}
	}
	return p_job_attr;
}

static
int GetKey(char *ptr, char *key, int size)
{
	int	key_len = 0;
	char	*p_key = NULL;

	p_key = key;
	while(1){
		if(*ptr == '*'){
			ptr++;
		}
		if(*ptr == '\n' || *ptr == '\r' || *ptr == '\0'){
			break;
		}
		if(*ptr == ' ' || *ptr == '\t'){
			break;
		}
		if(*ptr == ':'){
			break;
		}
		if(p_key - key >= size - 1){
			break;
		}
		*p_key = *ptr;
		ptr++;
		p_key++;
		key_len++;
	}
	*p_key = '\0';

	return key_len;
}

int GetValue(char *p_line_buf, char *p_value_buf, int value_len)
{
	int pos = 0;
	int val_pos = 0;

	while( IS_BLANK(p_line_buf[pos]) )
		pos++;
	while( p_line_buf[pos] == ':' )
		pos++;
	while( IS_BLANK(p_line_buf[pos]) )
		pos++;
	while( p_line_buf[pos] == '"')
		pos++;

	while( !IS_RETURN(p_line_buf[pos])
		&& p_line_buf[pos] != '"'
		&& p_line_buf[pos] != '\n'
		&& val_pos < value_len - 1 )
			p_value_buf[val_pos++] = p_line_buf[pos++];
	p_value_buf[val_pos++] = 0;

	return val_pos;
}

static
int GetUIValue(char *ptr, char *key, char *val, int size)
{
	char *p_key = NULL;
	char *p_val = NULL;
	int opt_flag = 0;

	p_key = key;
	while(1){
		if(*ptr == '\n' || *ptr == '\r' || *ptr == '\0')
			return 1;
		if(*ptr == ' ' || *ptr == '\t')
			ptr++;
		if(*ptr == '*')
			ptr++;
		if(*ptr == '(')
			break;
		if(p_key - key == size - 1)
			return 1;
		*p_key = *ptr;
		ptr++;
		p_key++;
	}
	*p_key = '\0';

	ptr++;
	p_val = val;
	while(1){
		if(*ptr == '\n' || *ptr == '\r' || *ptr == '\0')
			return 1;
		if(*ptr == ')')
			break;
		if(p_val - val == size - 1)
			break;
		*p_val = *ptr;
		ptr++;
		p_val++;
	}
	*p_val = '\0';

	if(ptr != NULL){
		if(strstr(ptr, "True") != NULL)
			opt_flag = 1;
	}

	return opt_flag;
}

static
int get_driver_table(char *ppd_file, char *p_buf, int buf_len)
{
	FILE	*fp = NULL;
	int	got_tbl_num = 0;

	if( ppd_file == NULL || p_buf == NULL || buf_len <= 0 )
	{
		return 0;
	}

	if( (fp = fopen(ppd_file, "r")) )
	{
		char	line_buf[LINE_BUF_SIZE];
		char	key[BUFSIZE];
		char	value[BUFSIZE];

		while( fgets(line_buf, sizeof(line_buf), fp) )
		{
			if( line_buf[0] == '*' )
			{
				int key_len = 0;
				int value_len = 0;
				int p_buf_len = 0;

				memset(key, 0, sizeof(key));
				memset(value, 0, sizeof(value));

				if( strncmp(line_buf, "*%CNUIValue:", 12 ) == 0 )
				{
					(void)GetUIValue(line_buf + 12, key, value, sizeof(key));
					key_len = strlen(key);
				}
				else
				{
					key_len = GetKey(line_buf + 1, key, sizeof(key));
					if(key_len)
					{
						(void)GetValue(line_buf + 1 + key_len, value, sizeof(value));
					}
				}

				value_len = strlen(value);

				if (key_len > 0 && value_len > 0)
				{
					p_buf_len = strlen(p_buf);

					debuglog(INDENT_NONE, "key:value = [%s]:[%s]\n", key, value);

					if( is_drv_tbl_name(key) )
					{
						int new_len = p_buf_len + key_len + value_len + 2;

						if( new_len < buf_len - 1 )
						{
							strcat(p_buf, key);
							strcat(p_buf, "=");
							strcat(p_buf, value);
							strcat(p_buf, ";");
						}
						got_tbl_num++;
					}
				}
			}
		}
		fclose(fp);
	}
	return got_tbl_num;
}

static
void append_lib_ext(char *p_driver, int buf_len)
{
	char *p_ext;
	int len = buf_len - strlen(p_driver) - 1;

	if( (p_ext = strrchr(p_driver, '.')) )
	{
		if( !strncmp(p_ext, ".so", 3) )
			*p_ext = '\0';
	}
#ifdef	APPLE_OS
	strncat(p_driver, ".dylib", len);
#else
	strncat(p_driver, ".so", len);
#endif
}

static
int get_driver_model(char *ppd_file,
	char *p_model, char *p_driver, char *p_device, int buf_len)
{
	FILE *fp = fopen(ppd_file, "r");
	int mdl_tbl_num = sizeof(mdl_tbl_name) / sizeof(char*);
	char line_buf[LINE_BUF_SIZE];
	int got_tbl_num = 0;

	if( fp && p_model != NULL && p_driver != NULL
	 	   && p_device != NULL && buf_len > 0 )
	{
		int cc;
		int pos = 0;

		*p_model  = '\0';
		*p_driver = '\0';
		*p_device = '\0';

		do
		{
			cc = fgetc(fp);

			if( cc != '\n' && cc != EOF )
			{
				if( pos < LINE_BUF_SIZE - 1 )
					line_buf[pos++] = cc;
			}
			else
			{
				int i;

				line_buf[pos] = 0;

				for( i = 0 ; i < mdl_tbl_num ; i++ )
				{
					int name_len = strlen(mdl_tbl_name[i]);
					char value[BUFSIZE];

					if( line_buf[0] == '*'
					 && strncmp(line_buf + 1, mdl_tbl_name[i], name_len) == 0 )
					{
						if( GetValue(line_buf + 1 + name_len, value, sizeof(value)) > 0 )
						{
							if( strlen(value) < buf_len - 1 )
							{
								switch( i )
								{
								case OPVP_MODEL:
									memset(p_model, 0, buf_len);
									strncpy(p_model, value, buf_len - 1);
									break;
								case OPVP_DRIVER:
									memset(p_driver, 0, buf_len);
									strncpy(p_driver, value, buf_len - 1);
									append_lib_ext(p_driver, buf_len);
									break;
								case OPVP_DEVICE:
									memset(p_device, 0, buf_len);
									strncpy(p_device, value, buf_len - 1);
									break;
								}
							}
							got_tbl_num++;
						}
					}
				}
				pos = 0;
			}
		}
		while( cc != EOF && got_tbl_num < mdl_tbl_num );

		fclose(fp);
	}
	return (got_tbl_num == mdl_tbl_num)? 1 : 0;
}

static
char *make_job_attr(ppd_file_t *p_ppd, char *p_ppd_name,
	 cups_option_t *p_cups_opt, int num_opt, int copies, char *user, char *title)
{
	char *p_job_attr = NULL;
	char table_buf[TABLE_BUF_SIZE];

	table_buf[0] = '\0';
	if( get_driver_table(p_ppd_name, table_buf, sizeof(table_buf)) )
	{
		debuglog(INDENT_NONE, "table_buf = [%s]\n", table_buf);

		p_job_attr = ppd_mark_to_job_attr(p_ppd,
			p_cups_opt, num_opt, table_buf, copies, user, title);
	}
	return p_job_attr;
}

static
int is_acro_util(char *p_data_buf, int read_bytes)
{
	if( strncmp(p_data_buf, ACRO_UTIL, ACRO_UTIL_LEN) == 0 )
		return 1;
	else
		return 0;
}

static
int is_end_resource(char *p_data_buf, int read_bytes)
{
	if( strncmp(p_data_buf, "%%EndResource", 13) == 0 )
		return 1;
	else
		return 0;
}

static
int read_line(int fd, char *p_buf, int bytes)
{
	static char read_buf[DATA_BUF_SIZE];
	static int buf_bytes = 0;
	static int buf_pos = 0;
	int read_bytes = 0;

	while( bytes > 0 )
	{
		if( buf_bytes == 0 && fd != -1 )
		{
			buf_bytes = read(fd, read_buf, DATA_BUF_SIZE);

			if( buf_bytes > 0 )
			{
#ifdef	DEBUG_IN_PS
				if( g_in_log_fd > 0 )
					write(g_in_log_fd, read_buf, buf_bytes);
#endif
				buf_pos = 0;
			}
		}

		if( buf_bytes > 0 )
		{
			*p_buf = read_buf[buf_pos++];
			bytes--;
			buf_bytes--;
			read_bytes++;

			if( IS_RETURN(*p_buf) )
				break;

			p_buf++;
		}
		else if( buf_bytes < 0 )
		{
			if( errno == EINTR )
				continue;
		}
		else
			break;
	}

	return read_bytes;
}

static
int get_copies(char *p_buf)
{
	char *p_code = p_buf;
	char value[MAX_VALUE_LEN + 1];
	int value_len = 0;
	int copies = -1;
	memset(value, 0, MAX_VALUE_LEN + 1);

	while(IS_BLANK(*p_code))
		p_code++;

	while(isdigit(*p_code))
	{
		value[value_len++] = *p_code++;
		if(value_len == MAX_VALUE_LEN)
			break;
	}

	if(value_len > 0)
		copies = atoi(value);

	return copies;
}

static
ParamList *get_ps_params(int ifd, BufList **ps_data, int *copies)
{
	char read_buf[DATA_BUF_SIZE];
	ParamList *p_list = NULL;
	int begin_page = 0;
	int read_bytes;
	int acro_util = 0;
	BufList *bl = NULL;
	BufList *prev_bl = NULL;
	int numcopies = -1;
	int is_banner = 0;
	char *p_read_buf;

	while( (read_bytes = read_line(ifd, read_buf, DATA_BUF_SIZE - 1)) > 0 )
	{
		{
			if( is_acro_util(read_buf, read_bytes) ){
				acro_util = 1;
			}
			else if( acro_util && is_end_resource(read_buf, read_bytes) ){
				acro_util = 0;
			}

			if( acro_util ){
				int line_bytes=0;
				while( line_bytes+29 < read_bytes ){
					if(!strncmp(&read_buf[line_bytes],
								" ct_BadResourceImplementation?", 30)){
						strcpy(&read_buf[line_bytes], " false");
						line_bytes+=6;
						strcpy(&read_buf[line_bytes],
							   &read_buf[line_bytes+24]);
						read_bytes-=24;
					}
					else{
						line_bytes++;
					}
				}
			}
		}

		bl = buflist_new((unsigned char *)read_buf, read_bytes);

		if( *ps_data == NULL )
			*ps_data = bl;
		else
			buflist_add_tail(prev_bl, bl);

		prev_bl = bl;

		if( read_bytes > 0 )
		{
			if( read_buf[read_bytes - 1] == '\n' )
				read_buf[read_bytes - 1] = '\0';
			else
				read_buf[read_bytes] = '\0';
		}
		else
		{
			read_buf[0] = '\0';
		}

		if( strncmp(read_buf, "%%BeginFeature:", 15) == 0 )
		{
			char key_buf[MAX_KEY_LEN + 1];
			char value_buf[MAX_VALUE_LEN + 1];
			int key_len = 0;
			int value_len = 0;
			char *p_code;

			p_code = read_buf + 15;

			while( *p_code != '\0' )
				if( *p_code++ == '*' )
					break;

			while( *p_code != '\0' )
			{
				if( IS_BLANK(*p_code)
				 || key_len >= MAX_KEY_LEN )
					break;
				key_buf[key_len++] = *p_code++;
			}
			while( *p_code != '\0' )
			{
				if( !IS_BLANK(*p_code)  )
					break;
				p_code++;
			}
			while( *p_code != '\0' )
			{
				if( IS_BLANK(*p_code)
				 || value_len >= MAX_VALUE_LEN )
					break;
				value_buf[value_len++] = *p_code++;
			}
			if( key_len > 0 && value_len > 0 )
			{
				key_buf[key_len] = '\0';
				value_buf[value_len] = '\0';

				param_list_add(&p_list, key_buf, value_buf, value_len + 1);
			}
		}
		else if( !begin_page && strncmp(read_buf, "%%Page:", 7) == 0 )
		{
			begin_page = 1;
		}
		else if((p_read_buf = strstr(read_buf, "NumCopies")) != NULL)
		{
			int tmp = get_copies(p_read_buf + 9);
			if(tmp > 0)
				numcopies = tmp;
		}
		else if(strncmp(read_buf, "/#copies", 8) == 0)
		{
			int tmp = get_copies(read_buf + 8);
			if(tmp > 0)
				numcopies = tmp;
		}
		else if(strncmp(read_buf, BANNER_DESC, BANNER_DESC_LEN) == 0)
		{
			is_banner = 1;
		}
		else if( begin_page )
		{
			if( strncmp(read_buf, "%%EndPageSetup", 14) == 0 )
				break;
			else if( strncmp(read_buf, "gsave", 5) == 0 )
				break;
			else if( read_buf[0] >= '0' && read_buf[0] <= '9' )
				break;
		}
	}

	while( (read_bytes = read_line(-1, read_buf, DATA_BUF_SIZE - 1)) > 0 )
	{
		BufList *bl = buflist_new((unsigned char *)read_buf, read_bytes);

		if( *ps_data == NULL )
			*ps_data = bl;
		else
			buflist_add_tail(prev_bl, bl);

		prev_bl = bl;

		if( read_buf[read_bytes - 1] == '\n' )
			read_buf[read_bytes - 1] = '\0';
		else
			read_buf[read_bytes] = '\0';
	}
	if(numcopies > 0)
	{
		*copies = numcopies;
	}
	if(is_banner)
	{
		*copies = -2;
	}

	return p_list;
}

static
void mark_ps_param_options(ppd_file_t *p_ppd, ParamList *p_param)
{
	int num = param_list_num(p_param);

	if( num > 0 )
	{
		cups_option_t *options
			= (cups_option_t*)malloc(num * sizeof(cups_option_t));
		if( options )
		{
			int i;
			for( i = 0 ; i < num ; i++ )
			{
				options[i].name = p_param->key;
				options[i].value = p_param->value;
				p_param = p_param->next;
			}
			for( i = num - 1 ; i >= 0 ; i-- )
			{
				cupsMarkOptions(p_ppd, 1, &options[i]);
			}

			free(options);
		}
	}
}

static
char *strdup2(char *str1, char *str2)
{
	int len;
	char *p_str;

	if( str1 == NULL || str2 == NULL )
		return NULL;

	len = strlen(str1) + strlen(str2) + 1;
	if( (p_str = malloc(len)) != NULL )
	{
		strcpy(p_str, str1);
		strcat(p_str, str2);
	}
	return p_str;
}

static
void free_cmd_arg(char **cmd_arg)
{
	if (cmd_arg != NULL) {
		int i = 0;

		for( i = 0 ; cmd_arg[i] != NULL ; i++ )
		{
			free(cmd_arg[i]);
			cmd_arg[i] = NULL;
		}
		free(cmd_arg);
	}
}

static
void set_cmd_arg(char *cmd_arg[], int index, char *str)
{
	if( cmd_arg[index] != NULL )
		free(cmd_arg[index]);

	cmd_arg[index] = str;
}


static
int make_cmd_arg(cups_option_t *p_cups_opt, int num_opt,
	ParamList *p_param, char *cmd_arg[], int copies, char *user, char *title)
{
#ifdef DEBUG_PPD
	char *p_ppd_name = "debug.ppd";
#else
	char *p_ppd_name = getenv("PPD");
#endif
	ppd_file_t *p_ppd;
	char device[PARAM_BUF_SIZE];
	char driver[PARAM_BUF_SIZE];
	char model[PARAM_BUF_SIZE];
	char reso_buf[PARAM_BUF_SIZE];
	char width_buf[PARAM_BUF_SIZE];
	char height_buf[PARAM_BUF_SIZE];
	ppd_choice_t *p_choice;
	ppd_size_t *p_size;
	char *p_size_name;
	int reso;
	int len = 0;
	char *p_job_attr = NULL;
	char *p_comp_job_attr = NULL;

	if( (p_ppd = ppdOpenFile(p_ppd_name)) == NULL )
		return -1;

	ppdMarkDefaults(p_ppd);
	cupsMarkOptions(p_ppd, num_opt, p_cups_opt);
	mark_ps_param_options(p_ppd, p_param);

	p_choice = ppdFindMarkedChoice(p_ppd, "PageSize");
	if( p_choice == NULL )
		p_choice = ppdFindMarkedChoice(p_ppd, "PageRegion");
	p_size_name = p_choice->choice;
	p_size = ppdPageSize(p_ppd, p_size_name);

	p_choice = ppdFindMarkedChoice(p_ppd, "Resolution");
	reso = atoi(p_choice->choice);

	snprintf(reso_buf, PARAM_BUF_SIZE-1, "%s%d",
		cmd_arg[ARGV_RESO], reso);
	snprintf(width_buf, PARAM_BUF_SIZE-1, "%s%d",
		cmd_arg[ARGV_WIDTH], (int)(p_size->width * (float)reso / 72.0));
	snprintf(height_buf, PARAM_BUF_SIZE-1, "%s%d",
		cmd_arg[ARGV_HEIGHT], (int)(p_size->length * (float)reso / 72.0));

	p_job_attr = make_job_attr(
		p_ppd, p_ppd_name, p_cups_opt, num_opt, copies, user, title);
	ppdClose(p_ppd);

	if( get_driver_model(
			p_ppd_name, model, driver, device, PARAM_BUF_SIZE) == 0 )
	{
		if( p_job_attr != NULL )
			free(p_job_attr);
		return -1;
	}

	if (p_job_attr != NULL )
	{
		len = strlen(p_job_attr) + strlen(cmd_arg[ARGV_JOB]);
		if (len >= JOBINFO_STR_MAX)
		{
			p_comp_job_attr = get_compress_job_attr(p_job_attr);

			if (p_comp_job_attr != NULL)
			{
				free(p_job_attr);
				p_job_attr = p_comp_job_attr;
			}
			else
			{
				debuglog(INDENT_NONE, "Fail to get compress job info.\n");
			}
		}
	}

	if( p_job_attr != NULL )
	{
		{
			char *p_gs_path = strdup2(EXEC_PATH, "/");
			set_cmd_arg(cmd_arg, ARGV_GS,
						strdup2(p_gs_path, EXEC_GS));
			free(p_gs_path);
		}
		set_cmd_arg(cmd_arg, ARGV_RESO, strdup(reso_buf));
		set_cmd_arg(cmd_arg, ARGV_WIDTH, strdup(width_buf));
		set_cmd_arg(cmd_arg, ARGV_HEIGHT, strdup(height_buf));
		set_cmd_arg(cmd_arg, ARGV_DEVICE,
						strdup2(cmd_arg[ARGV_DEVICE], device));
		set_cmd_arg(cmd_arg, ARGV_DRIVER,
						strdup2(cmd_arg[ARGV_DRIVER], driver));
		set_cmd_arg(cmd_arg, ARGV_MODEL,
						strdup2(cmd_arg[ARGV_MODEL], model));
		set_cmd_arg(cmd_arg, ARGV_JOB,
						strdup2(cmd_arg[ARGV_JOB], p_job_attr));

		free(p_job_attr);

		return 0;
	}
	else
		return -1;
}

int exec_filter(char *cmd_arg[], int ofd, int fds[2])
{
	int	child_pid = -1;

	if( pipe(fds) >= 0 )
	{
		child_pid = fork();

		if( child_pid == 0 )
		{
			close(0);
			dup2(fds[0], 0);
			close(fds[0]);
			close(fds[1]);

			if( ofd != 1 )
			{
				close(1);
				dup2(ofd, 1);
				close(ofd);
			}
			if( g_in_debug_fd > 0 ) {
				const time_t timer = time(NULL);
				char * const timeStr = ctime(&timer);

				FILE *fp_log = NULL;
				unsigned int loop_cnt= 0;

				fp_log = fopen( "/tmp/debug_optionlist.log", "w" );
				if( fp_log != NULL ) {
					fprintf(fp_log, "\n%s", timeStr);
					for( loop_cnt = 0; cmd_arg[loop_cnt] != NULL; loop_cnt++ ) {
						fprintf(fp_log, "cmd_arg[%d] : %s\n", loop_cnt, cmd_arg[loop_cnt]);
					}
					fclose(fp_log);
				}
			}

			execv(cmd_arg[0], cmd_arg);

			fprintf(stderr, "execl() error\n");
		}
		else if( child_pid != -1 )
		{
			close(fds[0]);
		}
	}
	return child_pid;
}

static
void
sigterm_handler(int sigcode)
{
	if( g_filter_pid != -1 )
		kill(g_filter_pid, SIGTERM);

	g_signal_received = 1;
}

int	main(int argc, char *argv[])
{
	cups_option_t *p_cups_opt = NULL;
	int num_opt = 0;
	ParamList *p_ps_param = NULL;
	BufList *p_ps_data = NULL;
	char *p_data_buf;
	int ifd = 0;
	int fds[2];
	struct sigaction sigact;
	char **cmd_arg = NULL;
	int cmd_arg_num = 0;
	int copies = -1;
#ifdef	UNIT_TEST
	freopen( "/tmp/unit_test/pstolipscpcar_leak_check.log", "a+", stderr);
#endif

#ifdef	DEBUG_SLEEP
	sleep(30);
#endif

	setbuf(stderr, NULL);
	fprintf(stderr, "DEBUG: pstolipscpcar start.\n");

	memset(&sigact, 0, sizeof(sigact));
	sigact.sa_handler = sigterm_handler;
	if( sigaction(SIGTERM, &sigact, NULL) )
	{
		fputs("ERROR: pstolipscpcar can't register signal hander.\n", stderr);
		return 1;
	}

	if( argc < 6 || argc > 7 )
	{
		fputs("ERROR: pstolipscpcar illegal parameter number.\n", stderr);
		return 1;
	}

	if( argv[5] != NULL )
	{
		num_opt = cupsParseOptions(argv[5], 0, &p_cups_opt);
		if( num_opt < 0 )
		{
			fputs("ERROR: illegal option.\n", stderr);
			return 1;
		}
	}
	if( argc == 7 )
	{
		if( (ifd = open(argv[6], O_RDONLY)) == -1 )
		{
			fputs("ERROR: can't open file.\n", stderr);
			return 1;
		}
	}

	cmd_arg = init_cmd_arg(org_cmd_arg, ARGV_NUM, &cmd_arg_num);
	if (cmd_arg == NULL) {
		fputs("ERROR: can't init parameter.\n", stderr);
		goto error_return;
	}

#ifdef	DEBUG_IN_PS
	g_in_log_fd = open("/tmp/debug_in_log.ps", O_WRONLY);
#endif
	p_ps_param = get_ps_params(ifd, &p_ps_data, &copies);

	if( make_cmd_arg(p_cups_opt, num_opt, p_ps_param, cmd_arg, copies, argv[2], argv[3]) < 0 )
	{
		fputs("ERROR: can't make parameter.\n", stderr);
		goto error_return;
	}

	if( g_signal_received )
		goto error_return;

	if( (p_data_buf = (char*)malloc(DATA_BUF_SIZE)) != NULL )
	{
#ifdef	DEBUG_PS
		int log_fd = open("/tmp/debug_data_log.ps", O_WRONLY);
#endif
		g_filter_pid = exec_filter(cmd_arg, 1, fds);

		buflist_write(p_ps_data, fds[1]);

#ifdef	DEBUG_PS
		buflist_write(p_ps_data, log_fd);
#endif
		if( g_in_debug_fd > 0 ) {
			buflist_write(p_ps_data, g_in_debug_fd);
		}

		while( !g_signal_received )
		{
			int read_bytes = read(ifd, p_data_buf, DATA_BUF_SIZE);

			if( read_bytes > 0 )
			{
				int write_bytes;
				char *p_data = p_data_buf;
#ifdef	DEBUG_PS
				write(log_fd, p_data_buf, read_bytes);
#endif
				if( g_in_debug_fd > 0 ) {
					write(g_in_debug_fd, p_data_buf, read_bytes);
				}
#ifdef	DEBUG_IN_PS
				if( g_in_log_fd > 0 )
					write(g_in_log_fd, p_data_buf, read_bytes);
#endif
				do
				{
					write_bytes = write(fds[1], p_data, read_bytes);

					if( write_bytes < 0 )
					{
						if( errno == EINTR )
							continue;
						fprintf(stderr,
							"ERROR: pstolipscpcar write error,%d.\n", errno);
						goto error_exit;
					}
					read_bytes -= write_bytes;
					p_data += write_bytes;
				}
				while( !g_signal_received && read_bytes > 0 );
			}
			else if( read_bytes < 0 )
			{
				if( errno == EINTR )
					continue;
				fprintf(stderr, "ERROR: pstolipscpcar read error,%d.\n", errno);
				goto error_exit;
			}
			else
				break;
		}

		free(p_data_buf);

#ifdef	DEBUG_PS
		if( log_fd != -1 ) close(log_fd);
#endif
	}

#ifdef	DEBUG_IN_PS
	if( g_in_log_fd > 0 )
		close(g_in_log_fd);
#endif
	if( g_in_debug_fd > 0 ) {
		close(g_in_debug_fd);
	}

	free_cmd_arg(cmd_arg);
	cmd_arg = NULL;

	if( p_ps_param != NULL )
		param_list_free(p_ps_param);
	if( p_ps_data != NULL )
		buflist_destroy(p_ps_data);
	if( ifd != 0 )
		close(ifd);

	close(fds[1]);

	if( g_filter_pid != -1 )
		waitpid(g_filter_pid, NULL, 0);

	return 0;

error_exit:
	if( p_data_buf != NULL )
		free(p_data_buf);

error_return:
	free_cmd_arg(cmd_arg);
	cmd_arg = NULL;

	if( p_ps_param != NULL )
		param_list_free(p_ps_param);
	if( p_ps_data != NULL )
		buflist_destroy(p_ps_data);
	if( ifd != 0 )
		close(ifd);

	return 1;
}

